package sequoia.path.alignment;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import sequoia.path.alignment.estimation.ScoreEvaluationParameters;
import sequoia.path.alignment.estimation.ScoreEvaluationParameters.ReadingFrameDifference;
import sequoia.path.alignment.filter.AlignmentFilter;
import sequoia.path.graph.BackTranslationGraph;
import sequoia.path.graph.Node;
import sequoia.path.scores.ScoreMatrix;
import sequoia.path.util.Codon;
import sequoia.path.util.DBSequence;
import sequoia.path.util.GenomicSequenceUtils;

public abstract class Alignment
{
    protected static boolean doTraceback = true;

    protected static final Log LOG = LogFactory.getLog(Alignment.class);

    public static final int DEFAULT_LIN_GAP_PENALTY = -1;

    public static final int DEFAULT_AFINE_GAP_OPEN_PENALTY = -4;

    public static final int DEFAULT_AFINE_GAP_EXTEND_PENALTY = -1;

    public static final int DEFAULT_MATCH_SCORE = 1;

    public static final int DEFAULT_MISMATCH_PENALTY = -1;

    public static final int DEFAULT_CODONSKIP_PENALTY = -3;

    public static final int DEFAULT_ALLOWED_FRAMESHIFTS = 4; // Integer.MAX_VALUE;

    public static final int DEFAULT_MIN_SCORE = 3;

    protected ScoreStatistics stat = new ScoreStatistics();

    protected boolean useCodonUsage = false;

    public boolean usesCodonUsage()
    {
        return this.useCodonUsage;
    }

    public void useCodonUsage()
    {
        this.useCodonUsage = true;
    }

    public void useCodonUsage(boolean use)
    {
        this.useCodonUsage = use;
    }

    protected static ScoreEvaluationParameters evaluator;

    protected static AlignmentFilter filter;

    public static void setFilter(AlignmentFilter f)
    {
        filter = f;
    }

    public BackTranslationGraph A, B;

    /** The protein sequences */
    protected DBSequence pseqA, pseqB;

    public DBSequence getPseqA()
    {
        return this.pseqA;
    }

    public void setPseqA(DBSequence pseqA)
    {
        this.pseqA = pseqA;
    }

    public DBSequence getPseqB()
    {
        return this.pseqB;
    }

    public void setPseqB(DBSequence pseqB)
    {
        this.pseqB = pseqB;
    }

    CellInfo alignmentMatrix[][];

    /** Alignment score */
    protected double bestScore = Double.MIN_VALUE;

    /** Performed frameshifts in the best alignment */
    protected int performedFrameshifts = 0;

    public int getPerformedFrameshifts()
    {
        return this.performedFrameshifts;
    }

    /** List of alignment fragments (within which the reading frame difference is conserved) */
    protected ArrayList<FrameshiftFragment> fragments = new ArrayList<FrameshiftFragment>();

    public List<FrameshiftFragment> getFragments()
    {
        return Collections.unmodifiableList(this.fragments);
    }

    /** Match score. */
    protected int match;

    /** Mismatch penalty. */
    protected int mismatch;

    /** Transition mismatch penalty. */
    protected int tmismatch;

    /** Gap opening penalty. */
    protected int d;

    /** Gap extension penalty. */
    protected int e;

    /** Size 3 gap penalty. */
    protected int codonSkip;

    /** Size 3 gap extension penalty. */
    protected int codonSkipExtension;

    public int allowedFrameshifts;

    // protected ScoreMatrix scoreMatrix = null;
    protected SubstitutionScoreOracle scores = null;

    protected AlignmentTracer tracer;

    // ************************************************************************
    // C O N S T R U C T O R S
    public Alignment(BackTranslationGraph A, BackTranslationGraph B)
    {
        this.d = DEFAULT_AFINE_GAP_OPEN_PENALTY;
        this.e = DEFAULT_AFINE_GAP_EXTEND_PENALTY;
        this.match = DEFAULT_MATCH_SCORE;
        this.mismatch = DEFAULT_MISMATCH_PENALTY;
        this.tmismatch = DEFAULT_MISMATCH_PENALTY;
        this.codonSkip = DEFAULT_CODONSKIP_PENALTY;
        this.codonSkipExtension = DEFAULT_CODONSKIP_PENALTY;
        this.allowedFrameshifts = DEFAULT_ALLOWED_FRAMESHIFTS;
        init(A, B);
    }

    public Alignment(BackTranslationGraph A, BackTranslationGraph B, int d, int e, int m, int x)
    {
        this.d = d;
        this.e = e;
        this.match = m;
        this.mismatch = x;
        this.tmismatch = x;
        this.codonSkip = DEFAULT_CODONSKIP_PENALTY;
        this.codonSkipExtension = DEFAULT_CODONSKIP_PENALTY;
        this.allowedFrameshifts = DEFAULT_ALLOWED_FRAMESHIFTS;
        init(A, B);
    }

    public Alignment(BackTranslationGraph A, BackTranslationGraph B, int d, int e, int m, int x, int tx)
    {
        this.d = d;
        this.e = e;
        this.match = m;
        this.mismatch = x;
        this.tmismatch = tx;
        this.codonSkip = DEFAULT_CODONSKIP_PENALTY;
        this.codonSkipExtension = DEFAULT_CODONSKIP_PENALTY;
        this.allowedFrameshifts = DEFAULT_ALLOWED_FRAMESHIFTS;
        init(A, B);
    }

    // ************************************************************************
    // I N I T I A L I Z A T I O N S

    public void setProteinSeqs(DBSequence pA, DBSequence pB)
    {
        this.pseqA = pA;
        this.pseqB = pB;
    }

    public void setScoreMatrix(ScoreMatrix scoreMatrix)
    {
        this.scores = new SubstitutionScoreOracle(scoreMatrix, this.match, this.mismatch, this.tmismatch);
    }

    public ScoreMatrix getScoreMatrix()
    {
        return this.scores.getScoreMatrix();
    }

    public SubstitutionScoreOracle getScoreOracle()
    {
        return this.scores;
    }

    public abstract void init(BackTranslationGraph A, BackTranslationGraph B);

    // ************************************************************************
    // S C O R E S / P E N A L T I E S
    protected double getGapPenalty(int crtDi, int crtDj, CellUnitInfo prevCell)
    {
        LOG.debug("Computing gap penalty for " + crtDi + " and " + crtDj);
        double score;
        if (crtDi == -Codon.CODON_LEN || crtDj == -Codon.CODON_LEN) {
            // if (prevCell.prevCell != null && Math.signum(prevCell.prevCell.Di) == Math.signum(crtDi)
            // && Math.signum(prevCell.prevCell.Dj) == Math.signum(crtDj)) {
            if (prevCell.prevCell != null && prevCell.prevCell.Di == crtDi && prevCell.prevCell.Dj == crtDj) {
                return this.codonSkipExtension;
            } else {
                return this.codonSkip;
            }
        }
        if (prevCell == null) {
            score = this.d;
            // } else if (Math.signum(prevCell.prevCell.Di) == Math.signum(crtDi)
            // && Math.signum(prevCell.prevCell.Dj) == Math.signum(crtDj)) {
        } else if (prevCell.prevCell.Di == crtDi && prevCell.prevCell.Dj == crtDj) {
            score = this.e;
        } else {
            score = this.d;
        }
        // stat.registerScore(score);
        return score;
    }

    protected double getScore(Node a, Node b)
    {
        double score = getRawScore(a, b);
        if (this.useCodonUsage) {
            if (a.getAA() != b.getAA()) {
                score += Math.log10(a.getWeight() * b.getWeight());
            }
        }
        this.stat.registerScore(score);
        return score;
    }

    protected double getRawScore(Node a, Node b)
    {
        if (this.scores != null) {
            double score = this.scores.getScore(a, b);
            return score;
        }
        if (a.getSymbol() == b.getSymbol()) {
            return this.match;
        } else if (GenomicSequenceUtils.isTransitionMutation(a.getSymbol(), b.getSymbol())) {
            return this.tmismatch;
        } else {
            return this.mismatch;
        }
    }

    /**
     * Compute the score of the cell at the given coordinates
     * 
     * @param i
     * @param j
     * @param k
     */
    protected abstract void computeBestPartialScore(int i, int j, CellUnitKey k);

    // Various choices for recursive score computation
    /**
     * match / mismatch
     */
    protected void partialScoreOptionMatch(CellUnitInfo c, Node aNodes[], Node bNodes[], int i, int j, CellUnitKey k)
    {
        int s, t;
        CellUnitInfo src;

        for (s = 0; s < aNodes.length; ++s) {
            for (t = 0; t < bNodes.length; ++t) {

                src = getCellUnit(i - 1, j - 1, aNodes[s], bNodes[t]);
                if (src != null) {
                    double thisscore = getScore(k.nA, k.nB) + src.score;
                    if (thisscore > c.score) {
                        c.score = thisscore;
                        c.prevCell = new Coord(-1, -1, aNodes[s], bNodes[t]);
                        if (src.score != 0) {
                            c.startedAt = src.getMyStart();
                        }
                        c.frameshiftCount = src.frameshiftCount;
                    }
                } else {
                    LOG.error("partialScoreOptionMatch: attempting to use unexisting node pair: " + (i - 1) + " "
                        + (j - 1) + " " + aNodes[s] + " " + bNodes[t]);
                }
            }
        }
    }

    /**
     * gap -- horizontal
     * 
     * @param c
     * @param aNodes
     * @param i
     * @param j
     * @param k
     */
    protected void partialScoreOptionHGap(CellUnitInfo c, Node aNodes[], int i, int j, CellUnitKey k)
    {
        int s;
        CellUnitInfo src;
        for (s = 0; s < aNodes.length; ++s) {
            src = getCellUnit(i - 1, j, aNodes[s], k.nB);
            double thisscore = getGapPenalty(-1, 0, src) + src.score;
            if (thisscore > c.score) {
                c.score = thisscore;
                c.prevCell = new Coord(-1, 0, aNodes[s], k.nB);
                c.startedAt = src.getMyStart();
                c.frameshiftCount = src.frameshiftCount;
            }
        }
    }

    /**
     * gap -- vertical
     * 
     * @param c
     * @param bNodes
     * @param i
     * @param j
     * @param k
     */
    protected void partialScoreOptionVGap(CellUnitInfo c, Node bNodes[], int i, int j, CellUnitKey k)
    {
        int t;
        CellUnitInfo src;
        for (t = 0; t < bNodes.length; ++t) {
            src = getCellUnit(i, j - 1, k.nA, bNodes[t]);
            double thisscore = getGapPenalty(0, -1, src) + src.score;
            if (thisscore > c.score) {
                c.score = thisscore;
                c.prevCell = new Coord(0, -1, k.nA, bNodes[t]);
                c.startedAt = src.getMyStart();
                c.frameshiftCount = src.frameshiftCount;
            }
        }
    }

    /**
     * frameshift -- horizontal
     * 
     * @param c
     * @param aNodes
     * @param i
     * @param j
     * @param k
     */
    protected void partialScoreOptionHFrameShift(CellUnitInfo c, Node aNodes[], int i, int j, CellUnitKey k)
    {
        partialScoreOptionHFrameShift(c, aNodes, i, j, k, 1);
    }

    protected void partialScoreOptionHFrameShift(CellUnitInfo c, Node aNodes[], int i, int j, CellUnitKey k, int gapSize)
    {
        int s;
        CellUnitInfo src;
        for (s = 0; s < aNodes.length; ++s) {
            src = getCellUnit(i - gapSize, j, aNodes[s], k.nB);
            if (src != null) {
                // boolean extendingGap = (src.prevCell.Di != 0 && src.prevCell.Dj
                // == 0);
                boolean extendingGap = (src.prevCell.Di == -gapSize && src.prevCell.Dj == 0);
                if (src.frameshiftCount < this.allowedFrameshifts || extendingGap) {
                    double thisscore = getGapPenalty(-gapSize, 0, src) + src.score;
                    if (thisscore > c.score) {
                        c.score = thisscore;
                        c.prevCell = new Coord(-gapSize, 0, aNodes[s], k.nB);
                        c.startedAt = src.getMyStart();
                        if (extendingGap) {
                            c.frameshiftCount = src.frameshiftCount;
                        } else {
                            c.frameshiftCount = src.frameshiftCount + 1;
                        }
                    }
                    // } else {
                    // System.out.println("H Frameshift not allowed: " + i + " " + j
                    // + " " + k);
                }
            } else {
                LOG.error("partialScoreOptionHFrameShift: attempting to use unexisting node pair: " + (i - 1) + " "
                    + (j - 1) + " " + aNodes[s] + " " + k.nB);
            }
        }
    }

    /**
     * frameshift -- vertical
     * 
     * @param c
     * @param aNodes
     * @param i
     * @param j
     * @param k
     */
    protected void partialScoreOptionVFrameShift(CellUnitInfo c, Node aNodes[], int i, int j, CellUnitKey k)
    {
        partialScoreOptionVFrameShift(c, aNodes, i, j, k, 1);
    }

    protected void partialScoreOptionVFrameShift(CellUnitInfo c, Node bNodes[], int i, int j, CellUnitKey k, int gapSize)
    {
        int t;
        CellUnitInfo src;
        for (t = 0; t < bNodes.length; ++t) {
            src = getCellUnit(i, j - gapSize, k.nA, bNodes[t]);
            if (src != null) {
                // boolean extendingGap = (src.prevCell.Di == 0 && src.prevCell.Dj
                // != 0);
                boolean extendingGap = (src.prevCell.Di == 0 && src.prevCell.Dj == -gapSize);
                if (src.frameshiftCount < this.allowedFrameshifts || extendingGap) {
                    double thisscore = getGapPenalty(0, -gapSize, src) + src.score;
                    if (thisscore > c.score) {
                        c.score = thisscore;
                        c.prevCell = new Coord(0, -gapSize, k.nA, bNodes[t]);
                        c.startedAt = src.getMyStart();
                        if (extendingGap) {
                            c.frameshiftCount = src.frameshiftCount;
                        } else {
                            c.frameshiftCount = src.frameshiftCount + 1;
                        }
                    }
                    // } else {
                    // System.out.println("H Frameshift not allowed: " + i + " " + j
                    // + " " + k);
                }
            } else {
                LOG.error("partialScoreOptionVFrameShift: attempting to use unexisting node pair: " + (i - 1) + " "
                    + (j - 1) + " " + k.nA + " " + bNodes[t]);
            }
        }
    }

    /**
     * codon gap -- horizontal
     * 
     * @param c
     * @param aNodes
     * @param i
     * @param j
     * @param k
     */
    protected void partialScoreOptionHCodonGap(CellUnitInfo c, Node aNodes[], int i, int j, CellUnitKey k)
    {
        int s;
        CellUnitInfo src;
        // if (A.isCodonskipPosition(i)) {
        for (s = 0; s < aNodes.length; ++s) {
            src = getCellUnit(i - Codon.CODON_LEN, j, aNodes[s], k.nB);
            double thisscore = getGapPenalty(-Codon.CODON_LEN, 0, src) + src.score;
            if (thisscore > c.score) {
                c.score = thisscore;
                c.prevCell = new Coord(-Codon.CODON_LEN, 0, aNodes[s], k.nB);
                c.startedAt = src.getMyStart();
                c.frameshiftCount = src.frameshiftCount;
            }
        }
        // }
    }

    /**
     * codon gap -- vertical
     * 
     * @param c
     * @param bNodes
     * @param i
     * @param j
     * @param k
     */
    protected void partialScoreOptionVCodonGap(CellUnitInfo c, Node bNodes[], int i, int j, CellUnitKey k)
    {
        int t;
        CellUnitInfo src;
        // if (B.isCodonskipPosition(j)) {
        for (t = 0; t < bNodes.length; ++t) {
            src = getCellUnit(i, j - Codon.CODON_LEN, k.nA, bNodes[t]);
            double thisscore = getGapPenalty(0, -Codon.CODON_LEN, src) + src.score;
            if (thisscore > c.score) {
                c.score = thisscore;
                c.prevCell = new Coord(0, -Codon.CODON_LEN, k.nA, bNodes[t]);
                c.startedAt = src.getMyStart();
                c.frameshiftCount = src.frameshiftCount;
            }
        }
        // }
    }

    // TODO intron gaps, when we have a DNA sequence to align to a protein
    // sequence

    // ************************************************************************
    /**
     * Actual alignment
     */
    public double align()
    {
        int i, j;
        // fill score matrix
        for (i = 1; i < this.alignmentMatrix.length; ++i) {
            for (j = 1; j < this.alignmentMatrix[i].length; ++j) {
                CellUnitKey[] cellKeys = this.alignmentMatrix[i][j].getKeys();
                for (CellUnitKey k : cellKeys) {
                    computeBestPartialScore(i, j, k);
                }
            }
        }
        // compute and return the score
        return computeBestScore();
    }

    public static void setDoTraceback(boolean doTraceback)
    {
        Alignment.doTraceback = doTraceback;
    }

    // ************************************************************************
    // Traceback functions
    abstract public String traceBack();

    abstract public String traceBack(CellUnitKey k);

    protected void updateFrameshiftPoints(int i, int j, CellUnitKey k)
    {
        CellUnitInfo cell = getCellUnit(i, j, k);
        double score = cell.score;
        Coord prev = cell.prevCell;
        if (prev.Di != prev.Dj && prev.Di + prev.Dj != -Codon.CODON_LEN) {
            insertFragmentStart(i, j, k, score);
            i += prev.Di;
            j += prev.Dj;
            k = prev.getFullKey();
            cell = getCellUnit(i, j, k);
            score = cell.score;
            if (cell != null) {
                if (cell.prevCell.Di != prev.Di || cell.prevCell.Dj != prev.Dj) {
                    // only for gap openings
                    insertFragmentEnd(i, j, k, cell.score);
                }
            }
        }
    }

    // displays: IDs of the aligned sequences, limits of the aligned zones if
    // necessary

    public AlignmentTracer getTracer()
    {
        return this.tracer;
    }

    /**
     * For fragments between frameshifts.
     * 
     * @param i
     * @param j
     * @param k
     */
    public void insertFragmentStart(int i, int j, CellUnitKey k, double score)
    {
        if (this.fragments.size() == 0) {
            // Create a fragment with this start
            this.fragments.add(new FrameshiftFragment(new Coord(i, j, k), score, null, score));
        } else {
            FrameshiftFragment fr = this.fragments.get(0);
            if (fr.from == null) {
                fr.setFrom(new Coord(i, j, k));
                fr.setScoreBefore(score);
            }
            // ignore otherwise
        }
    }

    /**
     * @param i
     * @param j
     * @param k
     */
    public void insertFragmentEnd(int i, int j, CellUnitKey k, double score)
    {
        if (this.fragments.size() == 0 || this.fragments.get(0).getFrom() != null) {
            this.fragments.add(0, new FrameshiftFragment(null, 0, new Coord(i, j, k), score));
        } else {
            FrameshiftFragment fr = this.fragments.get(0);
            fr.setTo(new Coord(i, j, k));
            fr.setScoreAfter(score);
        }
    }

    /**
     * Display the limits and scores for the alignment fragments between frameshifts
     * 
     * @param stream
     */
    public void displayFragments(PrintStream stream)
    {
        if (this.fragments.size() == 0) {
            stream.println("\nNo fragments.\n");
        } else {
            stream.println();
            for (FrameshiftFragment f : this.fragments) {
                stream.println(f);
            }
            stream.println();
        }
    }

    // ************************************************************************
    // Score & e-value
    abstract public double computeBestScore();

    public double getScore()
    {
        return this.bestScore;
    }

    public void setEvaluator(ScoreEvaluationParameters e)
    {
        evaluator = e;
    }

    public double getScoreSignificance()
    {
        return evaluator.evaluate(this.bestScore, this.getASize() * this.getBSize(), ReadingFrameDifference
            .valueOf(this.isRC()));
    }

    public double getScoreSignificance(FrameshiftFragment f)
    {
        return evaluator.evaluate(f.getFragmentScore(), this.getASize() * this.getBSize(), ReadingFrameDifference
            .valueOf(f.getPhaseDifference(this)));
    }

    // ************************************************************************
    // Data access functions
    public CellUnitInfo getCellUnit(Coord co)
    {
        return this.alignmentMatrix[co.Di][co.Dj].get(co.getFullKey());
    }

    public CellUnitInfo getCellUnit(int i, int j, CellUnitKey k)
    {
        return this.alignmentMatrix[i][j].get(k);
    }

    public CellInfo getCell(int i, int j)
    {
        return this.alignmentMatrix[i][j];
    }

    public CellUnitInfo getCellUnit(int i, int j, Node a, Node b)
    {
        return this.alignmentMatrix[i][j].get(a, b);
    }

    /**
     * For debug purposes only. TODO remove from the final version.
     * 
     * @return a string representation of the alignment matrix
     */
    public String displayMatrix()
    {
        StringBuilder str = new StringBuilder();
        for (int i = 0; i < this.alignmentMatrix.length; ++i) {
            str.append("==========================\n");
            str.append(i + "\n");
            str.append("--------------------------\n");
            for (int j = 0; j < this.alignmentMatrix[i].length; ++j) {
                str.append(" [" + i + ", " + j + "]\n");
                Set<CellUnitKey> keys = this.alignmentMatrix[i][j].keySet();
                for (CellUnitKey k : keys) {
                    str.append("  " + k.key + ": " + this.alignmentMatrix[i][j].get(k).score + ";\n");
                }
            }
        }
        return str.toString();
    }

    /**
     * Size of the first graph
     * 
     * @return The number of positions of the first graph.
     */
    public int getASize()
    {
        return this.A.size();
    }

    /**
     * Size of the second graph
     * 
     * @return The number of positions of the first graph.
     */
    public int getBSize()
    {
        return this.B.size();
    }

    /**
     * Is this a reverse complementary alignment?
     * 
     * @return True if one of the aligned graphs in reverse complementary, false otherwise
     */
    public boolean isRC()
    {
        return this.A.isRC() || this.B.isRC();
    }
}
