package sequoia.path.alignment;


import sequoia.path.alignment.format.PlainTextFormatter;
import sequoia.path.graph.Node;
import sequoia.path.util.Codon;
import sequoia.path.util.GenomicSequenceUtils;

public class AlignmentTracer
{
    protected StringBuilder strA, strB, matcher, strpA, strpB, strS;
    
	protected int startA, startB, endA, endB;

    public static final String GAP_KEY = "-";

    public static final String WILDCARD = "*";

    public static final char CGAP_KEY = '-';

    public static final String MATCH_KEY = "|";

    public static final char CMATCH_KEY = '|';

    public static final String TRMUT_KEY = ":";

    public static final char CTRMUT_KEY = ':';

    public static final String NOMATCH_KEY = ".";

    public static final char CNOMATCH_KEY = '.';

    public static final String MGAP_KEY = " ";

    public static final char CMGAP_KEY = ' ';

    public static final char AADISPLAY_OPEN = '[';

    public static final char AADISPLAY_CLOSE = ']';
    
    public final static char POSITIVE_SCORE = '+';
    public final static char ZERO_SCORE = ':';
    public final static char SMALL_POSITIVE_SCORE = '\'';
    public final static char SMALL_NEGATIVE_SCORE = ',';
    public final static char NEGATIVE_SCORE = ' ';
    public final static char UNDEFINED_SCORE = ' ';

    String prefix = "", prefixpA = "", prefixpB = "";

    Alignment forAlignment;

    // -----------------------------------------------------------------------------------------
    // Constructor
    public AlignmentTracer(Alignment a)
    {
        forAlignment = a;
        strA = new StringBuilder();
        strB = new StringBuilder();
        strpA = new StringBuilder();
        strpB = new StringBuilder();
        matcher = new StringBuilder();
        strS = new StringBuilder();
        startA = startB = endA = endB = -1;
    }
    // -----------------------------------------------------------------------------------------
    // When the traceback is completed, Formatter objects will need access to this information:
    public String getNucleotideSequenceA() {
		return strA.toString();
	}
	public String getNucleotideSequenceB() {
		return strB.toString();
	}
	public String getMatchSequence() {
		return matcher.toString();
	}
	public String getAASequenceA() {
		return strpA.toString();
	}
	public String getAASequenceB() {
		return strpB.toString();
	}
	public String getScoreSignSequence() {
		return strS.toString();
	}
	public int getStartA() {
		return startA;
	}
	public int getStartB() {
		return startB;
	}
	public int getEndA() {
		return endA;
	}
	public int getEndB() {
		return endB;
	}
	//-------------------------------------------------------------------------------------------
	/**
	 * Set the starting point of an alignment. Triggers the computation of string prefixes for correct display.
	 */
    public void setStart(int a, int b)
    {
        startA = a;// + forAlignment.A.getPhase();
        startB = b;// + forAlignment.B.getPhase();

        int shift = 0, shiftpA = 0, shiftpB = 0;
        final int CODON_LEN = Codon.CODON_LEN;

        int nstartPA = GenomicSequenceUtils.getPrevCodonStart(startA, forAlignment.A.getPhase(), true); // startA -
        if (nstartPA >= 0) {
            prefixpA += forAlignment.A.getProteinStr(nstartPA, startA);
        }
        int nstartPB = GenomicSequenceUtils.getPrevCodonStart(startB, forAlignment.B.getPhase(), true);// startB -
        if (nstartPB >= 0) {
            prefixpB += forAlignment.B.getProteinStr(nstartPB, startB);
        }

        shiftpA = startA % CODON_LEN;
        shiftpB = startB % CODON_LEN;
        shift = Math.max(shiftpA, shiftpB);
        shiftpA = shift - shiftpA;
        shiftpB = shift - shiftpB;
        for (int j = 0; j < shiftpA; ++j) {
            prefixpA += " ";
        }
        for (int j = 0; j < shiftpB; ++j) {
            prefixpB += " ";
        }
        for (int j = 0; j < shift; ++j) {
            prefix += " ";
        }

    }
    /**
     * Set the starting point of an alignment. Triggers the computation of string prefixes for correct display.
     * @param a
     * @param b
     * @param fullCodonSeq
     */
    public void setStart(int a, int b, boolean fullCodonSeq)
    {
        if (fullCodonSeq) {
            setStart(a, b);
        }
        startA = a;// + forAlignment.A.getPhase();
        startB = b;// + forAlignment.B.getPhase();
    }
	/**
	 * Set the starting point of an alignment. Triggers the computation of string suffixes for a more complete display.
	 */
    public void setEnd(int a, int b)
    {
        endA = a - 1;
        endB = b - 1;
        int j = endA + 1;
        while (j < GenomicSequenceUtils.getNextCodonStart(endA + 1, forAlignment.A.getPhase())
            && j < forAlignment.A.nodes.length)
        {
            strpA.insert(0, getAADisplayChar(forAlignment.A.nodes[j][0].getAA(), j, forAlignment.A.getPhase()));
            ++j;
        }
        j = endB + 1;
        while (j < GenomicSequenceUtils.getNextCodonStart(endB + 1, forAlignment.B.getPhase())
            && j < forAlignment.B.nodes.length)
        {
            strpB.insert(0, getAADisplayChar(forAlignment.B.nodes[j][0].getAA(), j, forAlignment.B.getPhase()));
            ++j;
        }
    }
	/**
	 * Set the starting point of an alignment. Triggers the computation of string suffixes for a more complete display.
	 */
    public void setEnd(int a, int b, boolean fullCodonSeq)
    {
        if (fullCodonSeq) {
            setEnd(a, b);
        }
        endA = a - 1;
        endB = b - 1;
    }

    void put(String pair, Coord prev)
    {
        put(pair, prev.Di, prev.Dj);
    }

    private void skipCodon(char forAA, String acodon, StringBuilder fromSeq, StringBuilder fromPSeq,
        StringBuilder inSeq, StringBuilder inPSeq)
    {
        for (int i = Codon.CODON_LEN - 1; i >= 0; --i) {
            fromSeq.append(acodon.charAt(i));
            fromPSeq.append(getAADisplayChar(forAA, i));
            inSeq.append(GAP_KEY);
            inPSeq.append(CMGAP_KEY);
            matcher.append(MGAP_KEY);
            strS.append(MGAP_KEY);
        }
    }

    String getMatchSymb(char cA, char cB)
    {
        if (GenomicSequenceUtils.isMatch(cA, cB)) {
            return MATCH_KEY;
        } else if (cA == CGAP_KEY || cB == CGAP_KEY) {
            return MGAP_KEY;
        } else if (GenomicSequenceUtils.isTransitionMutationExt(cA, cB)) {
            return TRMUT_KEY;
        } else {
            return NOMATCH_KEY;
        }
    }
    
    char getScoreSymbol(Node a, Node b, boolean gap) {
    	if (gap) return NEGATIVE_SCORE;
    	double EPSILON = 0.4;
    	double SEPSILON = 0.05;
    	
    	double score = forAlignment.getScoreOracle().getScore(a, b);
    	
    	if (Math.abs(score) < SEPSILON) {
    		return ZERO_SCORE;
    	}
    	if (score > 0) {
    		if (score <= EPSILON) {
    			return SMALL_POSITIVE_SCORE;
    		}
    		return POSITIVE_SCORE;
    	}
    	if (score >= -EPSILON) {
    	  return SMALL_NEGATIVE_SCORE;
    	}
    	return NEGATIVE_SCORE;
    }

    void put(CellUnitKey key, Coord prev, int i, int j)
    {
        int di = prev.Di, dj = prev.Dj;
        char cA = key.nA.getSymbol();
        char pA = key.nA.getAA();
        char cB = key.nB.getSymbol();
        char pB = key.nB.getAA();
        char cS = UNDEFINED_SCORE;
        if (this.forAlignment.getScoreOracle() != null) {
            String realSymbs = this.forAlignment.getScoreOracle().getNPair(key.nA, key.nB);
            cA = realSymbs.charAt(0);
            cB = realSymbs.charAt(1);
        }

        if (di == -Codon.CODON_LEN) {
            pA = this.forAlignment.A.nodes[i - 1][0].getAA();
            skipCodon(pA, forAlignment.A.getCodonAt(i - 1), strA, strpA, strB, strpB);
            return;
        }
        if (dj == -Codon.CODON_LEN) {
            pB = this.forAlignment.B.nodes[j - 1][0].getAA();
            skipCodon(pB, forAlignment.B.getCodonAt(j - 1), strB, strpB, strA, strpA);
            return;
        }
        String cM;
        pA = getAADisplayChar(pA, i - 1, this.forAlignment.A.getPhase());
        pB = getAADisplayChar(pB, j - 1, this.forAlignment.B.getPhase());
        if (di == 0) {
            cA = GAP_KEY.charAt(0);
            pA = CMGAP_KEY;
        }
        if (dj == 0) {
            cB = GAP_KEY.charAt(0);
            pB = CMGAP_KEY;
        }
        cM = getMatchSymb(cA, cB);
        cS = getScoreSymbol(key.nA, key.nB, di == 0 || dj == 0);
        update(cA, pA, cB, pB, cM.charAt(0), cS);
    }

    void put(String pair, int di, int dj)
    {
        put(pair.charAt(0), pair.charAt(1), di, dj);
    }

    void put(char cA, char cB, int di, int dj)
    {
        String cM;
        if (di == 0) {
            cA = GAP_KEY.charAt(0);
        }
        if (dj == 0) {
            cB = GAP_KEY.charAt(0);
        }
        cM = getMatchSymb(cA, cB);
        update(cA, cB, cM.charAt(0));
    }

    private void update(char cA, char pA, char cB, char pB, char cM, char cS)
    {
        strpA.append(pA);
        strA.append(cA);
        matcher.append(cM);
        strB.append(cB);
        strpB.append(pB);
        strS.append(cS);
    }

    private void update(char cA, char cB, char cM)
    {
        strA.append(cA);
        matcher.append(cM);
        strB.append(cB);
    }

    char getAADisplayChar(char c, int i)
    {
        if (i % Codon.CODON_LEN == 0) {
            return AADISPLAY_OPEN;
        }
        if (i % Codon.CODON_LEN == 2) {
            return AADISPLAY_CLOSE;
        }
        return c;
    }

    char getAADisplayChar(char c, int i, int phase)
    {
        int actualI = ((i - phase) + Codon.CODON_LEN) % Codon.CODON_LEN;
        if (actualI == 0) {
            return AADISPLAY_OPEN;
        }
        if (actualI == 2) {
            return AADISPLAY_CLOSE;
        }
        return c;
    }

    public void append(AlignmentTracer other)
    {
        this.strpA.append(other.strpA);
        this.strA.append(other.strA);
        this.matcher.append(other.matcher);
        this.strS.append(other.strS);
        this.strB.append(other.strB);
        this.strpB.append(other.strpB);
    }

    public void append(AlignmentTracer other, boolean mergeEnds)
    {
        if (mergeEnds && strA.length() != 0) {
            strpA.deleteCharAt(strpA.length() - 1);
            strA.deleteCharAt(strA.length() - 1);
            matcher.deleteCharAt(matcher.length() - 1);
            strS.deleteCharAt(strS.length() - 1);
            strB.deleteCharAt(strB.length() - 1);
            strpB.deleteCharAt(strpB.length() - 1);
        }
        this.strpA.append(other.strpA);
        this.strA.append(other.strA);
        this.matcher.append(other.matcher);
        this.strS.append(other.strS);
        this.strB.append(other.strB);
        this.strpB.append(other.strpB);
    }

    void unreverse()
    {
        strA.reverse();
        strA.insert(0, prefix);
        strB.reverse();
        strB.insert(0, prefix);
        matcher.reverse();
        matcher.insert(0, prefix);
        strS.reverse();
        strS.insert(0, prefix);
        strpA.reverse();
        strpA.insert(0, prefixpA);
        strpB.reverse();
        strpB.insert(0, prefixpB);
    }

    @Override
    public String toString()
    {
    	return PlainTextFormatter.instance.format(forAlignment);
    }
}
