package sequoia.path.alignment;

import sequoia.path.util.Codon;

/**
 * Marks a fragment of the alignment between two frameshifts (or border).
 * The fragment position is delimited by two coordinates (Coord) in the alignment matrix.
 * The scores BEFORE and AFTER the fragment are also memorized. 
 *
 */
public class FrameshiftFragment {
	Coord from = null;
	Coord to = null;
	double scoreBefore = 0;
	double scoreAfter = 0;
	
	public FrameshiftFragment(int i1, int j1, CellUnitKey k1, double score1, int i2, int j2, CellUnitKey k2, double score2) {
		from = new Coord(i1, j1, k1);
		to = new Coord(i2, j2, k2);
		scoreBefore = score1;
		scoreAfter = score2;
	}
	
	public FrameshiftFragment(Coord c1, double score1, Coord c2, double score2) {
		if (c1 != null) {
		  from = new Coord(c1.Di, c1.Dj, c1.getFullKey());
			scoreBefore = score1;
		}
		if (c2 != null) {
			to = new Coord(c2.Di, c2.Dj, c2.getFullKey());
			scoreAfter = score2;
		}
	}
	/**
	 * Shift a fragment. Useful when reconstructing a larger alignment from fragments.
	 * @param coord
	 * @param score
	 * @return the new score after the fragment
	 */
	public double shift(Coord coord, double score) {
		if (coord != null) {
			if (from != null) {
			  from.Di += coord.Di;
			  from.Dj += coord.Dj;
			}
			if (to != null) {
			  to.Di += coord.Di;
			  to.Dj += coord.Dj;
			}
			scoreBefore += score;
			scoreAfter += score;
		} else {
			System.err.println("NULL coordinate pn shift");
		}
		return scoreBefore;
	}
	/**
	 * Shift a fragment, using the coordinate and the score after another fragment point. Useful when reconstructing a larger alignment from pieces.
	 * @param frameshiftFragment
	 * @return the new score after the fragment
	 */
	public double shift(FrameshiftFragment frameshiftFragment) {
		return shift(frameshiftFragment.getTo(), frameshiftFragment.scoreAfter);
	}

	public Coord getFrom() {
		return from;
	}

	public void setFrom(Coord from) {
		this.from = from;
	}

	public Coord getTo() {
		return to;
	}

	public void setTo(Coord to) {
		this.to = to;
	}
	
	public double getScoreBefore() {
		return scoreBefore;
	}
	
	public double getScoreAfter() {
		return scoreAfter;
	}

	public void setScoreBefore(double scoreBefore) {
		this.scoreBefore = scoreBefore;
	}

	public void setScoreAfter(double scoreAfter) {
		this.scoreAfter = scoreAfter;
	}
	
	public double getFragmentScore() {
		return scoreAfter - scoreBefore;
	}
	
	public int getPhaseDifference(Alignment alignment) {
		int phaseDiff;
		if (alignment.isRC()) {
			phaseDiff = -((alignment.A.getPhase(this.from.Di) - alignment.B.getPhase(this.from.Dj) + Codon.CODON_LEN) % Codon.CODON_LEN);
			if (phaseDiff == 0) {
				phaseDiff = -Codon.CODON_LEN;
			}
		} else {
		   phaseDiff = ((alignment.A.getPhase(this.from.Di) - alignment.B.getPhase(this.from.Dj) + Codon.CODON_LEN) % Codon.CODON_LEN);
		}
		return phaseDiff;
	}
	
	public String toString() {
		//return "[" + from + "(" + String.format("%.2f", scoreBefore) + ") -> " + to + "(" + String.format("%.2f", scoreAfter) + ")] : " + String.format("%.2f", getFragmentScore()); 
		return "[" + String.format("%4d", from.Di) + ", " + String.format("%4d", from.Dj) + "] -> [" + 
		String.format("%4d", to.Di) + ", " + String.format("%4d", to.Dj) + "]     >> Score: " + String.format("%.2f", getFragmentScore()) + "; ";
	}
	
}

