package sequoia.path.alignment;

import java.util.ArrayList;

import sequoia.path.graph.BackTranslationGraph;
import sequoia.path.graph.Node;

public abstract class GlobalDPAlignment extends Alignment
{
    // ************************************************************************
    // C O N S T R U C T O R S
    public GlobalDPAlignment(BackTranslationGraph A, BackTranslationGraph B)
    {
        super(A, B);
    }

    public GlobalDPAlignment(BackTranslationGraph A, BackTranslationGraph B, int d, int e, int m, int x)
    {
        super(A, B, d, e, m, x);
    }

    // ************************************************************************
    /**
     * Initialization
     */
    @Override
    public void init(BackTranslationGraph A, BackTranslationGraph B)
    {
        int i, j, k, l;
        // TODO make sure there are no border effects
        this.A = A;
        this.B = B;
        this.alignmentMatrix = new CellInfo[A.size() + 1][B.size() + 1];
        this.alignmentMatrix[0][0] = new CellInfo();
        CellUnitKey key = new CellUnitKey(CellUnitKey.UNDEF_KEY);
        this.alignmentMatrix[0][0].put(key, new CellUnitInfo(0));

        for (i = 1; i < this.alignmentMatrix.length; ++i) {
            this.alignmentMatrix[i][0] = new CellInfo();
            for (k = 0; k < A.nodes[i - 1].length; ++k) {
                CellUnitKey thePrevKey;
                if (A.nodes[i - 1][k].hasPred(Node.ANY)) {
                    thePrevKey = this.alignmentMatrix[i - 1][0].getKeys()[0];
                } else {
                    thePrevKey = new CellUnitKey(A.getPrecNodes(i - 1, A.nodes[i - 1][k])[0], Node.ANY_NODE);
                    // TODO change this [0] restriction
                }
                this.alignmentMatrix[i][0].put(new CellUnitKey(A.nodes[i - 1][k], Node.ANY_NODE), new CellUnitInfo(
                    this.d + (i - 1) * this.e, -1, 0, thePrevKey));
            }
        }

        for (j = 1; j < this.alignmentMatrix[0].length; ++j) {
            this.alignmentMatrix[0][j] = new CellInfo();
            for (k = 0; k < B.nodes[j - 1].length; ++k) {
                CellUnitKey thePrevKey;
                if (B.nodes[j - 1][k].hasPred(Node.ANY)) {
                    thePrevKey = this.alignmentMatrix[0][j - 1].getKeys()[0];
                } else {
                    thePrevKey = new CellUnitKey(Node.ANY_NODE, B.getPrecNodes(j - 1, B.nodes[j - 1][k])[0]);
                    // TODO change this [0] restriction
                }
                this.alignmentMatrix[0][j].put(new CellUnitKey(Node.ANY_NODE, B.nodes[j - 1][k]), new CellUnitInfo(
                    this.d + (j - 1) * this.e, 0, -1, thePrevKey));
            }
        }

        for (i = 1; i < this.alignmentMatrix.length; ++i) {
            Node[] aNodes = A.getNodes(i - 1);
            for (j = 1; j < this.alignmentMatrix[i].length; ++j) {
                Node[] bNodes = B.getNodes(j - 1);
                this.alignmentMatrix[i][j] = new CellInfo();
                for (k = 0; k < aNodes.length; ++k) {
                    for (l = 0; l < bNodes.length; ++l) {
                        this.alignmentMatrix[i][j].put(new CellUnitKey(aNodes[k], bNodes[l]), new CellUnitInfo(
                            Integer.MIN_VALUE));
                    }
                }
            }
        }
    }

    // ************************************************************************
    /**
     * Traceback
     */
    @Override
    public double computeBestScore()
    {
        getBestUnitKeys(this.alignmentMatrix[this.A.size()][this.B.size()]);
        return this.bestScore;
    }

    @Override
    public String traceBack()
    {
        return traceBack(getBestUnitKey(this.alignmentMatrix[this.A.size()][this.B.size()]));
    }

    @Override
    public String traceBack(CellUnitKey k)
    {
        this.tracer = new AlignmentTracer(this);
        int i = this.A.size();// alignmentMatrix.length - 1;
        int j = this.B.size();// alignmentMatrix[i].length - 1;
        this.tracer.setEnd(i, j);
        CellUnitKey what = k;
        Coord prev;

        insertFragmentEnd(i, j, k, this.alignmentMatrix[i][j].get(what).score);

        this.performedFrameshifts = this.alignmentMatrix[i][j].get(what).frameshiftCount;

        // System.out.println("###### fs: " + this.getPerformedFrameshifts() + "( " + i + " "+ j + " " + k + ")");

        while (what.key.length() == 2 && !what.key.contains(CellUnitKey.GAP_KEY)) {
            // !what.key.equals(CellUnitKey.UNDEF_KEY
            // System.out.println(what.key + "  " + what);
            prev = this.alignmentMatrix[i][j].get(what).prevCell;
            this.tracer.put(what, prev, i, j);
            this.updateFrameshiftPoints(i, j, what);
            i += prev.Di;
            j += prev.Dj;
            what = prev.key;
        }

        insertFragmentStart(i, j, what, 0);

        this.tracer.setStart(0, 0);
        this.tracer.unreverse();
        return this.tracer.toString();
    }

    // ************************************************************************
    /**
     * Key of the best scoring alignment ending in a given cell of the matrix
     */
    public CellUnitKey getBestUnitKey(CellInfo cell)
    {
        this.bestScore = Integer.MIN_VALUE;
        CellUnitKey bestKey = null;
        CellUnitKey[] poss = cell.getKeys();
        for (int i = 0; i < poss.length; ++i) {
            if (cell.get(poss[i]).score > this.bestScore) {
                this.bestScore = cell.get(poss[i]).score;
                bestKey = poss[i];
            }
        }
        return bestKey;
    }

    /**
     * List of all best scoring alignments ending in a given cell of the matrix
     * 
     * @param cell
     * @return
     */
    ArrayList<CellUnitKey> getBestUnitKeys(CellInfo cell)
    {
        this.bestScore = Integer.MIN_VALUE;
        ArrayList<CellUnitKey> bestKey = new ArrayList<CellUnitKey>();
        CellUnitKey[] poss = cell.getKeys();
        for (int i = 0; i < poss.length; ++i) {
            if (cell.get(poss[i]).score >= this.bestScore) {
                if (cell.get(poss[i]).score > this.bestScore) {
                    this.bestScore = cell.get(poss[i]).score;
                    bestKey.clear();
                }
                bestKey.add(poss[i]);
            }
        }
        return bestKey;
    }
}
