package sequoia.path.alignment;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import sequoia.path.graph.BackTranslationGraph;
import sequoia.path.graph.Node;

public abstract class OverlapDPAlignment extends Alignment
{
    private static final Log LOG = LogFactory.getLog(OverlapDPAlignment.class);

    // ************************************************************************
    // C O N S T R U C T O R S
    public OverlapDPAlignment(BackTranslationGraph A, BackTranslationGraph B)
    {
        super(A, B);
    }

    public OverlapDPAlignment(BackTranslationGraph A, BackTranslationGraph B, int d, int e, int m, int x)
    {
        super(A, B, d, e, m, x);
    }

    // ************************************************************************
    // I N I T I A L I Z A T I O N
    @Override
    public void init(BackTranslationGraph A, BackTranslationGraph B)
    {
        LOG.debug("init");
        int i, j, k, l;
        // init
        this.A = A;
        this.B = B;
        this.alignmentMatrix = new CellInfo[A.size() + 1][B.size() + 1];
        this.alignmentMatrix[0][0] = new CellInfo();
        CellUnitKey key = new CellUnitKey(CellUnitKey.UNDEF_KEY);
        this.alignmentMatrix[0][0].put(key, new CellUnitInfo(0));

        for (i = 1; i < this.alignmentMatrix.length; ++i) {
            this.alignmentMatrix[i][0] = new CellInfo();
            for (k = 0; k < A.nodes[i - 1].length; ++k) {
                CellUnitKey ky = new CellUnitKey(A.nodes[i - 1][k], Node.ANY_NODE);
                CellUnitInfo c = new CellUnitInfo(0);
                c.setMyStart(new Coord(i, 0, ky));
                this.alignmentMatrix[i][0].put(ky, c);
            }
        }

        for (j = 1; j < this.alignmentMatrix[0].length; ++j) {
            this.alignmentMatrix[0][j] = new CellInfo();
            for (k = 0; k < B.nodes[j - 1].length; ++k) {
                CellUnitKey ky = new CellUnitKey(Node.ANY_NODE, B.nodes[j - 1][k]);
                CellUnitInfo c = new CellUnitInfo(0);
                c.setMyStart(new Coord(0, j, ky));
                this.alignmentMatrix[i][0].put(ky, c);
            }
        }

        for (i = 1; i < this.alignmentMatrix.length; ++i) {
            Node[] aNodes = A.getNodes(i - 1);
            if (i % 10 == 0) {
                LOG.debug("Init " + i);
            }
            for (j = 1; j < this.alignmentMatrix[i].length; ++j) {
                Node[] bNodes = B.getNodes(j - 1);
                this.alignmentMatrix[i][j] = new CellInfo();
                for (k = 0; k < aNodes.length; ++k) {
                    for (l = 0; l < bNodes.length; ++l) {
                        this.alignmentMatrix[i][j].put(new CellUnitKey(aNodes[k], bNodes[l]), new CellUnitInfo());
                    }
                }
            }
        }
    }

    // ************************************************************************
    // C O M P U T E T H E S C O R E O F E A C H C E L L
    @Override
    protected abstract void computeBestPartialScore(int i, int j, CellUnitKey k);

    // ************************************************************************
    // T R A C E B A C K F U N C T I O N S
    //
    @Override
    public double computeBestScore()
    {
        getBestCells();
        return this.bestScore;
    }

    @Override
    public String traceBack()
    {
        // System.out.println("traceback in local");
        StringBuilder str = new StringBuilder("\n");
        ArrayList<Coord> maxScorePoints;
        maxScorePoints = getBestCells();

        if (doTraceback) {
            for (int i = 0; i < maxScorePoints.size(); ++i) {
                str.append(traceBack(maxScorePoints.get(i)));
                str.append("\n");
                break;
            }
        }
        return str.toString();
    }

    String traceBack(Coord from)
    {
        // System.out.println("traceback in local2");
        this.tracer = new AlignmentTracer(this);
        int i = from.Di;
        int j = from.Dj;
        double score;
        this.tracer.setEnd(i, j);
        CellUnitKey what = from.key;
        Coord next;

        this.performedFrameshifts = this.alignmentMatrix[i][j].get(what).frameshiftCount;

        do {
            next = this.alignmentMatrix[i][j].get(what).prevCell;
            this.tracer.put(what, next, i, j);
            i += next.Di;
            j += next.Dj;
            what = next.key;
            score = this.alignmentMatrix[i][j].get(what).score;
        } while (score != 0);
        this.tracer.setStart(i, j);
        this.tracer.unreverse();
        return this.tracer.toString();
        // return tracer.getStrAlignment();
    }

    // ************************************************************************
    // B E S T A L I G N M E N T I N T H E M A T R I X S E A R C H
    //
    CellUnitKey getBestUnitKey(CellInfo cell)
    {
        double bestScore = Double.MIN_VALUE;
        CellUnitKey bestKey = null;
        CellUnitKey[] poss = cell.getKeys();
        for (int i = 0; i < poss.length; ++i) {
            if (cell.get(poss[i]).score > bestScore) {
                bestScore = cell.get(poss[i]).score;
                bestKey = poss[i];
            }
        }
        return bestKey;
    }

    ArrayList<CellUnitKey> getBestUnitKeys(CellInfo cell)
    {
        this.bestScore = Integer.MIN_VALUE;
        ArrayList<CellUnitKey> bestKey = new ArrayList<CellUnitKey>();
        CellUnitKey[] poss = cell.getKeys();
        for (int i = 0; i < poss.length; ++i) {
            if (cell.get(poss[i]).score >= this.bestScore) {
                if (cell.get(poss[i]).score > this.bestScore) {
                    this.bestScore = cell.get(poss[i]).score;
                    bestKey.clear();
                }
                bestKey.add(poss[i]);
            }
        }
        return bestKey;
    }

    private ArrayList<Coord> getBestCells()
    {
        this.bestScore = 0;
        ArrayList<Coord> ws = new ArrayList<Coord>();
        int N = this.alignmentMatrix.length, M = this.alignmentMatrix[0].length;
        for (int i = N - 1; i > 0; --i) {
            processCellScores(ws, i, M);
        }
        for (int i = M - 1; i > 0; --i) {
            processCellScores(ws, N, i);
        }
        return ws;
    }

    private void processCellScores(ArrayList<Coord> ws, int i, int j)
    {
        double crtScore;
        CellUnitKey[] keys = this.alignmentMatrix[i][j].getKeys();
        for (int k = 0; k < keys.length; ++k) {
            crtScore = this.alignmentMatrix[i][j].get(keys[k]).score;
            if (crtScore >= this.bestScore) {
                Coord where = new Coord(i, j, keys[k]);
                if (crtScore != this.bestScore) {
                    this.bestScore = crtScore;
                    ws.clear();
                }
                ws.add(where);
            }
        }
    }

    private ArrayList<Coord> getBestCells(double threshold)
    {
        this.bestScore = 0;
        HashMap<Coord, Double> bestScores = new HashMap<Coord, Double>();
        HashMap<Coord, Coord> bestPaths = new HashMap<Coord, Coord>();
        ArrayList<Coord> ws = new ArrayList<Coord>();

        int N = this.alignmentMatrix.length, M = this.alignmentMatrix[0].length;
        for (int i = N - 1; i > 0; --i) {
            processCellScores(bestScores, bestPaths, i, M, threshold);
        }
        for (int i = M - 1; i > 0; --i) {
            processCellScores(bestScores, bestPaths, N, i, threshold);
        }

        Iterator<Coord> it = bestPaths.keySet().iterator();
        while (it.hasNext()) {
            Coord from = it.next();
            ws.add(bestPaths.get(from));
        }
        return ws;
    }

    private void processCellScores(HashMap<Coord, Double> bestScores, HashMap<Coord, Coord> bestPaths, int i, int j,
        double threshold)
    {
        double crtScore;
        CellUnitKey[] keys = this.alignmentMatrix[i][j].getKeys();
        for (int k = 0; k < keys.length; ++k) {

            LOG.debug("Finding best " + i + " " + j + " " + k);

            crtScore = this.alignmentMatrix[i][j].get(keys[k]).score;
            Coord from = this.alignmentMatrix[i][j].get(keys[k]).startedAt;
            if (crtScore > this.bestScore) {
                this.bestScore = crtScore;
            }

            double val = threshold;
            Double crtIBest = bestScores.get(from);
            if (crtIBest != null) {
                val = crtIBest.doubleValue();
            }
            if (crtScore >= val) {
                Coord where = new Coord(i, j, keys[k]);
                bestScores.put(from, new Double(crtScore));
                bestPaths.put(from, where);
                LOG.debug("found; " + bestScores + " total");
            }
        }
    }

    @Override
    public String traceBack(CellUnitKey k)
    {
        // TODO Auto-generated method stub
        return null;
    }
}
