package sequoia.path.alignment.estimation;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;

import sequoia.path.alignment.Coord;
import sequoia.path.alignment.estimation.ScoreEvaluationParameters.ReadingFrameDifference;
import sequoia.path.alignment.spaceecon.AlignmentPoints;

public class AlignmentScoreSignificanceEstimation
{
	//--------------------------------------------------------------
	// Default parameters
	/** Default size of the sequences generated for estimation */
    public static final int DEFAULT_RAND_SEQ_LEN = 1800;

	/** Default number of sequences generated for estimation */
    public static final int DEFAULT_RAND_SEQ_NO = 4;

	/** By default, local alignments are only considered if their score is above this value */
    public static final double DEFAULT_MIN_SCORE = 160; // 1000;
    
    /** Where to put the computed parameters, by default */
    public static final String DEFAULT_PARAM_FILE = "score_sig_params";

    /** Where to put the computed parameters */
    protected String paramFile = DEFAULT_PARAM_FILE;

	//--------------------------------------------------------------
	// Estimation parameters
	/** Local alignments are only considered if their score is above this value */
    protected HashMap<ReadingFrameDifference, Double> minScore = new HashMap<ReadingFrameDifference, Double>();

	/** Number of sequences generated for estimation */
    protected int sequenceCount;

	/** Sizes of the sequences generated for estimation */
    protected int sequence1Length;
    protected int sequence2Length;
    
    /** Reading frame difference  |-> [encountered scores] */
    protected HashMap<ReadingFrameDifference, ArrayList<Double>> scores = new HashMap<ReadingFrameDifference, ArrayList<Double>>();
    
    /** The computed parameters */
    protected ScoreEvaluationParameters p = new ScoreEvaluationParameters();

    /** Where to send statistics */
    @Deprecated
    protected PrintStream globalStats;
    /** Should the estimation be performed? */
    @Deprecated
    protected boolean estimate = false;

	//--------------------------------------------------------------
	// Constructors
    public AlignmentScoreSignificanceEstimation()
    {
        sequenceCount = DEFAULT_RAND_SEQ_NO;
        sequence1Length = DEFAULT_RAND_SEQ_LEN;
        sequence2Length = DEFAULT_RAND_SEQ_LEN;
        setMinScore(DEFAULT_MIN_SCORE, ReadingFrameDifference.values());
        for (ReadingFrameDifference d : ReadingFrameDifference.values()) {
        	scores.put(d, new ArrayList<Double>());
        }
    }

    public AlignmentScoreSignificanceEstimation(int seqCount, int seq1Len, int seq2Len, double minScoreFW, double minScoreRC)
    {
        sequenceCount = seqCount;
        sequence1Length = seq1Len;
        sequence2Length = seq2Len;
        setMinScore(DEFAULT_MIN_SCORE, ReadingFrameDifference.valuesFW());
        setMinScore(DEFAULT_MIN_SCORE, ReadingFrameDifference.valuesRC());
        for (ReadingFrameDifference d : ReadingFrameDifference.values()) {
        	scores.put(d, new ArrayList<Double>());
        }
    }
	//--------------------------------------------------------------
	// Save
    /**
     * Save the computed parameters in the file indicated by the field paramFile
     */
    protected void saveParameters()
    {
        this.p.save(paramFile);
    }
	//--------------------------------------------------------------
	// Access to internal information
    /**
     * For the given reading frame differences, set the threshold above which the scores of local alignments are considered for the estimation
     * @param minScore
     * @param targetKeys
     * @return
     */
    protected void setMinScore(double minScore, ReadingFrameDifference[] targetKeys) {
        for (ReadingFrameDifference d : targetKeys) {
        	this.minScore.put(d, minScore);
        }
    }
    protected void setMinScore(double minScore) {
        setMinScore(minScore, ReadingFrameDifference.values());
    }
    public void setMinScoreFW(double minScore) {
        setMinScore(minScore, ReadingFrameDifference.valuesFW());
    }
    public void setMinScoreRC(double minScore) {
        setMinScore(minScore, ReadingFrameDifference.valuesRC());
    }
    public double getMinScoreFW() {
        return minScore(ReadingFrameDifference.FW);
    }
    public double getMinScoreRC() {
        return minScore(ReadingFrameDifference.RC);
    }
    public double getMinScore(boolean reverseComplementary) {
        return reverseComplementary ? minScore(ReadingFrameDifference.RC) : minScore(ReadingFrameDifference.FW);
    }
    public double minScore(ReadingFrameDifference d) {
        return this.minScore.get(d);
    }
    public int getSequenceCount()
    {
        return sequenceCount;
    }

    public void setSequenceCount(int sequenceCount)
    {
        this.sequenceCount = sequenceCount;
    }

    public int getSequence1Length()
    {
        return sequence1Length;
    }
    
    public int getSequence2Length()
    {
        return sequence2Length;
    }

    public void setSequence1Length(int sequenceLength)
    {
        this.sequence1Length = sequenceLength;
    }
    
    public void setSequence2Length(int sequenceLength)
    {
        this.sequence2Length = sequenceLength;
    }

    public String getParamFile() {
		return paramFile;
	}

	public void setParamFile(String paramFile) {
		if (paramFile != null && paramFile.trim().length() > 0) {
		   this.paramFile = paramFile;
		}
	}
    
    protected ArrayList<Double> scores(ReadingFrameDifference d) {
    	return this.scores.get(d);
    }
    public void addScore(double score, ReadingFrameDifference d) {
    	 this.scores(d).add(score);
    }
    
    /**
     * Extract the significant (i.e. with a score above the established threshold) local alignments from a large alignment matrix. 
     * @param a The alignment matrix from which the local alignments should be extracted and used for estimation.
     */
    public void extractLocalAlignments(ScoreEvaluationParametersAlignment a)
    {
    	// For statistics:
        double min = Double.MAX_VALUE, max = Double.MIN_VALUE, med = 0;
        int count = 0;
        
        ReadingFrameDifference d = a.getReadingFrameDifference();
        
        HashMap<Coord, AlignmentPoints> h = a.getIslands();
        
    	// Mostly for debug:
        System.out.println("\n----------------");
        System.out.println(  "      " + d);
    	System.out.println("Islands:    " + scores(d).size());
    	
    	for (Coord co : h.keySet()) {
        	double val = h.get(co).getScore();
        	if (val >= minScore(d)) {
              scores(d).add((h.get(co)).getScore());
              //globalStats.println(val);
        	}
        	if (val < min) {
            	  min = val;
              }
              if (val > max) {
            	  max = val;
              }
              med += val;
              ++count;
        }
    	// Mostly for debug:
        System.out.println("Min:        " + min);
        System.out.println("Max:        " + max);
        System.out.println("Med:        " + med/count);
        System.out.println("Count:      " + count);
    	System.out.println("Islands:    " + scores(d).size());
        System.out.println();
    }
    /**
     * Compute the actual estimation parameters for all setups
     */
    public void compute()
    {
    	for (ReadingFrameDifference d : ReadingFrameDifference.values()) {
    		compute(d);
    	}
    	saveParameters();
    }
    /**
     * Compute the estimation parameters for a given reading frame difference
     * @param d
     */
    protected void compute(ReadingFrameDifference d)
    {
        double sum = 0;
        int count = scores(d).size();
        for (double val : scores(d)) {
        	sum += (val - minScore(d));
        }

    	// Mostly for debug:
        System.out.println("Islands:    " + scores(d).size());
        System.out.println("Sum:        " + sum);
        System.out.println("Seq count:  " + this.sequenceCount);
        System.out.println("Seq length: " + this.sequence1Length + ", " + this.sequence2Length);

        sum /= scores(d).size();

        double lambda = Math.log(1 + 1.0 / sum);
        this.p.setLambda(lambda, d);
        this.p.setK(count * Math.exp(lambda * minScore(d)) / (this.sequenceCount * (this.sequenceCount - 1) / 2 * this.sequence1Length * this.sequence1Length), d);

        System.out.println(this.p.getLambda(d) + " " + this.p.getK(d));
    }

    public boolean setStatsOutputFile(String file)
    {
        try {
            globalStats = new PrintStream(file);
            return true;
        } catch (Exception ex) {
            System.err.println("Wrong stats file! " + file);
            ex.printStackTrace();
            globalStats = System.out;
        }
        return false;
    }
}
