package sequoia.path.alignment.estimation;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;

import sequoia.path.util.Codon;

/**
 * The set of
 *
 */
public class ScoreEvaluationParameters {
	/**
	 * Possible reading frame differences in an alignment between two back-translation graphs:
	 * forward sense:               +0, +1, +2, FW (any in forward sense);
	 * reverse complementary sense: -1, -2, -3, RC (any in reverse complementary sense);
	 */
	public enum ReadingFrameDifference {
		_0(0), _1(1), _2(2), RC_1(-1), RC_2(-2), RC_3(-3), FW(false), RC(true);

		private static final int ALL_READING_FRAMES = Integer.MAX_VALUE >> 1;
		
		int value;
		
		ReadingFrameDifference(boolean reverseComplementary) {
			value = (reverseComplementary ? -1 : 1) * ALL_READING_FRAMES;
		}
		
		ReadingFrameDifference(int v) {
			value = v;
		}
		
		boolean isRC() {
			return this.value < 0;
		}
		
		public String toString() {
			if (Math.abs(this.value) == ALL_READING_FRAMES) {
				return super.toString();
			}
			return String.format("%+2d", value);
		}
		
		//-------------------------------------------------------------------------------------
		// Static fields and methods for customized access  
		
		/** int |-> ReadingFrameDifference, where the key is the actual difference */
		private static HashMap<Integer, ReadingFrameDifference> map = new HashMap<Integer, ReadingFrameDifference>();
		
		/**
		 * Initialization of the map
		 */
		static {
			for (ReadingFrameDifference d : values()) {
				map.put(d.value, d);
			}
		}
		
		/**
		 * Obtain instance by integer value
		 * @param v 0, 1, 2, -1, -2, -3. If other values are provided, the MODULO 3 value is used (with -3 instead of 0 for negatives). 
		 * @return one of the instances _0 (for 0), _1(1), _2(2), RC_1(-1), RC_2(-2), RC_3(-3)
		 */
		public static ReadingFrameDifference valueOf(int v) {
			ReadingFrameDifference result = map.get(v);
			if (result == null) {
				if (v >= 0) {
					result = map.get(v % Codon.CODON_LEN);
				} else {
					result = map.get(-((Math.abs(v) - 1) % Codon.CODON_LEN + 1));
				}
			}
			return result;
		}
		
		/**
		 * Obtain instance by alignment sense
		 * @param reverseComplementary forward (false), reverse complementary (true)
		 * @return FW if the parameter is false, RC if it is true.
		 */
		public static ReadingFrameDifference valueOf(boolean reverseComplementary) {
			return map.get((reverseComplementary ? -1 : 1) * ALL_READING_FRAMES);
		}
		
		/**
		 * Obtain instance by name variants:
		 * @param name _0, _1, _2, RC_1, RC_2, RC_3, FW, RC, or 0, 1, 2, +0, +1, +2, -1, -2, -3.
		 * @param advanced
		 * @return _0 for -0 or +0, _1 for _1 or +1, etc.
		 */
		public static ReadingFrameDifference valueOf(String name, boolean advanced) {
			try {
				return valueOf(Integer.valueOf(name.substring(name.indexOf('+') + 1)));
			}  catch (NumberFormatException nfex) {
				try {
					return valueOf(name);
				} catch (IllegalArgumentException iaex){
					return FW;
				}
			}
		}
		
		/**
		 * Obtain all instances referring to alignments in the same sense 
		 * @return an array containing [_0, _1, _2, FW]
		 */
		public static ReadingFrameDifference[] valuesFW() {
			ReadingFrameDifference[] valuesFW = {_0, _1, _2, FW};
			return valuesFW;
		}
		/**
		 * Obtain all instances referring to alignments in the opposite sense 
		 * @return an array containing [RC_0, RC_1, RC_2, RC]
		 */
		public static ReadingFrameDifference[] valuesRC() {
			ReadingFrameDifference[] valuesRC = {RC_1, RC_2, RC_3, RC};
			return valuesRC;
		}
	}
	
	/**
	 * A (lambda, K) pair computed for a certain setup.
	 *
	 */
	protected class ScoreEvaluationParameter {
		double lambda;
		double K;
		
		ScoreEvaluationParameter() {
			lambda = 0;
			K = 1;
		}

		public double getLambda() {
			return lambda;
		}

		public void setLambda(double lambda) {
			this.lambda = lambda;
		}

		public double getK() {
			return K;
		}

		public void setK(double k) {
			K = k;
		}
		
		void set(double lambda, double K) {
			this.lambda = lambda;
			this.K = K;
		}
		/**
		 * Score evaluation
		 * @param score
		 * @param dim Typically, the product of the lengths of the two sequences
		 * @return E-value
		 */
		public double evaluate(double score, int dim) {
			return K * dim * Math.exp(-lambda * score);
			// P-value
//	    	double E = K * dim * Math.exp(-lambda * score);
//	        return 1 - Math.exp(-E);
		}
		/**
		 * Load parameter values from a string representation, of the form "lambda k"
		 * @param s
		 */
		public void load(String s) {
	       String[] tokens = s.trim().split(" ");
	       lambda = Double.valueOf(tokens[0]);
	       K      = Double.valueOf(tokens[1]);  
	    }

		/**
		 * String representation of the parameters: "lambda k"
		 */
	    public String toString() {
	    	return lambda + " " + K;
	    }
	}
	
	/** Score evaluation parameters for each of the possible reading frame differences in the aligned sequences */
	HashMap<ReadingFrameDifference, ScoreEvaluationParameter> p = new HashMap<ReadingFrameDifference, ScoreEvaluationParameter>();
	
	//-------------------------------------------------------------------
	// Constructors
	public ScoreEvaluationParameters() {
		for (ReadingFrameDifference f : ReadingFrameDifference.values()) {
			p.put(f, new ScoreEvaluationParameter());
		}
	}
	public ScoreEvaluationParameters(String fileName) {
		this();
		load(fileName);
	}
	//-------------------------------------------------------------------
	// Getters/Setters
	public void set(double lambda, double K, ReadingFrameDifference d) {
		 this.p.get(d).set(lambda, K);
	}
	
	public void setLambda(double lambda, ReadingFrameDifference d) {
		 this.p.get(d).setLambda(lambda);
	}
	
	public void setK(double K, ReadingFrameDifference d) {
		 this.p.get(d).setK(K);
	}
	
	public double getLambda(ReadingFrameDifference d) {
		 return this.p.get(d).getLambda();
	}
	
	public double getK(ReadingFrameDifference d) {
		 return this.p.get(d).getK();
	}
	//-------------------------------------------------------------------
	// Load/Save
	/**
	 * Load the parameters from a file. File format: @see #toString()
	 * @param fileName
	 */
	public void load(String fileName) {
		try {
			BufferedReader br = new BufferedReader(new FileReader(fileName));
	        // System.out.println(new File(fileName).getAbsolutePath());
	        String line;
	        String[] tokens;
	        while ((line = br.readLine()) != null && (tokens = line.split(":")).length == 2) {
	           this.p.get(ReadingFrameDifference.valueOf(tokens[0], true)).load(tokens[1]);
	        }
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	/**
	 * Save the parameters to a file. File format: @see #toString()
	 * @param fileName
	 */
	public void save(String fileName) {
		try {
			PrintWriter p = new PrintWriter(fileName);
			p.print(this);
			p.flush();
			p.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}
	//-------------------------------------------------------------------
	/**
	 * String representation of the parameters: 
	 * "FW: lambda_fw K_fw
	 *  RC: lambda_rc K_rc
	 *  +0: lambda__0 k__0
	 *  +1: lambda__1 k__1
	 *  +2: lambda__2 k__2
	 *  -1: lambda_-1 k_-1
	 *  -2: lambda_-2 k_-2
	 *  -3: lambda_-3 k_-3
	 *  "
	 */
	public String toString() {
		StringBuilder str = new StringBuilder();
		for (ReadingFrameDifference d : ReadingFrameDifference.values()) {
		    str.append(d);
		    str.append(": ");
		    str.append(this.p.get(d));
		    str.append("\n");
		}
		return str.toString();
	}
	//-------------------------------------------------------------------
	/**
	 * Score evaluation
	 * @param score
	 * @param dim Typically, the product of the lengths of the two sequences
	 * @param d The reading frame difference
	 * @return The score significance with respect to the given reading frame difference.
	 */
	public double evaluate(double score, int dim, ReadingFrameDifference d) {
		return this.p.get(d).evaluate(score, dim);
	}
}
