package sequoia.path.alignment.format;

import sequoia.path.alignment.Alignment;
import sequoia.path.alignment.AlignmentTracer;
import sequoia.path.alignment.FrameshiftFragment;

public enum PlainTextFormatter implements AlignmentFormatter {
	
	instance;

    private final static int CHARS_PER_ROW = 90;
    
    private final static String OUTPUT_INFO_MARKER = ">> "; 
    
    /**
     * Creates a plain text representation of the alignment, including information such as:
     * - identifiers of the aligned sequences:
	 * - number of frameshifts encountered in the alignment. 
     * - score, e-value, frameshifted fragments with their scores and evalues.
     */
	public String format(Alignment alignment) {
        StringBuilder str = new StringBuilder();
        str.append(alignmentHeader(alignment));
        str.append(toWrappedString(alignment));
        str.append(alignmentSummary(alignment));
        return str.toString();
    }
	/**
	 * Breaks the alignments in rows of 90 characters, for better readability
	 * @param alignment
	 * @return  A string representation of the alignment
	 */
	public String toWrappedString(Alignment alignment) {
        StringBuilder str = new StringBuilder();
        AlignmentTracer t = alignment.getTracer();
        String strA = t.getNucleotideSequenceA();
        String strB = t.getNucleotideSequenceB();
        String strpA = t.getAASequenceA();
        String strpB = t.getAASequenceB();
        String matcher = t.getMatchSequence();
        String strS = t.getScoreSignSequence();

        for (int i = 0; i < strA.length(); i += CHARS_PER_ROW) {
            if (alignment.A.isBackTranslation()) {
                str.append(substring(strpA, i, CHARS_PER_ROW));
                str.append('\n');
            }
            str.append(substring(strA, i, CHARS_PER_ROW));
            str.append('\n');
            str.append(substring(matcher, i, CHARS_PER_ROW));
            str.append('\n');
            str.append(substring(strS, i, CHARS_PER_ROW));
            str.append('\n');
            str.append(substring(strB, i, CHARS_PER_ROW));
            str.append('\n');
            if (alignment.B.isBackTranslation()) {
                str.append(substring(strpB, i, CHARS_PER_ROW));
                str.append('\n');
            }
            str.append('\n');
        }
        return str.toString();
    }
	/**
	 * Combines the sequences built by the alignment tracer in a single string.
	 * @param alignment
	 * @return A string representation of the alignment
	 */
    public String toUnwrappedString(Alignment alignment)
    {
        StringBuilder str = new StringBuilder();
        AlignmentTracer t = alignment.getTracer();
        String strA = t.getNucleotideSequenceA();
        String strB = t.getNucleotideSequenceB();
        String strpA = t.getAASequenceA();
        String strpB = t.getAASequenceB();
        String matcher = t.getMatchSequence();
        String strS = t.getScoreSignSequence();
        
        if (alignment.A.isBackTranslation()) {
            str.append(strpA);
            str.append('\n');
        }
        str.append(strA);
        str.append('\n');
        str.append(matcher);
        str.append('\n');
        str.append(strS);
        str.append('\n');
        str.append(strB);
        str.append('\n');
        if (alignment.B.isBackTranslation()) {
            str.append(strpB);
            str.append('\n');
        }
        return str.toString();
    }
	/**
	 * Generates a (plain text) header for an alignment:
	 * - identifiers of the aligned sequences, and intervals that are actually aligned;
	 * - number of frameshifts encountered in the alignment. 
	 * @param alignment
	 * @return A string representation of the alignment
	 */
    public String alignmentHeader(Alignment alignment)
    {
        StringBuilder meta = new StringBuilder();
        meta.append(alignment.A.getSeqID());
        meta.append('[');
        meta.append(alignment.getTracer().getStartA());
        meta.append(", ");
        meta.append(alignment.getTracer().getEndA());
        meta.append("] || ");
        meta.append(alignment.B.getSeqID());
        meta.append('[');
        meta.append(alignment.getTracer().getStartB());
        meta.append(", ");
        meta.append(alignment.getTracer().getEndB());
        meta.append("]\n");
        int fs = 0;
        for (FrameshiftFragment f : alignment.getFragments()) {
        	try {
	        	if (f.getPhaseDifference(alignment) != 0) {
	        		++fs;
	        	}
        	} catch (Exception ex) {
        		ex.printStackTrace();
        	}
        }
        if (fs > 0) {
        	meta.append(OUTPUT_INFO_MARKER + "Frameshifted subsequences: " + fs + "\n");
        } else {
        	meta.append(OUTPUT_INFO_MARKER + "No frameshifts\n");
        }
        meta.append("\n================================================================================\n\n");
        return meta.toString();
    }
	
    /**
     * Creates a (plain text) string representation of the alignment summary:
     * score, e-value, frameshifted fragments with their scores and evalues.
     * @param alignment
     * @return
     */
    protected String alignmentSummary(Alignment alignment) {
    	StringBuilder str = new StringBuilder();
    	str.append("\n================================================================================\n\n");
    	str.append(OUTPUT_INFO_MARKER);
        str.append("Score:   " + String.format("%.2f", alignment.getScore()) + "\n");
    	str.append(OUTPUT_INFO_MARKER);
        str.append("E-value: " + alignment.getScoreSignificance() + "\n");
        str.append('\n');
        if (alignment.getFragments().size() > 0) {
        	str.append("\n--------------------------------------------------------------------------------\n\n");
        	str.append(OUTPUT_INFO_MARKER + "Scores per frameshifted fragment:\n");
        	for (FrameshiftFragment f : alignment.getFragments()) {
            	str.append(OUTPUT_INFO_MARKER);
        		str.append(f);
        		str.append(" E-value: ");
        		str.append(alignment.getScoreSignificance(f));
        		str.append("   (Phase diff: ");
        		str.append(String.format("%+2d", f.getPhaseDifference(alignment)));
        		str.append(")\n");
        	}
        }
        return str.toString();
    }
    
    /**
     * For internal use only
     * @param str
     * @param from
     * @param len
     * @return
     */
    private String substring(String str, int from, int len){
    	return str.substring(from, Math.min(from+len, str.length()));
    }

}
