package sequoia.path.alignment.spaceecon;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import sequoia.path.alignment.Alignment;
import sequoia.path.alignment.AlignmentTracer;
import sequoia.path.alignment.CellUnitInfo;
import sequoia.path.alignment.CellUnitKey;
import sequoia.path.alignment.Coord;
import sequoia.path.alignment.FrameshiftFragment;
import sequoia.path.graph.BackTranslationGraph;
import sequoia.path.graph.Node;
import sequoia.path.util.Codon;

abstract public class SpaceEconomicAligner extends Alignment
{
    private static final Log LOG = LogFactory.getLog(SpaceEconomicAligner.class);

    // Sequence A labels the columns (so it is horizontal); it is traversed by index i (or c...)
    // Sequence B labels the rows (so it is vertical ); it is traversed by index j
    public static final int ALLOWED_SEQ_LEN = 128;

    protected ScoreList list;

    ScanWindow win;

    Alignment fullAlignment;

    int[] milestones; // col coordinates

    int crtMilestoneIdx = 0;

    // ************************************************************************
    // C O N S T R U C T O R S
    public SpaceEconomicAligner(BackTranslationGraph a, BackTranslationGraph b, double scoreAtLeast)
    {
        super(a, b);
        this.list = new ScoreList(scoreAtLeast);
        this.allowedFrameshifts = DEFAULT_ALLOWED_FRAMESHIFTS;
    }

    public SpaceEconomicAligner(BackTranslationGraph A, BackTranslationGraph B, int d, int e, int m, int x,
        int codonSkip, int codonSkipExtension, int fsNo, double acceptedScore)
    {
        super(A, B, d, e, m, x);
        this.codonSkip = codonSkip;
        this.codonSkipExtension = codonSkipExtension;
        this.allowedFrameshifts = fsNo;
        this.list = new ScoreList(acceptedScore);
    }

    // ************************************************************************
    /**
     * Initialization
     */
    @Override
    public void init(BackTranslationGraph a, BackTranslationGraph b)
    {
        this.A = a;
        this.B = b;
        int milestonesNo = 0;
        if (this.A.size() > ALLOWED_SEQ_LEN) {
            milestonesNo = (int) Math.floor((this.A.size() / (double) ALLOWED_SEQ_LEN));
        }
        this.milestones = new int[milestonesNo];
        for (int i = 0; i < this.milestones.length; ++i) {
            int col = (a.size() + 1) / (this.milestones.length + 1) * (i + 1);
            this.milestones[i] = col;
            // System.out.println("Milestone [" + i + "] = " + milestones[i]);
        }
        if (this.milestones.length > 0) {
            this.win = new ScanWindow(this, this.getWindowSize());
        }

    }

    // ************************************************************************
    /**
     * Alignment
     */
    @Override
    public double align()
    {
        if (this.milestones.length > 0) {
            int i;
            // compute the first column in the window -> done at init
            // compute the next c-2 columns in the window
            for (i = 1; i < this.win.size() - 1; ++i) {
                computeColumn(i);
            }
            // Recursively compute the last column & shift
            boolean hasMoreCols = true;
            i = this.win.size() - 1;
            while (hasMoreCols) {
                computeColumn(i);
                hasMoreCols = this.win.shiftWindow();
            }
        } else {
            this.list.putGlobalAlignmentFor(this);
            this.fullAlignment = this.getFullmatrixAligner();
            this.fullAlignment.align();
            this.bestScore = this.fullAlignment.getScore();
        }

        if (filter.accepts(this)) {
            rebuildAlignment();
        }
        return this.list.getScore();

    }

    // ************************************************************************
    /**
     * Reconstruction from fragments
     */
    protected void rebuildAlignment()
    {
        // tracer = new AlignmentTracer(this);
        int i;
        Alignment subal = null;
        if (this.milestones.length > 0) {
            for (int I = 0; I < this.list.size(); ++I) {
                this.tracer = new AlignmentTracer(this);
                AlignmentPoints p = this.list.get(I);
                this.tracer.setStart(p.getStart().getDi(), p.getStart().getDj(), false);
                this.tracer.setEnd(p.getEnd().getDi(), p.getEnd().getDj(), false);

                // System.out.println(p);
                for (i = 0; i < p.size() - 1; ++i) {
                    Coord from = p.getRight(i);
                    Coord to = p.getLeft(i + 1);
                    try {

                        BackTranslationGraph sA =
                            this.A.subgraph(from.getDi() - 1, to.getDi() - 1, from.getFullKey().nA);
                        BackTranslationGraph sB =
                            this.B.subgraph(from.getDj() - 1, to.getDj() - 1, from.getFullKey().nB);

                        // System.out.println("\n ------ \n" + from.getFullKey());
                        // System.out.println("PHASES: " + sA.getPhase() + " " + sB.getPhase());

                        subal = this.getSubmatrixAligner(sA, sB, subal);

                        // System.out.println(from + " -> " + to);

                        subal.align();

                        // System.out.println(subal.traceBack(to.getFullKey()));
                        subal.traceBack(to.getFullKey());

                        // subal.displayFragments(System.out);

                        this.tracer.append(subal.getTracer(), false);

                        if (i == 0) {
                            for (FrameshiftFragment f : subal.getFragments()) {
                                f.shift(from, 0);
                            }
                        }
                        this.mergeFrameshiftPoints(subal);

                        // System.out.println("Full alignment fragments: ");
                        // this.displayFragments(System.out);

                        // System.out.println("Frameshifts: " + subal.getCellUnit(sA.size(), sB.size(),
                        // to.getFullKey()).frameshiftCount + "  " + subal.getPerformedFrameshifts());
                        this.performedFrameshifts += subal.getPerformedFrameshifts();
                    } catch (NullPointerException npe) {
                        npe.printStackTrace();
                        break;
                    }
                }
            }
        } else {
            if (this.fullAlignment != null) {
                this.fullAlignment.traceBack();
                this.tracer = this.fullAlignment.getTracer();
            }
        }
    }

    /**
     * The frameshift fragments from the other alignment are appended to the current ones. The last point from the
     * current list is used to shift the new points. The junction point is eliminated.
     * 
     * @param other
     */
    public void mergeFrameshiftPoints(Alignment other)
    {
        List<FrameshiftFragment> otherFragments = other.getFragments();
        if (otherFragments.size() > 0) {
            if (this.fragments.size() > 0) {
                FrameshiftFragment crtLast = this.fragments.get(this.fragments.size() - 1);
                for (FrameshiftFragment f : otherFragments) {
                    f.shift(crtLast);
                }
                this.fragments.remove(this.fragments.size() - 1);
                otherFragments.get(0).setFrom(crtLast.getFrom());
                otherFragments.get(0).setScoreBefore(crtLast.getScoreBefore());
            }
            this.fragments.addAll(otherFragments);
        }
    }

    @Override
    public String traceBack()
    {
        if (this.tracer != null) {
            return this.tracer.toString();
        }
        return this.list.toString();
    }

    public int windowToMartixIdx(int i)
    {
        return this.win.windowToMatrixIdx(i);
    }

    /**
     * Compute the scores on a column
     * 
     * @param c the index of the column relative to the sliding window
     */
    protected void computeColumn(int c)
    {
        int C = this.win.windowToMatrixIdx(c);
        int j;
        for (j = 1; j < this.win.getColumn(c).length; ++j) {
            CellUnitKey[] cellKeys = this.win.getCell(c, j).getKeys();
            for (CellUnitKey k : cellKeys) {
                computeBestPartialScore(c, j, k);
                // update milestones
                ScanWindowCellUnitInfo crtCU = (ScanWindowCellUnitInfo) getCellUnit(c, j, k);
                Coord prevCoords = crtCU.getPrevCell();
                ScanWindowCellUnitInfo prevCU =
                    (ScanWindowCellUnitInfo) getCellUnit(prevCoords.getDi() + c, prevCoords.getDj() + j, prevCoords
                        .getFullKey());
                if (prevCU != null) {
                    crtCU.addMilestones(prevCU.milestones);
                }
                int lastMilestone =
                    (crtCU.milestones.size() > 0) ? (crtCU.milestones.get(crtCU.milestones.size() - 1).getRight()
                        .getDi()) : crtCU.getMyStart().getDi();
                for (int m = crtCU.milestones.size(); m < this.milestones.length; ++m) {
                    // int m = crtMilestoneIdx;
                    if (this.milestones[m] == C && // i'm in the milestone column
                        this.milestones[m] > lastMilestone + Codon.CODON_LEN && // not too close to the previous cut
                        // point
                        prevCoords.getDi() != 0 // not coming from above (if in the same column, i already added this
                    // milestone)
                    ) {
                        crtCU.addMilestone(C, j, k, prevCoords);
                    } else if (this.milestones[m] < C && // the milestone is behind
                        this.milestones[m] > C - prevCoords.getDi()// inside a gap
                    ) {
                        // there's a codon gap that jumped over the milestone
                        crtCU.addMilestone(C, j, k, prevCoords);
                    } else if (this.milestones[m] > C) { // the other milestones will be greater too; no use to continue
                        break;
                    }
                }
                LOG.debug("Aligning " + c + " " + j + " " + k);
                this.acknowledgeScore(c, j, k);
            }
        }
    }

    // ************************************************************************
    // Data access functions
    public ScoreList getScoreList()
    {
        final ScoreList l = this.list;
        return l;
    }

    @Override
    public CellUnitInfo getCellUnit(Coord co)
    {
        return this.win.getCellUnit(co);
    }

    @Override
    public CellUnitInfo getCellUnit(int i, int j, CellUnitKey k)
    {
        return this.win.getCellUnit(i, j, k);
    }

    @Override
    public CellUnitInfo getCellUnit(int i, int j, Node a, Node b)
    {
        return this.win.getCellUnit(i, j, a, b);
    }

    List<AlignmentMatrixMilestone> getMilestonesFor(Coord end)
    {
        // NB! this result depends on the current state of the alignment
        // for what alignment do we want the intermediary points?
        if (end == null) {
            return new ArrayList<AlignmentMatrixMilestone>();
        }
        if (this.win.getCellUnit(end, false) == null) {
            System.out.println("Does not exist: " + end);
            // return new ArrayList<Coord>();
        }
        return (this.win.getCellUnit(end, false)).getMilestones();
    }

    protected int getWindowSize()
    {
        return 2;
    }

    void setScoreFromScoreList()
    {
        this.bestScore = this.list.score;
    }

    void setScore(double score)
    {
        this.bestScore = score;
    }

    @Override
    public int getASize()
    {
        return this.A.size();
    }

    @Override
    public int getBSize()
    {
        return this.B.size();
    }

    // ************************************************************************
    // ABSTRACT FUNCTIONS, TO BE IMPLEMENTED BY GLOBAL AND LOCAL ALIGNERS
    public abstract double getCellUnitInitValue();

    public abstract double getBorderInitValue(int where);

    public abstract void initCellUnitStartField(ScanWindowCellUnitInfo c, int i, int j, CellUnitKey k);

    public abstract void acknowledgeScore(int i, int j, CellUnitKey k);

    public abstract Alignment getSubmatrixAligner(BackTranslationGraph sA, BackTranslationGraph sB, Alignment prev);

    public abstract Alignment getFullmatrixAligner();

    @Override
    public double computeBestScore()
    {
        this.bestScore = this.list.score;
        return this.bestScore;
    }
}
