package sequoia.path.data;

import java.util.ArrayList;

import sequoia.path.alignment.Alignment;
import sequoia.path.alignment.GlobalDPAfGapAlignment;
import sequoia.path.alignment.GlobalDPLinGapAlignment;
import sequoia.path.alignment.LocalDPAfGapAlignment;
import sequoia.path.alignment.LocalDPLinGapAlignment;
import sequoia.path.alignment.OverlapDPAfGapAlignment;
import sequoia.path.alignment.OverlapDPLinGapAlignment;
import sequoia.path.alignment.estimation.ScoreEvaluationParametersAlignment;
import sequoia.path.alignment.estimation.ScoreEvaluationParameters.ReadingFrameDifference;
import sequoia.path.alignment.spaceecon.SpaceEconomicGlobalDPFSCAligner;
import sequoia.path.alignment.spaceecon.SpaceEconomicLocalDPFSCAligner;
import sequoia.path.alignment.spaceecon.SpaceEconomicOverlapDPFSCAligner;
import sequoia.path.graph.BackTranslationGraph;
import sequoia.path.params.Parameters;
import sequoia.path.params.Parameters.AlignmentType;
import sequoia.path.params.Parameters.GapType;
import sequoia.path.scores.ScoreMatrix;
import sequoia.path.util.DBSequence;

public abstract class Data
{
    protected int seqNo;

    protected Parameters params;

    protected boolean r = false;

    private ArrayList<BackTranslationGraph> graphs = new ArrayList<BackTranslationGraph>();

    protected ArrayList<DBSequence> seqs = new ArrayList<DBSequence>();

    private ScoreMatrix scoreMatrix;

    protected int startI = 0, startJ = 1;

    protected int grI = 0, grJ = 1;
    
    protected int rfdI = 0;
    protected ReadingFrameDifference rfdsFW[] = ReadingFrameDifference.valuesFW();
    protected ReadingFrameDifference rfdsRC[] = ReadingFrameDifference.valuesRC();

    protected int endI = -1, endJ = -1;

    public void setStartIdx(int val)
    {
        setStartIdx(val, firstJFor(val));
    }

    void setStartIdx(int valI, int valJ)
    {
        startI = valI;
        startJ = valJ;
        reset();
    }

    void reset()
    {
        grI = startI;
        grJ = startJ;
    }

    protected abstract int firstJFor(int i);
    public abstract int alignmentsCount();

    public void setEndIdx(int val)
    {
        endI = val;
        endJ = -1;
    }

    void setEndIdx(int valI, int valJ)
    {
        endI = valI;
        endJ = valJ;
    }

    public abstract void doneLoading();

    Data(Parameters p)
    {
        params = p;
    }

    public void addSeq(String seq, String ID)
    {
        seqs.add(new DBSequence(seq, ID));
    }

    public void addSeq(String seq)
    {
        seqs.add(new DBSequence(seq, String.format("SEQ%4d", seqNo)));
        ++seqNo;
    }

    public void addAllSeqs(ArrayList<String> theSeqs)
    {
        for (int i = 0; i < theSeqs.size(); ++i) {
            addSeq(theSeqs.get(i));
        }
    }

    void addSeq(DBSequence seq)
    {
        seqs.add(seq);
    }

    public int seqsCount()
    {
        return seqs.size();
    }

    public DBSequence getSeq(int i)
    {
        return seqs.get(i);
    }

    public abstract boolean hasMoreAlignments();

    public Alignment nextAlignment()
    {
        int lgrJ = grJ, lgrI = grI;

        if (!params.isQuiet()) {
	        System.err.println(" Aligning:   " + lgrI + " (size " +  getGraphs().get(lgrI).size() + ")   to   " + lgrJ + " (size " + getGraphs().get(lgrJ).size() + ")");
	        System.err.println(seqs.get(lgrI));
	        System.err.println(seqs.get(lgrJ));
        }
        Alignment result = null;
        if (params.estimator == null) {
	        if ((r || !params.reverseCompl) && !params.reverseComplExclusive) {
	            result =
	                createAlignment(params.aType, params.gType, params.getMatch(), params.getMismatch(), params.getG(),
	                    params.getD(), params.getE(), params.getCodonSkip(), params.getCodonSkipExtend(), params.getAllowedFrameshifts(), params
	                        .getMinScore(), getGraphs().get(lgrI), getGraphs().get(lgrJ));
	            result.setProteinSeqs(seqs.get(lgrI), seqs.get(lgrJ));
	            result.setScoreMatrix(getScoreMatrix());
	            grJ++;
	            if (grJ == getGraphs().size()) {
	                grI++;
	                grJ = firstJFor(grI);
	            }
	        } else {
	            result =
	                createAlignment(params.aType, params.gType, params.getMatch(), params.getMismatch(), params.getG(),
	                    params.getD(), params.getE(), params.getCodonSkip(), params.getCodonSkipExtend(), params.getAllowedFrameshifts(), params
	                        .getMinScore(), getGraphs().get(lgrI), getGraphs().get(lgrJ).reverseComplementary());
	            result.setProteinSeqs(seqs.get(lgrI), seqs.get(lgrJ));
	            result.setScoreMatrix(getScoreMatrix());
	        }
	        if (params.reverseCompl) {
	            r = !r;
	        }
	        if (params.reverseComplExclusive) {
	        	grJ++;
	            if (grJ == getGraphs().size()) {
	                grI++;
	                grJ = firstJFor(grI);
	            }
	        }
        } else {

        	ReadingFrameDifference v[] = (r ? rfdsRC : rfdsFW);
        	ReadingFrameDifference df = v[rfdI];
        	 
         	result = new ScoreEvaluationParametersAlignment(getGraphs().get(lgrI), 
         			r ? getGraphs().get(lgrJ).reverseComplementary() : getGraphs().get(lgrJ), 
         			params.getD(), params.getE(), params.getMatch(), params.getMismatch(), 
         			params.getCodonSkip(), params.getCodonSkipExtend(), params.getAllowedFrameshifts(), 
        			params.estimator.minScore(df), df);
            result.setProteinSeqs(seqs.get(lgrI), seqs.get(lgrJ));
            result.setScoreMatrix(getScoreMatrix());
            
            
            // increment counters
        	r = r ^ (++rfdI >= v.length);
        	rfdI %= v.length;
        	if (!r && rfdI == 0) {
	        	grJ++;
	            if (grJ == getGraphs().size()) {
	                grI++;
	                grJ = firstJFor(grI);
	            }
        	}
        	
        	//System.out.println(r + " " + rfdI + " " + grI + " " + grJ);
        }
        return result;
    }

    Alignment createAlignment(AlignmentType aType, GapType gType, int match, int mismatch, int g, int d, int e, int codonSkip,
        int codonSxipExtension, int allowedFrameshifts, int minScore, BackTranslationGraph g1, BackTranslationGraph g2)
    {
        Alignment a;
  
        if (aType.equals(AlignmentType.LOCAL)) { // local
            if (gType.equals(GapType.FS_CODON)) { // frameshift/codon
                a =
                    new SpaceEconomicLocalDPFSCAligner(g1, g2, d, e, match, mismatch, codonSkip, codonSxipExtension, allowedFrameshifts,
                        minScore);
            } else if (gType.equals(GapType.AFFINE)) { // afine
                a = new LocalDPAfGapAlignment(g1, g2, d, e, match, mismatch);
            } else { // linear
                a = new LocalDPLinGapAlignment(g1, g2, g, match, mismatch);
            }
        } else if (aType.equals(AlignmentType.SEMI_GLOBAL)) { // overlap
            if (gType.equals(GapType.FS_CODON)) { // frameshift/codon
                a =
                    new SpaceEconomicOverlapDPFSCAligner(g1, g2, d, e, match, mismatch, codonSkip, codonSxipExtension, allowedFrameshifts,
                        minScore);
            } else if (gType.equals(GapType.AFFINE)) { // afine
                a = new OverlapDPAfGapAlignment(g1, g2, d, e, match, mismatch);
            } else { // linear
                a = new OverlapDPLinGapAlignment(g1, g2, g, match, mismatch);
            }
        } else { // global
            if (gType.equals(GapType.AFFINE)) {
                a = new GlobalDPAfGapAlignment(g1, g2, d, e, match, mismatch);
            } else if (gType.equals(GapType.FS_CODON)) { // frameshift/codon
            	a = new SpaceEconomicGlobalDPFSCAligner(g1, g2, d, e, match, mismatch, codonSkip, codonSxipExtension, allowedFrameshifts, minScore);
            } else { // linear
                a = new GlobalDPLinGapAlignment(g1, g2, g, match, mismatch);
            }
        }
        a.setEvaluator(this.params.evaluator);
        a.useCodonUsage(params.useCodonUsage);
        return a;
    }

    public void setScoreMatrix(ScoreMatrix scoreMatrix)
    {
        this.scoreMatrix = scoreMatrix;
    }

    public ScoreMatrix getScoreMatrix()
    {
        return scoreMatrix;
    }

    public void setGraphs(ArrayList<BackTranslationGraph> graphs)
    {
        this.graphs = graphs;
    }

    public ArrayList<BackTranslationGraph> getGraphs()
    {
        return graphs;
    }
}
