package sequoia.path.graph;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import sequoia.path.util.Codon;
import sequoia.path.util.DBSequence;
import sequoia.path.util.GeneticCodeMap;
import sequoia.path.util.GenomicSequenceUtils;

public class BackTranslationGraph
{
    /**
     * The graph nodes are grouped on positions', each group i containing nodes for the symbols (nucleotides) that can
     * appear on position i in the DNA sequences combined in this graph
     */
    public Node nodes[][];

    /** Is the graph obtained from back-translating a protein or is it a simple DNa sequence */
    boolean isBackTranslation = true;

    /** Is the graph a reverse complementary representation ? */
    boolean isReversedComplemented = false;

    /** Identifier of the represented sequence */
    String seqID;

    /** Quick access to the nodes on the first position */
    Node start[];

    /** Quick access to the nodes on the last position */
    Node end[];

    /** The graph's first position relative to the reading frame (possible values: 0, 1, 2) */
    private int phase = 0;

    /** Internal use only: the progress of graph construction by adding the codons for each amino acid */
    private int fillIndex = 0;

    /** Number of nodes in the graph TODO is this right? */
    private int nodeCount = -1;

    // ------------------------------------------------------------------
    /** Sequence Identifier */
    public String getSeqID()
    {
        return this.seqID;
    }

    // ------------------------------------------------------------------
    // Constructors
    /**
     * Constructor from a list of codons. The obtained graph has size 3.
     * 
     * @param codons
     */
    @Deprecated
    public BackTranslationGraph(ArrayList<Codon> codons)
    {
        this(codons, Codon.ANY_AA);
    }

    /**
     * Constructor from a list of codons, mentioning if the graph should be in compressed form. The obtained graph has
     * size 3.
     * 
     * @param codons
     * @param compressed
     */
    @Deprecated
    public BackTranslationGraph(ArrayList<Codon> codons, boolean compressed)
    {
        this(codons, Codon.ANY_AA, compressed);
    }

    /**
     * Constructor from a list of codons corresponding to a given amino-acid. The obtained graph has size 3.
     * 
     * @param codons
     * @param AA
     */
    @Deprecated
    public BackTranslationGraph(ArrayList<Codon> codons, char AA)
    {
        this(codons, AA, true);
    }

    /**
     * Constructor from a codon.
     * 
     * @param c
     */
    public BackTranslationGraph(Codon c)
    {
        this.nodes = new Node[Codon.CODON_LEN][];
        this.seqID = "Codon " + c.getSeq() + " of " + c.getAA();
        Node prev = Node.ANY_NODE;
        for (int i = 0; i < c.getSeq().length(); ++i) {
            this.nodes[i] = new Node[1];
            this.nodes[i][0] = new Node(c.getSymbol(i), prev, 1.0, c.getAA(), i);
            prev.next = this.nodes[i];
            prev = this.nodes[i][0];
        }
    }

    /**
     * This constructor builds the subgraph for the given codon set.
     */
    public BackTranslationGraph(ArrayList<Codon> codons, char AA, boolean compressed)
    {
        this.nodes = new Node[Codon.CODON_LEN][];
        this.seqID = "Back-translation of " + AA;
        if (codons == null) {
            System.err.println("No codons for the amino acid [" + AA + "]! Skipping...");
            return;
        }
        for (int i = 0; i < Codon.CODON_LEN - (compressed ? 1 : 0); ++i) {
            HashMap<String, Node> crtNodes = new HashMap<String, Node>();
            HashMap<Node, HashSet<Node>> crtLinks = new HashMap<Node, HashSet<Node>>();
            HashSet<Node> first = new HashSet<Node>();
            for (int j = 0; j < codons.size(); ++j) {
                Codon codon = codons.get(j);
                String n = getIdentifyingPrefix(codon, i);
                Node node = crtNodes.get(n);
                if (node == null) {
                    node =
                        new Node(n.charAt(2), getNode(this.fillIndex - 1, n.charAt(1), n.charAt(0)), codon
                            .getFrequency(), AA, i);
                    crtNodes.put(n, node);
                } else {
                    node.setWeight(node.getWeight() + codon.getFrequency());
                }
                // we are FOR SURE in the special case, when each node has only
                // one predecessor...
                if (node.pred[0].symbol != Node.ANY) {
                    markAsNext(crtLinks, node.pred[0], node);
                } else if (this.fillIndex != 0) {
                    for (int k = 0; k < this.end.length; ++k) {
                        markAsNext(crtLinks, this.end[k], node);
                    }
                }

                if (this.fillIndex == 0) {
                    first.add(node);
                }
            }
            HashSet<Node> newEnd = new HashSet<Node>();
            if (this.fillIndex != 0) {
                for (int k = 0; k < this.end.length; ++k) {
                    HashSet<Node> next = crtLinks.get(this.end[k]);
                    if (next != null) {
                        this.end[k].next = cast(next.toArray());
                        newEnd.addAll(next);
                    }
                }
                this.nodes[this.fillIndex] = cast(newEnd.toArray());
                this.end = this.nodes[this.fillIndex];
            } else {
                Object[] nextarr = first.toArray();
                this.nodes[this.fillIndex] = cast(nextarr);
                this.start = this.nodes[0];
                this.end = this.nodes[this.fillIndex];
            }
            this.fillIndex++;
        }
        if (compressed) {
            int i = Codon.CODON_LEN - 1;

            HashMap<String, HashSet<Character>> crtNodes = new HashMap<String, HashSet<Character>>();
            HashMap<String, Double> freqs = new HashMap<String, Double>();
            HashMap<Node, HashSet<Node>> crtLinks = new HashMap<Node, HashSet<Node>>();

            for (Codon codon : codons) {
                char lastSymb = codon.getSeq().charAt(i);
                String prefix = codon.getSeq().substring(0, i);
                HashSet<Character> set = crtNodes.get(prefix);
                if (set == null) {
                    set = new HashSet<Character>();
                    crtNodes.put(prefix, set);
                    freqs.put(prefix, 0.0);
                }
                set.add(lastSymb);
                crtNodes.put(prefix, set);
                freqs.put(prefix, NodeWeightEncDec.putSymbolWeight(freqs.get(prefix), codon.getFrequency(), lastSymb));
            }

            HashSet<Node> newEnd = new HashSet<Node>();

            for (String key : crtNodes.keySet()) {
                Node node = new Node(
                    		GenomicSequenceUtils.IUPACreverseMapping.get(crtNodes.get(key)), 
                    		getNode(this.fillIndex - 1, key.charAt(1), key.charAt(0)),
                    		freqs.get(key), AA, i);

                // we are FOR SURE in the special case, when each node has only
                // one
                // predecessor...
                // System.out.println(node.pred);
                if (node.pred[0].symbol != Node.ANY) {
                    markAsNext(crtLinks, node.pred[0], node);
                } else {
                    for (int k = 0; k < this.end.length; ++k) {
                        markAsNext(crtLinks, this.end[k], node);
                    }
                }
            }

            for (int k = 0; k < this.end.length; ++k) {
                HashSet<Node> next = crtLinks.get(this.end[k]);
                if (next != null) {
                    this.end[k].next = cast(next.toArray());
                    newEnd.addAll(next);
                }
            }
            this.nodes[this.fillIndex] = cast(newEnd.toArray());
            this.end = this.nodes[this.fillIndex];
            this.fillIndex++;
        }
    }

    /**
     * Build a graph with a given size, from an AA sequence; to be 'filled' using the append(codon graph) method
     * 
     * @param size
     * @param fromAAseq
     * @param id
     */
    public BackTranslationGraph(int size, boolean fromAAseq, String id)
    {
        if (fromAAseq) {
            this.nodes = new Node[Codon.CODON_LEN * size][];
        } else {
            this.nodes = new Node[size][];
        }
        this.seqID = id;
    }

    /**
     * Build a graph with a given size;
     * 
     * @param size
     * @param id
     */
    public BackTranslationGraph(int size, String id)
    {
        this(size, false, id);
    }

    /**
     * Build a graph from a DNA string; a trivial one-path graph is built
     * 
     * @param fromDNA
     */
    public BackTranslationGraph(String fromDNA)
    {
        this.isBackTranslation = false;
        this.nodes = new Node[fromDNA.length()][1];
        Node prev = Node.ANY_NODE;
        for (int i = 0; i < fromDNA.length(); ++i) {
            this.nodes[i][0] = new Node(fromDNA.charAt(i), prev);
            prev.next = this.nodes[i];
            prev = this.nodes[i][0];
        }
        this.start = this.nodes[0];
        this.end = this.nodes[this.nodes.length - 1];
    }

    /**
     * Build a graph from a DNA string; a trivial one-path graph is built
     * 
     * @param fromDNA
     * @param id
     */
    public BackTranslationGraph(String fromDNA, String id)
    {
        this(fromDNA);
        this.seqID = id;
    }

    /**
     * Build a graph from a DNA string; a trivial one-path graph is built. The sequence and its id are comprised in the
     * fromDNA object
     * 
     * @param fromDNA
     */
    public BackTranslationGraph(DBSequence fromDNA)
    {
        this(fromDNA.getSeq());
        this.seqID = fromDNA.getId();
    }

    // ------------------------------------------------------------------
    // Methods for generating the reverse complementary
    // ------------------------------------------------------------------
    /**
     * reverse index (internal use method)
     */
    private int ridx(int i)
    {
        return this.nodes.length - 1 - i;
    }

    /**
     * Build the reverse complementary
     * 
     * @return A new graph instance, in which the positions of the nodes and the arcs are reversed, and the nucleotides
     *         labeling the nodes are complemented
     */
    public BackTranslationGraph reverseComplementary()
    {
        return reverse(true);
    }

    /**
     * Reverse the graph
     * 
     * @return A new graph instance, in which the positions and the arcs are reversed.
     */
    public BackTranslationGraph reverse()
    {
        return reverse(false);
    }

    // --------------------------------------------------------------------------------
    /**
     * Build the reverse (complementary) of a graph
     * 
     * @param complementary
     * @return a new graph instance, in which the positions of the nodes and the arcs are reversed, and the nucleotides
     *         labeling the nodes are complemented of the provided parameter is true.
     */
    public BackTranslationGraph reverse(boolean complementary)
    {
        // HashMap<Node, Node> revNodeMap = new HashMap<Node, Node>(this.size());
        GraphTransformationCorrespondence revNodeMap = new GraphTransformationCorrespondence(this.size());
        // this hashmap maps node from the original graph to nodes from the
        // reversed one
        // useful for easily creating the pred/next links
        BackTranslationGraph g = new BackTranslationGraph(this.size(), this.seqID + "(reversed)");
        g.isReversedComplemented = true;
        g.phase = (this.size() - this.phase) % Codon.CODON_LEN;
        // create nodes
        for (int i = this.size() - 1; i >= 0; --i) {
            int l = this.nodes[i].length;
            int ri = ridx(i);
            g.nodes[ri] = new Node[l];

            for (int j = 0; j < l; ++j) {
                Node from = this.nodes[i][j];
                if (complementary) {
                    g.nodes[ri][j] = from.getComplementary();
                } else {
                    g.nodes[ri][j] = from.getReverse();
                }
                revNodeMap.put(ri, from, g.nodes[ri][j]);
            }
        }
        // create prev/next links
        for (int i = this.size() - 1; i >= 0; --i) {
            int l = this.nodes[i].length;
            int ri = ridx(i);
            for (int j = 0; j < l; ++j) {
                Node from = this.nodes[i][j];
                Node to = revNodeMap.get(ri, from);
                // build the prev set for the node in the reversed graph
                // from the next set of the node in the original graph
                if (from.next == null || i == 0
                    || (from.next.length != 1 && from.next.length == this.nodes[i + 1].length)) {
                    to.pred = new Node[1];
                    to.pred[0] = Node.ANY_NODE;
                } else {
                    to.pred = new Node[from.next.length];
                    for (int k = 0; k < to.pred.length; ++k) {
                        to.pred[k] = revNodeMap.get(ri - 1, from.next[k]);
                    }
                }
                // build the next set for the node in the reversed graph
                // from the prev set of the node in the original graph
                if (from.pred == null || i == 0) {
                    to.next = null; // end of the line...
                } else if (from.pred[0].symbol == Node.ANY) {
                    to.next = g.nodes[ri + 1];
                } else {
                    to.next = new Node[from.pred.length];
                    for (int k = 0; k < to.next.length; ++k) {
                        to.next[k] = revNodeMap.get(ri + 1, from.pred[k]);
                    }
                }
            }
        }

        g.start = g.nodes[0];
        g.end = g.nodes[this.size() - 1];
        return g;
    }

    // --------------------------------------------------------------------------------
    /**
     * Obtain a Subgraph with the columns [from..to-1]
     * 
     * @param from
     * @param to
     */
    public BackTranslationGraph subgraph(int from, int to, boolean fullCodons)
    {
        ++to;
        int nPhase = 0;
        if (from < 0) {
            from = 0;
        }
        if (fullCodons) {
            from = GenomicSequenceUtils.getPrevCodonStart(from, this.phase);
            to = GenomicSequenceUtils.getNextCodonStart(to, this.phase);
        } else {
            nPhase = from - GenomicSequenceUtils.getPrevCodonStart(from, this.phase);
            if (nPhase != 0) {
                nPhase = Codon.CODON_LEN - nPhase;
            }
        }
        BackTranslationGraph subg =
            new BackTranslationGraph(to - from, this.seqID + "(part: " + from + " -> " + to + ")");
        subg.phase = nPhase;
        subg.isBackTranslation = this.isBackTranslation;
        // this hashmap maps nodes from the original graph to nodes from the fragment one
        // useful for easily creating the pred/next links
        GraphTransformationCorrespondence nodeMap = new GraphTransformationCorrespondence(subg.size());
        for (int i = 0; i < subg.size(); ++i) {
            subg.nodes[i] = new Node[this.nodes[from + i].length];
            for (int j = 0; j < subg.nodes[i].length; ++j) {
                subg.nodes[i][j] = this.nodes[from + i][j].clone();
                nodeMap.put(i, this.nodes[from + i][j], subg.nodes[i][j]);
            }
        }

        int i = 0;
        Node any = Node.ANY_NODE;
        for (int j = 0; j < subg.nodes[i].length; ++j) {
            subg.nodes[i][j].pred = new Node[1];
            subg.nodes[i][j].pred[0] = any;
            if (subg.nodes[i][j].pred[0] == null) {
                System.out.println("1 NULL predecessor " + i + " " + j + " " + any);
                throw new NullPointerException();
            }
            int ns = this.nodes[from + i][j].next.length;
            subg.nodes[i][j].next = new Node[ns];
            for (int n = 0; n < ns; ++n) {
                subg.nodes[i][j].next[n] = nodeMap.get(i + 1, this.nodes[from + i][j].next[n]);
            }
        }
        for (i = 1; i < subg.size() - 1; ++i) {
            for (int j = 0; j < subg.nodes[i].length; ++j) {
                int ps = this.nodes[from + i][j].pred.length;
                subg.nodes[i][j].pred = new Node[ps];
                for (int p = 0; p < ps; ++p) {
                    subg.nodes[i][j].pred[p] = nodeMap.get(i - 1, this.nodes[from + i][j].pred[p]);
                    if (subg.nodes[i][j].pred[p] == null) {
                        System.out.println("2 NULL predecessor " + i + " " + j + " " + p + " "
                            + this.nodes[from + i][j].pred[p]);
                        throw new NullPointerException();
                    }
                }
                if (this.nodes[from + i][j].next == null) {
                    System.out.println(i + " " + (from + i) + " " + j);
                }
                int ns = this.nodes[from + i][j].next.length;
                subg.nodes[i][j].next = new Node[ns];
                for (int n = 0; n < ns; ++n) {
                    subg.nodes[i][j].next[n] = nodeMap.get(i + 1, this.nodes[from + i][j].next[n]);
                }
            }
        }

        i = subg.size() - 1;
        for (int j = 0; j < subg.nodes[i].length; ++j) {
            int ps = this.nodes[from + i][j].pred.length;
            subg.nodes[i][j].pred = new Node[ps];
            for (int p = 0; p < ps; ++p) {
                subg.nodes[i][j].pred[p] = nodeMap.get(i - 1, this.nodes[from + i][j].pred[p]);
                if (subg.nodes[i][j].pred[p] == null) {
                    System.out.println("3 NULL predecessor " + i + " " + j + " " + p + " "
                        + this.nodes[from + i][j].pred[p]);
                    throw new NullPointerException();
                }
            }

            subg.nodes[i][j].next = null;
        }
        subg.start = subg.nodes[0];
        subg.end = subg.nodes[subg.size() - 1];
        subg.isReversedComplemented = this.isReversedComplemented;

        return subg;
    }

    /**
     * Obtain a Subgraph with the columns [from..to-1], also fixing the first node of the graph (the other alternatives
     * on the first position are removed, along with the descendants for which they are unique predecessors)
     * 
     * @param from
     * @param to
     * @param startNode
     * @return a new graph instance containing the corresponding subset of nodes and preserving all the arcs.
     */
    public BackTranslationGraph subgraph(int from, int to, Node startNode)
    {
        BackTranslationGraph subg = subgraph(from, to, false);

        // int bbb = 0;
        // System.err.println("start node " + startNode);
        if (subg.nodes[0].length > 1) {
            /* remove from the first position nodes other than the start node, and then their successors */
            Node tmp[] = subg.nodes[0];

            ArrayList<Node> keptNodes = new ArrayList<Node>();

            for (Node n : tmp) {
                // System.err.println("start node " + (bbb++) + ": " + n);
                if (n.isolatedEquals(startNode)) {
                    // System.err.println("found");
                    subg.nodes[0] = new Node[1];
                    subg.nodes[0][0] = n; /* not putting startNode directly, in order to preserve graph links */
                    keptNodes.add(n);

                    boolean removed = true;

                    int i = 1, j;

                    ArrayList<Node> nodesToKeep = new ArrayList<Node>();

                    while (removed) {
                        // System.err.println("column " + i);
                        nodesToKeep.clear();
                        tmp = subg.nodes[i];
                        for (Node ni : tmp) {
                            // System.err.println("keep " + ni + "?");
                            for (Node nj : ni.pred) {
                                if (nj.symbol == Node.ANY || keptNodes.contains(nj)) {
                                    nodesToKeep.add(ni);
                                    // System.err.println("yes  " + ni);

                                    break;
                                }
                            }
                        }
                        if (removed = (nodesToKeep.size() < subg.nodes[i].length)) {
                            // System.err.println("must remove");
                            subg.nodes[i] = new Node[nodesToKeep.size()];
                            for (j = 0; j < nodesToKeep.size(); ++j) {
                                subg.nodes[i][j] = nodesToKeep.get(j);
                            }
                            keptNodes.clear();
                            keptNodes.addAll(nodesToKeep);
                        }
                        ++i;
                    }

                    subg.start = subg.nodes[0];
                    subg.end = subg.nodes[subg.size() - 1];

                    break;
                }
            }

            // TODO: remove me!
            if (subg.nodes[0].length != 1) {
                System.err.println("Wrong first node : " + startNode);
                System.exit(0);
            }
        }
        return subg;
    }

    /**
     * Obtain the codon "containing" the given position
     * 
     * @param pos
     * @return A string representing the sequence of the codon.
     */
    public String getCodonAt(int pos)
    {
        String s = "";
        if (!this.isBackTranslation) {
            for (int i = pos; i < Codon.CODON_LEN + pos; ++i) {
                s += this.nodes[i][0].symbol;
            }
        } else {
            String AA = "" + this.nodes[pos][0].getAA();
            ArrayList<Codon> cs = GeneticCodeMap.getInstance().getCodons(AA);
            double w = 0;
        	s = "***";
            for (int i = 0; i < cs.size(); ++i) {
                if (cs.get(i).getFrequency() > w) {
                    w = cs.get(i).getFrequency();
                    s = cs.get(i).getSeq();
                }
            }
        }
        return s;
    }

    /**
     * Obtain the position relative to the containing codon
     * 
     * @param idx
     * @return 0, 1 or 2.
     */
    public int getPositionInCodon(int idx)
    {
        return idx - GenomicSequenceUtils.getPrevCodonStart(idx, this.phase, true);
    }

    // --------------------------------------------------------------------------------
    // General graph info methods
    // --------------------------------------------------------------------------------
    /**
     * The size of the sequences combine in the graph, i.e. the number ofcolumns
     */
    public int size()
    {
        return this.nodes.length;
    }

    /**
     * Counts the nodes in the graph
     * 
     * @return the sum of node counts on each position of the graph
     */
    public int nodeCount()
    {
        if (this.nodeCount == -1) { // undefined
            this.nodeCount = 0;
            for (int i = 0; i < this.nodes.length; ++i) {
                this.nodeCount += this.nodes[i].length;
            }
        }
        return this.nodeCount;
    }

    /**
     * Decide whether this position is between codons
     * 
     * @param pos
     * @return true if the position marks the limit of a codon, false otherwise.
     */
    public boolean isCodonskipPosition(int pos)
    {
        if (pos - this.phase - 1 < Codon.CODON_LEN) {
            return false;
        }
        if (!this.isBackTranslation) {
            return true;
        }
        return (((pos - this.phase) % Codon.CODON_LEN) == 0);
        // return true;
    }

    // --------------------------------------------------------------------------------
    // Methods for filling in the graph with codons
    // --------------------------------------------------------------------------------
    // These methods help building a small subgraph from the set of codons that
    // encode
    // one AA
    // On each position of the codons, identical symbols with identical prefixes
    // are
    // merged into one node
    // These method return the prefix for a given codon on a given position
    // For internal use only
    private String getIdentifyingPrefix(String codon, int charPos)
    {
        StringBuilder is = new StringBuilder();
        for (int i = 0; i < (codon.length() - charPos - 1); ++i) {
            is.append(Node.ANY);
        }
        is.append(codon.substring(0, charPos + 1));

        return is.toString();
    }

    private String getIdentifyingPrefix(Codon codon, int charPos)
    {
        return getIdentifyingPrefix(codon.getSeq(), charPos);
    }

    // When building the subgraphs for a codon, the nodes that succeed a given
    // node in
    // the graph are memorized in a set
    private void markAsNext(HashMap<Node, HashSet<Node>> crtLinks, Node forNode, Node nextNode)
    {
        HashSet<Node> next = crtLinks.get(forNode);
        if (next == null) {
            next = new HashSet<Node>();
            crtLinks.put(forNode, next);
        }
        next.add(nextNode);
    }

    // This method shouldn't be here... it's not the graph's job!
    // TODO: move me!
    // transforms an Object array into an Node array
    private Node[] cast(Object[] from)
    {
        Node[] to = new Node[from.length];
        for (int i = 0; i < to.length; ++i) {
            to[i] = (Node) from[i];
        }
        return to;
    }

    /**
     * Append a graph constructed from a set of codons to a larger one corresponding to an AA sequence
     * 
     * @param graph
     * @return False if the graph cannot be appended because all the pre-established positions were filled; true
     *         otherwise.
     */
    public boolean append(BackTranslationGraph graph)
    {
        int appendSize = Math.min(graph.size(), this.size() - this.fillIndex);
        if (appendSize <= 0) {
            return false;
        }
        // Map from original nodes to cloned nodes
        HashMap<Node, Node> nodeMap = new HashMap<Node, Node>();
        // Pseudo-node needed for covering ANY predecessors.
        nodeMap.put(Node.ANY_NODE, Node.ANY_NODE);
        // Clone the protein graph nodes in the current back-translated graph.
        // c = column
        for (int c = 0; c < appendSize; ++c) {
            int thisC = c + this.fillIndex;
            this.nodes[thisC] = new Node[graph.getNodes(c).length];
            for (int n = 0; n < this.nodes[thisC].length; ++n) {
                this.nodes[thisC][n] = graph.getNodes(c)[n].clone();
                nodeMap.put(graph.getNodes(c)[n], this.nodes[thisC][n]);
            }
        }
        // Clone the links, too.
        for (int c = 0; c < appendSize; ++c) {
            int thisC = c + this.fillIndex;
            for (int n = 0; n < this.nodes[thisC].length; ++n) {
                Node gNode = graph.getNodes(c)[n];
                Node tNode = nodeMap.get(gNode);
                tNode.pred = new Node[gNode.pred.length];
                for (int p = 0; p < gNode.pred.length; ++p) {
                    tNode.pred[p] = nodeMap.get(gNode.pred[p]);
                }
                if (gNode.next != null) {
                    tNode.next = new Node[gNode.next.length];
                    for (int x = 0; x < gNode.next.length; ++x) {
                        tNode.next[x] = nodeMap.get(gNode.next[x]);
                    }
                }
            }
        }
        // Add back links between the new graph segment and the previous graph
        // end.
        for (int i = 0; i < this.nodes[this.fillIndex].length; ++i) {
            Node n = this.nodes[this.fillIndex][i];
            n.pred = new Node[1];
            n.pred[0] = Node.ANY_NODE;
        }
        // Add forward links between the previous graph end and the new graph
        // segment.
        if (this.fillIndex > 0) {
            for (int i = 0; i < this.nodes[this.fillIndex - 1].length; ++i) {
                Node n = this.nodes[this.fillIndex - 1][i];
                n.next = new Node[this.nodes[this.fillIndex].length];
                for (int j = 0; j < this.nodes[this.fillIndex].length; ++j) {
                    n.next[j] = this.nodes[this.fillIndex][j];
                }
            }
        }

        this.fillIndex += appendSize;
        return true;
    }

    // -----------------------------------------------------------------------------------
    // Some information about the graph
    // -----------------------------------------------------------------------------------
    public boolean isRC()
    {
        return this.isReversedComplemented;
    }

    public boolean isBackTranslation()
    {
        return this.isBackTranslation;
    }

    public int getPhase()
    {
        return this.phase;
    }

    public void setPhase(int phase)
    {
        this.phase = phase;
    }

    /**
     * The relative position inside a codon
     * 
     * @param i a position in the graph
     * @return the position in the corresponding codon, with respect to the known reading frame for this graph
     */
    public int getPhase(int i)
    {
        return (i - this.phase + Codon.CODON_LEN) % Codon.CODON_LEN;
    }

    // -----------------------------------------------------------------------------------
    // Methods for accessing graph data: positions (node groups), nodes, links,
    // node data...
    // -----------------------------------------------------------------------------------
    /**
     * The source nodes of the directed graph
     */
    Node[] getFirst()
    {
        return this.start;
    }

    /**
     * The final nodes of the directed graph
     */
    Node[] getLast()
    {
        return this.end;
    }

    // get the symbols from a node group
    private char[] get(Node[] theNodes)
    {
        char chars[] = new char[theNodes.length];
        for (int i = 0; i < chars.length; ++i) {
            chars[i] = theNodes[i].symbol;
        }
        return chars;
    }

    /**
     * Obtain the symbols that appear on the given position in the graph
     */
    public char[] get(int position)
    {
        return get(this.nodes[position]);
    }

    /**
     * Obtain the nodes from a given position in the graph
     */
    public Node[] getNodes(int position)
    {
        try {
            return (this.nodes[position]);
        } catch (Exception ex) {
            return null;
        }
    }

    /**
     * Get the node from the graph with the given coordinates.
     * 
     * @param position the position in the graph
     * @param symbol the labeling symbol
     * @param prev the predecessor symbol
     * @return the corresponding node, or {@code null} if no such node is found.
     */
    public Node getNode(int position, char symbol, char prev)
    {
        if (position == -1 || symbol == Node.ANY) {
            return Node.ANY_NODE;
        }
        for (int i = 0; i < this.nodes[position].length; ++i) {
            if (this.nodes[position][i].symbol == symbol) {
                if (this.nodes[position][i].hasPred(prev)) {
                    return this.nodes[position][i];
                }
            }
        }
        return null;
    }

    /**
     * Return the symbols labeling a node's successors
     * 
     * @param position the position of the queried node
     * @param symbol the symbol of the queried node
     * @return an array of {@code char} containing the symbols
     */
    public char[] getNextSymbols(int position, char symbol)
    {
        if (position == 0) {
            return get(this.nodes[0]);
        }
        for (int i = 0; i < this.nodes[position - 1].length; ++i) {
            if (this.nodes[position - 1][i].symbol == symbol) {
                get(this.nodes[position - 1][i].next);
            }
        }
        return new char[0];
    }

    /**
     * Return the symbols labeling a node's predecessors
     * 
     * @param position the position of the queried node
     * @param symbol the symbol of the queried node
     * @return an array of {@code char} containing the symbols
     */
    public char[] getPrecSymbols(int position, char symbol)
    {
        char prevs[] = {'-'};
        if (position >= this.nodes.length - 1) {
            return this.get(this.nodes.length - 1);
        }
        for (int i = 0; i < this.nodes[position].length; ++i) {
            if (this.nodes[position][i].symbol == symbol) {
                if (this.nodes[position][i].pred[0].symbol == Node.ANY) {
                    prevs = get(this.nodes[position - 1]);
                } else {
                    prevs = get(this.nodes[position][i].pred);
                }
            }
        }
        return prevs;
    }

    /**
     * Return the a node's predecessors
     * 
     * @param position the position of the queried node
     * @param symbol the symbol of the queried node
     * @return an array of {@code Node}
     */
    public Node[] getPrecNodes(int position, char symbol)
    {
        if (position >= this.nodes.length - 1) {
            return this.nodes[this.nodes.length - 1];
        }
        for (int i = 0; i < this.nodes[position].length; ++i) {
            if (this.nodes[position][i].symbol == symbol) {
                if (this.nodes[position][i].pred[0].symbol == Node.ANY) {
                    return (this.nodes[position - 1]);
                } else {
                    return this.nodes[position][i].pred;
                }
            }
        }
        return new Node[0];
    }

    /**
     * Return the a node's predecessors
     * 
     * @param position the position of the queried node
     * @param node the queried node
     * @return an array of {@code Node}
     */
    public Node[] getPrecNodes(int position, Node node)
    {
        if (position > this.nodes.length - 1) {
            return this.nodes[this.nodes.length - 1];
        }
        if (position == 0) {
            Node start[] = new Node[1];
            start[0] = Node.ANY_NODE;
            return start;
        }
        if (node.pred[0].symbol == Node.ANY) {
            return (this.nodes[position - 1]);
        } else {
            return node.pred;
        }
    }

    // ------------------------------------------------------------------------------
    // Methods for displaying info in the graph
    // TODO clean these methods up! they are inconsistent
    // -----------------------------------------------------------------------------
    // prints the columns
    public void display(PrintStream out)
    {
        for (int i = 0; i < this.nodes.length; ++i) {
            for (int j = 0; j < this.nodes[i].length; ++j) {
                out.print(this.nodes[i][j].symbol + "(" + i + ") -> ");
                if (this.nodes[i][j].next != null) {
                    for (int k = 0; k < this.nodes[i][j].next.length; ++k) {
                        out.print(this.nodes[i][j].next[k].symbol + "(" + (i + 1) + ") ");
                    }
                } else {
                    out.print("next is null!");
                }
                out.println();
            }
        }
    }

    // designed for debug: prints each node with its predecessors, successors,
    // position in the graph, weight...
    @Override
    public String toString()
    {
        StringBuilder s = new StringBuilder();
        // if (nodes == null) {
        // System.out.println ("nodes null");
        // System.exit(0);
        // } else {
        // System.out.println ("nodes: " + nodes.length);
        // }
        for (int i = 0; i < this.nodes.length; ++i) {
            if (this.nodes[i] == null) {
                // System.out.println ("nodes[" + i + "] null");
                // continue;
            }
            for (int j = 0; j < this.nodes[i].length; ++j) {
                s.append("<" + i + "," + j + ">: " + this.nodes[i][j].toString() + "\n");
            }
        }
        return s.toString();
    }

    /**
     * Re-translate a section of the back-translated graph
     * 
     * @param from the start position in the back-translation graph
     * @param to the start position (exclusive) in the back-translation graph
     * @return a string representing the amino acid sequence encoded by the demanded subgraph
     */
    // return
    public String getProteinStr(int from, int to)
    {
        StringBuilder strb = new StringBuilder();
        for (int i = from; i < to; i++) {
            if (((i - this.phase) + Codon.CODON_LEN) % Codon.CODON_LEN == 0) {
                strb.append('[');
            } else if (((i - this.phase) + Codon.CODON_LEN) % Codon.CODON_LEN == 1) {
                strb.append(this.nodes[i][0].AA);
            } else {
                strb.append(']');
            }

        }
        return strb.toString();
    }
}
