package sequoia.path.graph;

import sequoia.path.util.Codon;
import sequoia.path.util.GenomicSequenceUtils;
import sequoia.path.util.ProteinSeqUtils;

/**
 * A node in the {@link BackTranslationGraph}. Represents a nucleotide symbol, labeled with the position within the
 * codon, and the amino-acid encoded by that codon. Two identical nucleotides on the same position of two different
 * codons encoding the same amino-acid are represented by the same node if and only if they have identical prefixes in
 * their respective codons.
 */
public class Node
{
    /**
     * Symbol representing a "wildcard" node, used to improve memory efficiency. Nodes that should be linked with all
     * the nodes on a neighbor position have in fact a single link to this wildcard node.
     */
    public static final char ANY = '-';

    /**
     * A "wildcard" node, used to improve memory efficiency. Nodes that should be linked with all the nodes on a
     * neighbor position have in fact a single link to this wildcard node.
     */
    public static final Node ANY_NODE = new Node('-');

    /** Default value for the {@link #place} field. */
    public static final int UNDEFINED_POSITION = -1;

    /** Default value for the {@link #weight} field. */
    public static final double DEFAULT_WEIGHT = 1;

    /** The nucleotide represented by this node. */
    final char symbol;

    /** The position within the corresponding codon. */
    final int place;

    /**
     * The previous possible nodes in the next column, on valid paths. If ALL the nodes in the previous column precede
     * this node in valid paths, then just one node, with the ANY symbol will be in this array (for the alignment
     * algorithm to run faster) In most of the cases, pred contains just one specific node or the ANY node Exception:
     * the reverse complementary graph, when pred contains 2 or 4 nodes from a column of 6
     */
    Node pred[];

    /** The next possible nodes in the next column, on valid paths. */
    Node next[];

    /** Each node is associated a weight, computed from the codon frequency table. */
    private double weight;

    /** The amino acid encoded by the codon this node 'belongs' to. */
    char AA = Codon.ANY_AA;

    // -----------------------------------------------------------------------------
    // Constructors
    //
    public Node(char N)
    {
        this(N, null, DEFAULT_WEIGHT, Codon.ANY_AA, UNDEFINED_POSITION);
    }

    public Node(char N, double w)
    {
        this(N, null, w, Codon.ANY_AA, UNDEFINED_POSITION);
    }

    public Node(char N, char prev)
    {
        this(N, new Node(prev), DEFAULT_WEIGHT, Codon.ANY_AA, UNDEFINED_POSITION);
    }

    public Node(char N, char prev, double w)
    {
        this(N, new Node(prev), w, Codon.ANY_AA, UNDEFINED_POSITION);
    }

    public Node(char N, Node prev)
    {
        this(N, prev, DEFAULT_WEIGHT, Codon.ANY_AA, UNDEFINED_POSITION);
    }

    public Node(char N, Node prev, double w)
    {
        this(N, prev, w, Codon.ANY_AA, UNDEFINED_POSITION);
    }

    public Node(char N, Node prev, double w, char AA, int pos)
    {
        this.symbol = N;
        this.place = pos;
        if (prev != null) {
            this.pred = new Node[1];
            this.pred[0] = prev;
        }
        setWeight(w);
        this.AA = AA;
    }

    // ----------------------------------------------------------------------------
    /**
     * Compute the complementary symbol node. This method does NOT preserve links, since link preservation is never
     * needed in practice (new links must be built)
     * 
     * @return a new node, with the complementary symbol and reversed position
     */
    Node getComplementary()
    {
        return new Node(GenomicSequenceUtils.getComplementary(this.symbol), null, this.weight, ProteinSeqUtils
            .toRC(this.AA), Codon.CODON_LEN - 1 - this.place);
    }

    /**
     * Compute the reverse symbol node. This method does NOT preserve links, since link preservation is never needed in
     * practice (new links must be built)
     * 
     * @return a new node, with the same symbol and reversed position
     */
    Node getReverse()
    {
        return new Node(this.symbol, null, this.weight, ProteinSeqUtils.toRC(this.AA), Codon.CODON_LEN - 1 - this.place);
    }

    /**
     * Check if there is an arc between this node and a node on the previous position labeled with the symbol s.
     * 
     * @param s the label of the target predecessor
     * @return {@code true} if there is a predecessor labeled with 's', {@code false} otherwise
     */
    public boolean hasPred(char s)
    {
        if (this.pred == null) {
            return (s == ANY);
        }
        // unnecessary: if (pred[0].symbol == ANY) return true;
        for (int i = 0; i < this.pred.length; ++i) {
            if (this.pred[i].symbol == s) {
                return true;
            }
        }
        return false;
    }

    public Node[] getNext()
    {
        final Node[] n = this.next;
        return n;
    }

    // -----------------------------------------------------------------------------
    // hashcode, equals...
    //
    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode()
    {
        final int PRIME = 31;
        int result = 17;
        if (this.pred != null) {
            for (Node i : this.pred) {
                result = PRIME * result + i.hashCode();
            }
        } else {
            result = PRIME * result;
        }
        // result = PRIME * result + ((this.pred == null) ? 0 : this.pred.hashCode());
        result = PRIME * result + this.symbol;
        result = PRIME * result + this.place;
        // long temp;
        // temp = Double.doubleToLongBits(weight);
        // result = PRIME * result + (int) (temp ^ (temp >>> 32));
        return result;
    }

    /**
     * Checks the equality of two nodes without comparing their neighbors.
     * 
     * @param obj the other object to test for equality
     * @return {@code true} if the nodes have the same symbol and place, {@code false} otherwise
     */
    public boolean isolatedEquals(Object obj)
    {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Node other = (Node) obj;

        if (this.place != other.place) {
            return false;
        }
        if (this.symbol != other.symbol) {
            return false;
        }
        // if (Double.doubleToLongBits(weight) != Double.doubleToLongBits(other.weight))
        // return false;
        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object obj)
    {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Node other = (Node) obj;
        if (this.pred == null) {
            if (other.pred != null) {
                return false;
            }
        } else if (!isPredEqual(other)) {
            return false;
        }
        if (this.place != other.place) {
            return false;
        }
        if (this.symbol != other.symbol) {
            return false;
        }
        // if (Double.doubleToLongBits(weight) != Double.doubleToLongBits(other.weight))
        // return false;
        return true;
    }

    private boolean isPredEqual(Node other)
    {
        // TODO is it necessary to check?
        if (other.pred == null) {
            return this.pred == null;
        }
        if (this.pred.length != other.pred.length) {
            return false;
        }
        for (int i = 0; i < this.pred.length; ++i) {
            if (!this.pred[i].equals(other.pred[i])) {
                return false;
            }
        }

        return true;
    }

    /**
     * {@inheritDoc}
     * <p>
     * The arcs are not preserved in the clone.
     * </p>
     */
    @Override
    public Node clone()
    {
        return new Node(this.symbol, null, this.weight, this.AA, this.place);
    }

    // -----------------------------------------------------------------------------
    // String display
    //
    /**
     * {@inheritDoc}
     */
    @Override
    public String toString()
    {
        return predString() + "<-[[" + this.symbol + "@" + this.AA + "<" + this.place + "> ("
            + getTruncatedWeightString() + ")]]->" + nextString();
    }

    /**
     * A simpler representation of the node.
     * 
     * @return A string displaying the symbol and the weight.
     */
    public String toShortString()
    {
        return this.symbol + "(" + getTruncatedWeightString() + ")";
    }

    // ------------------------------------------------------------------------------
    // Various methods that generate representations, for internal use
    //
    private String predString()
    {
        return linkToString(this.pred);
    }

    private String nextString()
    {
        return linkToString(this.next);
    }

    private String getTruncatedWeightString()
    {
        return getTruncatedWeightString(2);
    }

    private String getTruncatedWeightString(int precision)
    {
        return String.format("%." + precision + "f", this.weight);
    }

    private String linkToString(Node[] links)
    {
        if (links == null) {
            return "[ ]";
        }
        StringBuilder str = new StringBuilder(2 * (links.length + 1) + 1);
        str.append("[ ");
        for (int k = 0; k < links.length; ++k) {
            try {
                str.append(links[k].symbol + "(" + links[k].getTruncatedWeightString() + ") ");
            } catch (NullPointerException e) {
                System.err.print("index in column: " + k);
                // throw e;
            }
        }
        str.append("]");
        return str.toString();
    }

    // -----------------------------------------------------------------------------
    // Field access
    //
    void setWeight(double w)
    {
        this.weight = w;
        if (this.weight > 0.9999999) {
            this.weight = 1;
        }
    }

    public double getWeight()
    {
        return this.weight;
    }

    public char getSymbol()
    {
        return this.symbol;
    }

    public char getAA()
    {
        return this.AA;
    }

    public int getPlace()
    {
        return this.place;
    }

    public Node[] getPred()
    {
        final Node[] p = this.pred;
        return p;
    }
}
