package sequoia.path.params;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.PrintStream;
import java.sql.Date;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.OptionGroup;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.PosixParser;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.math.NumberUtils;

import sequoia.path.alignment.Alignment;
import sequoia.path.alignment.estimation.AlignmentScoreSignificanceEstimation;
import sequoia.path.alignment.filter.EValueAlignmentFilter;
import sequoia.path.alignment.filter.HasFrameshiftsFilter;
import sequoia.path.data.DoubleDBData;
import sequoia.path.data.SingleDBData;
import sequoia.path.params.Parameters.AlignmentType;
import sequoia.path.params.Parameters.DBFormatType;
import sequoia.path.params.Parameters.GapType;
import sequoia.path.params.Parameters.SequenceType;
import sequoia.path.scores.CompressedContextScoreMatrix;
import sequoia.path.scores.ContextScoreMatrix;
import sequoia.path.scores.ScoreMatrix;
import sequoia.path.scores.SimpleScoreMatrix;
import sequoia.path.translation.BackTranslator;
import sequoia.path.util.RandSeqGenerator;

public class ParamProcessor
{
    static final long serialVersionUID = 110108;

    final String fileNameSufix;

    private Option helpOption;

    private Option quietOption;

    private Option verboseOption;
    
    private Option progressBarOption;

    private Option sequenceTypeOption;

    private Option firstSequenceOption;

    private Option firstSequenceFileOption;

    private Option secondSequenceOption;

    private Option secondSequenceFileOption;

    private Option databaseFileOption;

    private Option databaseFormatOption;

    private Option databaseSamplingOption;

    private Option databaseFilterStringOption;

    private Option databaseIdxStartOption;

    private Option databaseSeqIDStartOpion;

    private Option databaseIdxEndOption;

    private Option outputOption;

    private Option selectionOutputOption;

    private Option xmlOutputOption;

    private Option selectionEValThresholdOption;

    private Option enableMandatoryFSOption;

    private Option alignmentTypeOption;

    private Option gappingTypeOption;

    private Option scoreMatrixOption;

    private Option enableCodonUsageOption;

    private Option matchScoreOption;

    private Option mismatchScoreOption;

    private Option transitionMutationScoreOption;

    private Option linearGapPenaltyOption;

    private Option affineGapOpeningPenaltyOption;

    private Option affineGapExtensionPenaltyOption;

    private Option codonGapPenaltyOption;

    private Option codonGapExtensionPenaltyOption;

    private Option allowedFrameshiftCount;

    private Option minLocalScoreOption;

    private Option reverseComplOption;

    private Option reverseComplOnlyOption;

    private Option compressBackTranslationOption;

    private Option expandGenomicSequencesOption;

    private Option estimationMinScoreOption;

    private Option estimationMinScoreRCOption;

    private Option estimationSeq1LengthOption;

    private Option estimationSeq2LengthOption;

    private Option estimationSeqCountOption;

    private Option estimationParamsFileOption;

    private Option doEstimateOption;

    Options validArguments;

    CommandLineParser parser;

    CommandLine arguments;

    public ParamProcessor()
    {
        this.fileNameSufix = "-" + new Date(System.currentTimeMillis()).toString() + ".out";
        initializeValidArguments();
        this.parser = new PosixParser();

    }

    @SuppressWarnings("static-access")
    private void initializeOptions()
    {
        this.helpOption = new Option("h", "help", false, "Print this help message and exit");
        this.quietOption = new Option("q", "quiet", false, "Do not print any messages. -q has priority over -v.");
        this.verboseOption = new Option("v", "verbose", false, "Print messages. -q has priority.");
        this.progressBarOption = new Option("P", "progress", false, "Print progress bar.");

        this.sequenceTypeOption =
            OptionBuilder
                .withDescription("Sequence type; possible values: dna, rna, prot, auto (autodetect, default).")
                .hasArg().withArgName(SequenceType.generateOptionList()).create("st");
        this.firstSequenceOption =
            OptionBuilder.withDescription("First sequence").hasArg().withArgName("seq").create("s1");
        this.firstSequenceFileOption =
            OptionBuilder.withDescription("First sequence (from file)").hasArg().withArgName("file").create("S1");
        this.secondSequenceOption =
            OptionBuilder.withDescription("Second sequence").hasArg().withArgName("seq").create("s2");
        this.secondSequenceFileOption =
            OptionBuilder.withDescription("Second sequence (from file)").hasArg().withArgName("file").create("S2");
        this.databaseFileOption =
            OptionBuilder.withLongOpt("database").withDescription("Path to the database").hasArg().withArgName("file")
                .create("b");
        this.databaseFormatOption =
            OptionBuilder
                .withLongOpt("database-format")
                .withDescription(
                    "Database format; possible values: s (=simple data file), f (=fasta file), fd (=directory of fasta files), r (=randomly generated sequences); default: "
                        + DBFormatType.DEFAULT + ". In case the value is 'r', the '-b' param is used for statistics.")
                .hasArg().withArgName(DBFormatType.generateOptionList()).create("f");
        this.databaseSamplingOption =
            OptionBuilder.withLongOpt("sampling").withDescription(
                "Percetage of sequences to use from the database (default: " + Parameters.DEFAULT_DB_RATE + ")")
                .hasArg().withArgName("percentage").create("r");
        this.databaseFilterStringOption =
            OptionBuilder.withLongOpt("database-filter-string").withDescription(
                "Select those entries in the database whose identifier matches this string").hasArg().withArgName(
                "string").create("filter");
        this.databaseIdxStartOption =
            OptionBuilder.withLongOpt("database-index-start").withDescription(
                "Take into account sequences from this index forward for alignments.").hasArg().withArgName("number")
                .create("iStart");
        this.databaseIdxEndOption =
            OptionBuilder.withLongOpt("database-index-end").withDescription(
                "Take into account sequences up to this index for alignments.").hasArg().withArgName("number").create(
                "iEnd");
        this.databaseSeqIDStartOpion =
            OptionBuilder.withLongOpt("database-seqid-start").withDescription(
                "Take into account sequences from this identifier forward for alignments.").hasArg().withArgName(
                "sequence ID").create("sIDStart");
        this.outputOption =
            OptionBuilder.withLongOpt("output").withDescription("Output to a file; default: output to stdout").hasArg()
                .withArgName("file").create("o");

        this.xmlOutputOption =
            OptionBuilder.withLongOpt("xml-output").withDescription("Output to a file in xml format").hasArg()
                .withArgName("file").create("xml");

        this.selectionOutputOption =
            OptionBuilder.withLongOpt("selection-output").withDescription(
                "Output best alignments (i;e; with lowest e-value) to a file; default: output to stdout").hasArg()
                .withArgName("file").create("os");
        this.selectionEValThresholdOption =
            OptionBuilder.withLongOpt("max-evalue").withDescription(
                "Maximum accepted e-value for alignments to be considered significant (default: "
                    + EValueAlignmentFilter.DEFAULT_EVALUE_THRESHOLD + ")").hasArg().withArgName("number")
                .create("mev");
        this.enableMandatoryFSOption =
            new Option("FS", "frameshifts-only", false, "Only output alignments that contain at least a frameshift");

        this.alignmentTypeOption =
            OptionBuilder.withLongOpt("alignment-type").withDescription(
                "Alignment type; values: g (=global), l (=local), o (=overlap); default: "
                    + Parameters.AlignmentType.DEFAULT.label).hasArg().withArgName(AlignmentType.generateOptionList())
                .create("a");
        this.gappingTypeOption =
            OptionBuilder.withLongOpt("gapping-type").withDescription(
                "Gapping type; values: l (=linear), a (=affine), f(=frameshift/codon); default: "
                    + Parameters.GapType.DEFAULT.label).hasArg().withArgName(GapType.generateOptionList()).create("G");
        this.scoreMatrixOption =
            OptionBuilder.withLongOpt("score-matrix").withDescription(
                "nucleotide pair score matrix to use; if provided, params m, x, tx are ignored").hasArg().withArgName(
                "file").create("sm");
        this.matchScoreOption =
            OptionBuilder.withLongOpt("match").withDescription(
                "Match score (default: " + Alignment.DEFAULT_MATCH_SCORE + ")").hasArg().withArgName("number").create(
                "m");
        this.mismatchScoreOption =
            OptionBuilder.withLongOpt("mismatch").withDescription(
                "Mismatch score (default: " + Alignment.DEFAULT_MISMATCH_PENALTY + ")").hasArg().withArgName("number")
                .create("x");
        this.transitionMutationScoreOption =
            OptionBuilder.withLongOpt("transition").withDescription(
                "Transition mutation mismatch score (default: the mismatch score)").hasArg().withArgName("number")
                .create("t");
        this.linearGapPenaltyOption =
            OptionBuilder.withLongOpt("lgap").withDescription(
                "Linear gap penalty (default: " + Alignment.DEFAULT_LIN_GAP_PENALTY + ")").hasArg().withArgName(
                "number").create("g");
        this.affineGapOpeningPenaltyOption =
            OptionBuilder.withLongOpt("gap-opening").withDescription(
                "Affine gap opening penalty (default: " + Alignment.DEFAULT_AFINE_GAP_OPEN_PENALTY + ")").hasArg()
                .withArgName("number").create("d");
        this.affineGapExtensionPenaltyOption =
            OptionBuilder.withLongOpt("gap-extension").withDescription(
                "Affine gap extension penalty (default: " + Alignment.DEFAULT_AFINE_GAP_EXTEND_PENALTY + ")").hasArg()
                .withArgName("number").create("e");
        this.codonGapPenaltyOption =
            OptionBuilder.withLongOpt("codon-gap-penalty").withDescription(
                "Penalty for full codon gaps (default: " + Alignment.DEFAULT_CODONSKIP_PENALTY + ")").hasArg()
                .withArgName("number").create("c");
        this.codonGapExtensionPenaltyOption =
            OptionBuilder.withLongOpt("codon-gap-extension-penalty").withDescription(
                "Penalty for extending full codon gaps (default: " + Alignment.DEFAULT_CODONSKIP_PENALTY + ")")
                .hasArg().withArgName("number").create("ce");
        this.allowedFrameshiftCount =
            OptionBuilder.withLongOpt("allowed-frameshifts").withDescription(
                "Number of allowed frameshifts (default: "
                    + ((Alignment.DEFAULT_ALLOWED_FRAMESHIFTS == Integer.MAX_VALUE) ? "unbounded"
                        : Alignment.DEFAULT_ALLOWED_FRAMESHIFTS) + ")").hasArg().withArgName("number").create("fs");
        this.minLocalScoreOption =
            OptionBuilder.withLongOpt("min-local-score").withDescription(
                "Minimum acceptable score for local alignments (default: " + Alignment.DEFAULT_MIN_SCORE + ")")
                .hasArg().withArgName("number").create("msc");
        this.reverseComplOption =
            new Option("rc", "reverse-complementary", false, "Also align with reverse complementary");
        this.reverseComplOnlyOption =
            new Option("rcOnly", "reverse-complementary-only", false, "Only align with reverse complementary");
        this.compressBackTranslationOption =
            new Option("z", "compression", false,
                "Compression of Back-Translation graphs using IUPAC nucleotide ambiguity codes (default: "
                    + BackTranslator.DEFAULT_GRAPH_FORMAT + "; becomes " + (!(BackTranslator.DEFAULT_GRAPH_FORMAT))
                    + " if this option is present).");
        this.expandGenomicSequencesOption =
            new Option("gx", "expand", false,
                "Expand genomic sequences into complete back-translated graphs of the encoded protein");

        this.enableCodonUsageOption =
            new Option("cu", "include-codon-usage-in-score", false,
                "Consider the codon usage probabilities when computing the score");
        this.doEstimateOption =
            new Option(
                "estim",
                "compute-statistical-significance",
                false,
                "Compute statistical significance parameters. Currently available only for frameshift/codon local alignments, with randomly generated sequences.");

        this.estimationMinScoreOption =
            OptionBuilder
                .withLongOpt("min-estimation-score")
                .withDescription(
                    "Minimum score of local alignments considered for estimating the score significance parameters on sequences aligned on the same translation sense.")
                .hasArg().withArgName("number").create("eC");
        this.estimationMinScoreRCOption =
            OptionBuilder
                .withLongOpt("min-estimation-score")
                .withDescription(
                    "Minimum score of local alignments considered for estimating the score significance parameters on reverse complementary alignments. If this"
                        + "parameter is missing, the value for the same translation sense is used.").hasArg()
                .withArgName("number").create("eCrc");
        this.estimationParamsFileOption =
            OptionBuilder.withLongOpt("estim-params-file").withDescription(
                "Name of the file that contains the parameters used to determine score significance").hasArg()
                .withArgName("file").create("eP");
        this.estimationSeq1LengthOption =
            OptionBuilder.withLongOpt("estimation-sequence-1-length").withDescription(
                "Length of the first sequence from each pair generated for score significance parameters estimation")
                .hasArg().withArgName("number").create("eL1");
        this.estimationSeq2LengthOption =
            OptionBuilder.withLongOpt("estimation-sequence-2-length").withDescription(
                "Length of the second sequence from each pair generated for score significance parameters estimation")
                .hasArg().withArgName("number").create("eL2");
        this.estimationSeqCountOption =
            OptionBuilder.withLongOpt("estimation-sequences-number").withDescription(
                "Number of sequences generated for score significance parameters estimation").hasArg().withArgName(
                "number").create("eN");
    }

    private void initializeValidArguments()
    {
        initializeOptions();
        this.validArguments = new Options();
        this.validArguments.addOption(this.helpOption);
        this.validArguments.addOption(this.quietOption);
        this.validArguments.addOption(this.verboseOption);
        this.validArguments.addOption(this.progressBarOption);

        OptionGroup firstSequenceParams = new OptionGroup();
        firstSequenceParams.addOption(this.firstSequenceOption);
        firstSequenceParams.addOption(this.firstSequenceFileOption);
        this.validArguments.addOptionGroup(firstSequenceParams);

        OptionGroup secondSequenceParams = new OptionGroup();
        secondSequenceParams.addOption(this.secondSequenceOption);
        secondSequenceParams.addOption(this.secondSequenceFileOption);
        this.validArguments.addOptionGroup(secondSequenceParams);

        this.validArguments.addOption(this.sequenceTypeOption);

        this.validArguments.addOption(this.databaseFileOption);
        this.validArguments.addOption(this.databaseFormatOption);
        this.validArguments.addOption(this.databaseSamplingOption);
        this.validArguments.addOption(this.databaseFilterStringOption);
        this.validArguments.addOption(this.databaseIdxStartOption);
        this.validArguments.addOption(this.databaseIdxEndOption);
        this.validArguments.addOption(this.databaseSeqIDStartOpion);

        this.validArguments.addOption(this.outputOption);
        this.validArguments.addOption(this.xmlOutputOption);
        this.validArguments.addOption(this.selectionOutputOption);
        this.validArguments.addOption(this.selectionEValThresholdOption);
        this.validArguments.addOption(this.enableMandatoryFSOption);

        this.validArguments.addOption(this.alignmentTypeOption);
        this.validArguments.addOption(this.gappingTypeOption);
        this.validArguments.addOption(this.scoreMatrixOption);
        this.validArguments.addOption(this.matchScoreOption);
        this.validArguments.addOption(this.mismatchScoreOption);
        this.validArguments.addOption(this.transitionMutationScoreOption);
        this.validArguments.addOption(this.linearGapPenaltyOption);
        this.validArguments.addOption(this.affineGapOpeningPenaltyOption);
        this.validArguments.addOption(this.affineGapExtensionPenaltyOption);
        this.validArguments.addOption(this.codonGapPenaltyOption);
        this.validArguments.addOption(this.codonGapExtensionPenaltyOption);
        this.validArguments.addOption(this.allowedFrameshiftCount);
        this.validArguments.addOption(this.minLocalScoreOption);
        this.validArguments.addOption(this.reverseComplOption);
        this.validArguments.addOption(this.reverseComplOnlyOption);
        this.validArguments.addOption(this.compressBackTranslationOption);
        this.validArguments.addOption(this.expandGenomicSequencesOption);
        this.validArguments.addOption(this.enableCodonUsageOption);

        this.validArguments.addOption(this.doEstimateOption);
        this.validArguments.addOption(this.estimationMinScoreOption);
        this.validArguments.addOption(this.estimationMinScoreRCOption);
        this.validArguments.addOption(this.estimationSeq1LengthOption);
        this.validArguments.addOption(this.estimationSeq2LengthOption);
        this.validArguments.addOption(this.estimationSeqCountOption);
        this.validArguments.addOption(this.estimationParamsFileOption);
    }

    public void showUsage()
    {
        showUsage(System.out);
        System.exit(0);
    }

    public void showUsage(PrintStream out)
    {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp(100, "java sequoia.path.main.Main",
            "Either a pair of sequences, a database file, or a random database format must be selected.",
            this.validArguments, "", true);
    }

    public Parameters parse(String args[]) throws Exception
    {
        this.arguments = this.parser.parse(this.validArguments, args);
        if (this.arguments.hasOption("h")) {
            showUsage();
            return null;
        }
        Parameters params = new Parameters();
        if (this.arguments.hasOption(this.quietOption.getOpt())) {
            params.setQuiet();
        } else if (this.arguments.hasOption(this.verboseOption.getOpt())) {
            params.setVerbose();
        }
        if (this.arguments.hasOption(this.progressBarOption.getOpt())) {
            params.setProgressBar();
        }
        if (this.arguments.hasOption(this.sequenceTypeOption.getOpt())) {
            params.sequenceType = SequenceType.valueOfLabel(this.sequenceTypeOption.getValue());
        }
        if (this.arguments.hasOption(this.estimationParamsFileOption.getOpt())) {
            if (!this.arguments.hasOption("estim")) {
                params.evaluator.load(this.estimationParamsFileOption.getValue());
            }
        }
        if ((!StringUtils.isBlank(this.firstSequenceFileOption.getValue()) || !StringUtils
            .isBlank(this.firstSequenceOption.getValue()))
            && (!StringUtils.isBlank(this.secondSequenceFileOption.getValue()) || !StringUtils
                .isBlank(this.secondSequenceOption.getValue()))) {
            params.setData(new DoubleDBData(params));
            if (!StringUtils.isBlank(this.firstSequenceOption.getValue())) {
                params.getData().addSeq(this.firstSequenceOption.getValue());
            } else {
                loadSequenceFromFile(this.firstSequenceFileOption.getValue(), params);
            }
            ((DoubleDBData) params.getData()).setBound(params.getData().seqsCount());
            if (!StringUtils.isBlank(this.secondSequenceOption.getValue())) {
                params.getData().addSeq(this.secondSequenceOption.getValue());
            } else {
                loadSequenceFromFile(this.secondSequenceFileOption.getValue(), params);
            }
            params.getData().doneLoading();
        } else if (DBFormatType.DEFAULT.label.equals(this.databaseFormatOption.getValue())) {
            if (this.arguments.hasOption("estim")) {
                if (this.gappingTypeOption.getValue() != null
                    && !this.gappingTypeOption.getValue().equals(Parameters.GapType.FS_CODON.label)) {
                    System.err.println("Estimation is currently available only for frameshift/codon gapping type.");
                    System.exit(0);
                }
                params.estimator = new AlignmentScoreSignificanceEstimation();
                params.estimator.setParamFile(this.estimationParamsFileOption.getValue());
                params.estimator.setSequenceCount(NumberUtils.toInt(this.estimationSeqCountOption.getValue(),
                    AlignmentScoreSignificanceEstimation.DEFAULT_RAND_SEQ_NO));
                params.estimator.setSequence1Length(NumberUtils.toInt(this.estimationSeq1LengthOption.getValue(),
                    AlignmentScoreSignificanceEstimation.DEFAULT_RAND_SEQ_LEN));
                params.estimator.setSequence2Length(NumberUtils.toInt(this.estimationSeq2LengthOption.getValue(),
                    AlignmentScoreSignificanceEstimation.DEFAULT_RAND_SEQ_LEN));
                params.estimator.setMinScoreFW(NumberUtils.toDouble(this.estimationMinScoreOption.getValue(),
                    AlignmentScoreSignificanceEstimation.DEFAULT_MIN_SCORE));
                params.estimator.setMinScoreRC(NumberUtils.toDouble(this.estimationMinScoreRCOption.getValue(),
                    params.estimator.getMinScoreFW()));
            }
            params.dbType =
                DBFormatType.valueOfLabel(StringUtils.defaultIfEmpty(this.databaseFormatOption.getValue(),
                    DBFormatType.DEFAULT.label));
            params.databaseFile = this.databaseFileOption.getValue();
            loadDatabase(params);
        } else if (!StringUtils.isBlank(this.databaseFileOption.getValue())) {
            params.dbType =
                DBFormatType.valueOfLabel(StringUtils.defaultIfEmpty(this.databaseFormatOption.getValue(),
                    DBFormatType.DEFAULT.label));
            params.databaseFile = this.databaseFileOption.getValue();
            params.sampleRate =
                NumberUtils.toDouble(this.databaseSamplingOption.getValue(), Parameters.DEFAULT_DB_RATE);
            params.getData().setStartIdx(NumberUtils.toInt(this.databaseIdxStartOption.getValue(), 0));
            ((SingleDBData) params.getData()).setStartId(this.databaseSeqIDStartOpion.getValue());
            if (!params.isQuiet()) {
            	System.out.println("Starting at " + this.databaseSeqIDStartOpion.getValue());
            }
            params.getData().setEndIdx(NumberUtils.toInt(this.databaseIdxEndOption.getValue(), -1));
            params.dbFilter = this.databaseFilterStringOption.getValue();
            loadDatabase(params);
        } else {
            showUsage();
            return null;
        }

        params.aType =
            AlignmentType.valueOfLabel(StringUtils.defaultIfEmpty(this.alignmentTypeOption.getValue(),
                Parameters.AlignmentType.DEFAULT.label));
        params.gType =
            GapType.valueOfLabel(StringUtils.defaultIfEmpty(this.gappingTypeOption.getValue(),
                Parameters.GapType.DEFAULT.label));

        params.setMatch(NumberUtils.toInt(this.matchScoreOption.getValue(), Alignment.DEFAULT_MATCH_SCORE));
        params.setMismatch(NumberUtils.toInt(this.mismatchScoreOption.getValue(), Alignment.DEFAULT_MISMATCH_PENALTY));
        if (params.getMismatch() > 0) {
            params.setMismatch(-params.getMismatch());
        }
        params.tmismatch = NumberUtils.toInt(this.transitionMutationScoreOption.getValue(), params.getMismatch());
        if (params.tmismatch > 0) {
            params.tmismatch = -params.tmismatch;
        }
        params.setG(NumberUtils.toInt(this.linearGapPenaltyOption.getValue(), Alignment.DEFAULT_LIN_GAP_PENALTY));
        if (params.getG() > 0) {
            params.setG(-params.getG());
        }
        params.setD(NumberUtils.toInt(this.affineGapOpeningPenaltyOption.getValue(),
            Alignment.DEFAULT_AFINE_GAP_OPEN_PENALTY));
        if (params.getD() > 0) {
            params.setD(-params.getD());
        }
        params.setE(NumberUtils.toInt(this.affineGapExtensionPenaltyOption.getValue(),
            Alignment.DEFAULT_AFINE_GAP_EXTEND_PENALTY));
        if (params.getE() > 0) {
            params.setE(-params.getE());
        }
        params.setCodonSkip(NumberUtils.toInt(this.codonGapPenaltyOption.getValue(),
            Alignment.DEFAULT_CODONSKIP_PENALTY));
        if (params.getCodonSkip() > 0) {
            params.setCodonSkip(-params.getCodonSkip());
        }
        params.setCodonSkipExtend(NumberUtils.toInt(this.codonGapExtensionPenaltyOption.getValue(), params
            .getCodonSkip()));
        if (params.getCodonSkipExtend() > 0) {
            params.setCodonSkipExtend(-params.getCodonSkip());
        }
        params.setAllowedFrameshifts(NumberUtils.toInt(this.allowedFrameshiftCount.getValue(),
            Alignment.DEFAULT_ALLOWED_FRAMESHIFTS));
        params.setMinScore(NumberUtils.toInt(this.minLocalScoreOption.getValue(), Alignment.DEFAULT_MIN_SCORE));

        if (this.arguments.hasOption(this.compressBackTranslationOption.getOpt())) {
            params.compressGraphs = !BackTranslator.DEFAULT_GRAPH_FORMAT;
        }

        if (this.arguments.hasOption(this.expandGenomicSequencesOption.getOpt())) {
            params.expandGenomicSequences = !BackTranslator.DEFAULT_EXPAND_GENOMIC_SEQUENCE_GRAPHS;
        }

        if (!StringUtils.isBlank(this.outputOption.getValue())) {
            try {
                params.out = new PrintStream(new FileOutputStream(this.outputOption.getValue() + this.fileNameSufix));
                if (!params.isQuiet()) {
                  System.out.println("Output file (2): " + this.outputOption.getValue() + this.fileNameSufix);
                }
            } catch (Exception ex) {
                System.err.println("The output file " + this.outputOption.getValue() + this.fileNameSufix
                    + " cannot be opened. The alignments will be printed to stdout instead.");
                // ex.printStackTrace();
                // System.exit(0);
                params.out = System.out;
            }
        }

        if (!StringUtils.isBlank(this.selectionOutputOption.getValue())) {
            try {
                params.setSelOut(new PrintStream(new FileOutputStream(this.selectionOutputOption.getValue()
                    + this.fileNameSufix)));
                if (!params.isQuiet()) {
                  System.out.println("Output file: " + this.selectionOutputOption.getValue() + this.fileNameSufix);
                }
            } catch (Exception ex) {
                System.err.println("The output file " + this.selectionOutputOption.getValue() + this.fileNameSufix
                    + " cannot be opened. The alignments will be printed to the default file (" + "selectedAlignments"
                    + this.fileNameSufix + ") instead.");
                params.setSelOut(new PrintStream(new FileOutputStream("selectedAlignments" + this.fileNameSufix)));
                ;
            }
        }

        if (!StringUtils.isBlank(this.xmlOutputOption.getValue())) {
            try {
                params.setXmlOut(new PrintStream(new FileOutputStream(this.xmlOutputOption.getValue())));
                if (!params.isQuiet()) {
                   System.out.println("XML output: " + this.xmlOutputOption.getValue());
                }
            } catch (Exception ex) {
                ex.printStackTrace();
                params.setXmlOut(null);
                ;
            }
        }

        EValueAlignmentFilter evFilter =
            new EValueAlignmentFilter(NumberUtils.toDouble(this.selectionEValThresholdOption.getValue(),
                EValueAlignmentFilter.DEFAULT_EVALUE_THRESHOLD));
        params.filters.add(evFilter);
        Alignment.setFilter(evFilter);
        if (this.arguments.hasOption("FS")) {
            params.filters.add(new HasFrameshiftsFilter());
        }

        try {
            params.getData().setScoreMatrix(new ContextScoreMatrix(this.scoreMatrixOption.getValue()));
            if (params.compressGraphs) {
                params.getData().setScoreMatrix(
                    new CompressedContextScoreMatrix((ContextScoreMatrix) params.getData().getScoreMatrix()));
            }
        } catch (Exception ex) {
            try {
                ObjectInputStream ois =
                    new ObjectInputStream(new java.io.BufferedInputStream(new java.io.FileInputStream(
                        this.scoreMatrixOption.getValue())));
                params.getData().setScoreMatrix((ScoreMatrix) ois.readObject());
                boolean isCompressedMatrix =
                    params.getData().getScoreMatrix().getClass().getName().contains("CompressedContextScoreMatrix");
                if (isCompressedMatrix) {
                    params.compressGraphs = true;
                } else if (params.compressGraphs) {
                    params.getData().setScoreMatrix(
                        new CompressedContextScoreMatrix((ContextScoreMatrix) params.getData().getScoreMatrix()));
                }
            } catch (Exception ex2) {
                // ex.printStackTrace();
                params.getData().setScoreMatrix(loadScoreMatrix(this.scoreMatrixOption.getValue(), params));
            }
        }

        if (this.arguments.hasOption("rc")) {
            params.reverseCompl = true;
        }
        if (this.arguments.hasOption("rcOnly")) {
            params.reverseCompl = true;
            params.reverseComplExclusive = true;
        }
        if (this.arguments.hasOption("cu")) {
            params.useCodonUsage = true;
        }
        params.process();
        return params;
    }

    private void loadSequenceFromFile(String file, Parameters params) throws Exception
    {
        BufferedReader br = new BufferedReader(new FileReader(file));
        StringBuilder seq = new StringBuilder();
        String seqPart = br.readLine();
        // is it a fasta file?
        if (seqPart != null && seqPart.startsWith(">")) {
            // might even be a multi fasta...
            br.close();
            loadFastaFile(file, params);
            return;
        }
        while (seqPart != null) {
            seq.append(seqPart.trim());
            seqPart = br.readLine();
        }
        br.close();
    }

    private void loadDatabase(Parameters params)
    {
        if (params.dbType.equals(DBFormatType.MULTI_FASTA_DIR)) {
            loadFasta(params);
        } else if (params.dbType.equals(DBFormatType.MULTI_FASTA)) {
            loadFastaFile(params.databaseFile, params);
        } else if (params.dbType.equals(DBFormatType.RANDOMLY_GENERATED)) {
            generateRandSeq(params);
        } else if (params.dbType.equals(DBFormatType.PLAIN_TEXT)) {
            loadSimpleDataFile(params);
        }
        params.getData().doneLoading();
    }

    private void generateRandSeq(Parameters params)
    {
        RandSeqGenerator.loadAAintervals();
        RandSeqGenerator rsg = new RandSeqGenerator();
        int len1 = params.estimator.getSequence1Length();
        int len2 = params.estimator.getSequence2Length();
        if (len1 < len2) {
            len2 = len1;
            len1 = params.estimator.getSequence2Length();
        }
        params.setData(new DoubleDBData(params));
        for (int i = 0; i < params.estimator.getSequenceCount() / 2; i++) {
            params.getData().addSeq(rsg.generateSeq(len1));
        }
        ((DoubleDBData) params.getData()).setBound(params.getData().seqsCount());
        for (int i = 0; i < params.estimator.getSequenceCount() / 2; i++) {
            params.getData().addSeq(rsg.generateSeq(len2));
        }
        params.getData().doneLoading();
    }

    private void loadSimpleDataFile(Parameters params)
    {
        BufferedReader in = null;
        try {
            in = new BufferedReader(new FileReader(params.databaseFile));
            String line = in.readLine();

            while (line != null) {
                params.getData().addSeq(line);
                line = in.readLine();
            }
        } catch (IOException ioex) {
            ioex.printStackTrace();
        } finally {
            if (in != null) {
                try {
                    in.close();
                } catch (IOException e) {
                    // Shouldn't happen...
                    e.printStackTrace();
                }
            }
        }
    }

    private void loadFasta(Parameters params)
    {
        // *.fasta
        FilenameFilter filter = new FilenameFilter()
        {
            public boolean accept(File dir, String name)
            {
                return name.endsWith(".fasta");
            }
        };
        File db = new File(params.databaseFile);
        File[] fams = db.listFiles(filter);
        for (int i = 0; i < fams.length; ++i) {
            loadFastaFile(fams[i], params);
        }
    }

    private void loadFastaFile(File f, Parameters params)
    {
        try {
            BufferedReader in = new BufferedReader(new FileReader(f));
            loadFastaFile(in, params);
            in.close();
        } catch (IOException ioex) {
            ioex.printStackTrace();
        }
    }

    private void loadFastaFile(String filename, Parameters params)
    {
        try {
            BufferedReader in = new BufferedReader(new FileReader(filename));
            loadFastaFile(in, params);
            in.close();
        } catch (IOException ioex) {
            ioex.printStackTrace();
        }
    }

    private void loadFastaFile(BufferedReader in, Parameters params) throws IOException
    {
        String line = in.readLine(), seq, seqID = "";
        StringBuilder seqB = new StringBuilder();
        while (line != null) {
            if (line.startsWith("#")) {
                // comment
            } else if (line.startsWith(">")) {
                seq = seqB.toString();// .toUpperCase(); //TODO fix me
                if (seq.length() != 0) { // it's not the first sequence
                    params.getData().addSeq(seq, seqID);
                    seqB.delete(0, seqB.length());
                }
                seqID = (line.substring(line.indexOf('>'))).trim();
            } else {
                seqB.append(line.trim());
            }
            line = in.readLine();
            continue;
        }
        // add the last sequence that remained in the builder
        params.getData().addSeq(seqB.toString().trim(), seqID);
    }

    private SimpleScoreMatrix loadScoreMatrix(String inputFile, Parameters params)
    {
        try {
            return new SimpleScoreMatrix(inputFile);

        } catch (Exception e) {
            System.out.println("No score matrix...");
            // e.printStackTrace();
            return new SimpleScoreMatrix(params.getMatch(), params.getMismatch(), params.tmismatch);
        }
    }
}
