package sequoia.path.params;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import sequoia.path.alignment.estimation.AlignmentScoreSignificanceEstimation;
import sequoia.path.alignment.estimation.ScoreEvaluationParameters;
import sequoia.path.alignment.filter.AlignmentFilter;
import sequoia.path.data.Data;
import sequoia.path.data.SingleDBData;
import sequoia.path.translation.BackTranslator;

public class Parameters
{
	public static enum AlignmentType {
		GLOBAL("g"), LOCAL("l"), SEMI_GLOBAL("o");
		
		public static final AlignmentType DEFAULT = LOCAL;
		
		private static Map<String, AlignmentType> map = new HashMap<String, AlignmentType>();
		static {
			for (AlignmentType v : values()) {
				map.put(v.label, v);
			}
		}
		
		String label;
		
		AlignmentType(String label) {
			this.label = label;
		}
		
		public static AlignmentType valueOfLabel(String label) {
			return map.get(label) != null ? map.get(label) : DEFAULT;
		}
		
		public static String generateOptionList() {
			StringBuilder str = new StringBuilder();
			for (AlignmentType v : values()) {
				str.append('|');
				str.append(v.label);
			}
			return str.substring(1);
		}
	}
	
	public static enum GapType {
		LINEAR("l"), AFFINE("a"), FS_CODON("f");
		
		public static final GapType DEFAULT = FS_CODON;
		
		private static Map<String, GapType> map = new HashMap<String, GapType>();
		static {
			for (GapType v : values()) {
				map.put(v.label, v);
			}
		}
		
		String label;
		
		GapType(String label) {
			this.label = label;
		}
		
		public static GapType valueOfLabel(String label) {
			return map.get(label) != null ? map.get(label) : DEFAULT;
		}
		
		public static String generateOptionList() {
			StringBuilder str = new StringBuilder();
			for (GapType v : values()) {
				str.append('|');
				str.append(v.label);
			}
			return str.substring(1);
		}
	}

	public static enum SequenceType {
		PROTEIN("prot"), DNA("dna"), RNA("rna"), AUTO_DETECT("auto");
		
		public static final SequenceType DEFAULT = AUTO_DETECT;
		
		private static Map<String, SequenceType> map = new HashMap<String, SequenceType>();
		static {
			for (SequenceType v : values()) {
				map.put(v.label, v);
			}
		}
		
		String label;
		
		SequenceType(String label) {
			this.label = label;
		}
		
		public static SequenceType valueOfLabel(String label) {
			return map.get(label) != null ? map.get(label) : DEFAULT;
		}
		
		public static String generateOptionList() {
			StringBuilder str = new StringBuilder();
			for (SequenceType v : values()) {
				str.append('|');
				str.append(v.label);
			}
			return str.substring(1);
		}
	}
	
	public static enum DBFormatType {
		MULTI_FASTA("f"), MULTI_FASTA_DIR("fd"), PLAIN_TEXT("s"), RANDOMLY_GENERATED("r");
		
		public static final DBFormatType DEFAULT = MULTI_FASTA;
		
		private static Map<String, DBFormatType> map = new HashMap<String, DBFormatType>();
		static {
			for (DBFormatType v : values()) {
				map.put(v.label, v);
			}
		}
		
		String label;
		
		DBFormatType(String label) {
			this.label = label;
		}
		
		public static DBFormatType valueOfLabel(String label) {
			return map.get(label) != null ? map.get(label) : DEFAULT;
		}
		
		public static String generateOptionList() {
			StringBuilder str = new StringBuilder();
			for (DBFormatType v : values()) {
				str.append('|');
				str.append(v.label);
			}
			return str.substring(1);
		}
	}

    public static final double DEFAULT_DB_RATE = 0.1;


    public ScoreEvaluationParameters evaluator = new ScoreEvaluationParameters();
    
    public AlignmentScoreSignificanceEstimation estimator = null;

    public ArrayList<AlignmentFilter> filters = new ArrayList<AlignmentFilter>();

    PrintStream out;

    private PrintStream selOut;
    
    private PrintStream xmlOut;
    
    public boolean progressBar = false;

    public boolean verbose = false;

    public boolean reverseCompl = false;
    
    public boolean reverseComplExclusive = false;

    public boolean compressGraphs = BackTranslator.DEFAULT_GRAPH_FORMAT;

    public boolean expandGenomicSequences = BackTranslator.DEFAULT_EXPAND_GENOMIC_SEQUENCE_GRAPHS;

    public boolean useCodonUsage = false;

    public boolean quiet = false;

    public SequenceType sequenceType = SequenceType.DEFAULT;

    public double sampleRate;

    private Data data = new SingleDBData(this);

    /** Alignment type: global, local, semi-global.  */
    public AlignmentType aType;

    /** Gapping type: linear/afine/non-monotonous */
    public GapType gType;

    public String databaseFile;

    public String dbFilter;

    public DBFormatType dbType;

    int tmismatch;

    private int e;

    private int g;

    private int d;

    private int minScore;

    private int allowedframeshifts;

    private int mismatch;

    private int match;

    private int codonSkip;
    
    private int codonSkipExtend;

    public boolean isQuiet()
    {
        return quiet;
    }

    public void setQuiet()
    {
        setQuiet(true);
    }

    public void setQuiet(boolean quiet)
    {
        this.quiet = quiet;
    }

    public boolean isVerbose()
    {
        return verbose;
    }

    public void setVerbose()
    {
        setVerbose(true);
    }

    public void setVerbose(boolean verbose)
    {
        this.verbose = verbose;
    }
    
    public boolean showProgressBar()
    {
        return progressBar;
    }

    public void setProgressBar()
    {
    	setProgressBar(true);
    }

    public void setProgressBar(boolean progressBar)
    {
        this.progressBar = progressBar;
    }

    public void process()
    {
        BackTranslator at = new BackTranslator(this.compressGraphs, this.expandGenomicSequences);

        if (SequenceType.AUTO_DETECT.equals(sequenceType)) {
            at.setSeqType(BackTranslator.BACK_TRANSLATE_UNDEFINED);
        } else if (SequenceType.DNA.equals(sequenceType)) {
            at.setSeqType(BackTranslator.BACK_TRANSLATE_DNA);
        } else if (SequenceType.RNA.equals(sequenceType)) {
            at.setSeqType(BackTranslator.BACK_TRANSLATE_RNA);
        } else if (SequenceType.PROTEIN.equals(sequenceType)) {
            at.setSeqType(BackTranslator.BACK_TRANSLATE_PROTEIN);
        }
        for (int j = 0; j < getData().seqsCount(); ++j) {
            getData().getGraphs().add(at.backTranslate(getData().getSeq(j)));
        }
    }

    public void closeOutput()
    {
        try {
            if (out != System.out) {
                out.flush();
                out.close();
            }
            if (getSelOut() != System.out) {
                getSelOut().flush();
                getSelOut().close();
            }
            if (getXmlOut() != null) {
            	getXmlOut().flush();
            	getXmlOut().close();
            }
        } catch (Exception ex) {
        }
    }

    public void setCodonSkip(int codonSkip)
    {
        this.codonSkip = codonSkip;
    }

    public int getCodonSkip()
    {
        return codonSkip;
    }

    public int getCodonSkipExtend() {
		return codonSkipExtend;
	}

	public void setCodonSkipExtend(int codonSkipExtend) {
		this.codonSkipExtend = codonSkipExtend;
	}

	public void setMatch(int match)
    {
        this.match = match;
    }

    public int getMatch()
    {
        return match;
    }

    public void setMismatch(int mismatch)
    {
        this.mismatch = mismatch;
    }

    public int getMismatch()
    {
        return mismatch;
    }

    public void setAllowedFrameshifts(int allowedframeshifts)
    {
        this.allowedframeshifts = allowedframeshifts;
    }

    public int getAllowedFrameshifts()
    {
        return allowedframeshifts;
    }

    public void setMinScore(int minScore)
    {
        this.minScore = minScore;
    }

    public int getMinScore()
    {
        return minScore;
    }

    public void setD(int d)
    {
        this.d = d;
    }

    public int getD()
    {
        return d;
    }

    public void setG(int g)
    {
        this.g = g;
    }

    public int getG()
    {
        return g;
    }

    public void setE(int e)
    {
        this.e = e;
    }

    public int getE()
    {
        return e;
    }

    public void setData(Data data)
    {
        this.data = data;
    }

    public Data getData()
    {
        return data;
    }

    public void setSelOut(PrintStream selOut)
    {
        this.selOut = selOut;
    }

    public PrintStream getSelOut()
    {
        return selOut;
    }
    
    public void setXmlOut(PrintStream selOut)
    {
        this.xmlOut = selOut;
    }

    public PrintStream getXmlOut()
    {
        return xmlOut;
    }
}
