package sequoia.path.scorematrixbuilder.directcomputation;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;

import sequoia.path.scorematrixbuilder.matrix.SquareMatrix;
import sequoia.path.scores.ContextScoreMatrix;
import sequoia.path.scores.MatrixAnalyzer;
import sequoia.path.scores.SymbolData;
import sequoia.path.scores.SymbolPairAlignmentData;
import sequoia.path.util.Codon;
import sequoia.path.util.GenomicSequenceUtils;

public class ContextScoreMatrixBuilder
{
    public static final long serialVersionUID = 20080616;

    // 
    public static final int MATRIX_SEPARATION_NONE = 1;

    public static final int MATRIX_SEPARATION_SENSE = 2;

    public static final int MATRIX_SEPARATION_PHASE = 2 * Codon.CODON_LEN;

    public static final int MATRIX_SEPARATION_POS = 2 * Codon.CODON_LEN * Codon.CODON_LEN;

    public static final int DEFAULT_MATRIX_SEPARATION = MATRIX_SEPARATION_POS;

    /**
     * Chosen matrix separation criterion; it also gives the number of actually obtained matrices. This separation
     * affects the normalization procedure.
     */
    private int indepMatrixCount;

    /** The result: computed score matrix */
    private ContextScoreMatrix scores = new ContextScoreMatrix();

    // Inputs:
    /**
     * Input 1): The codons (with all the info: sequence, encoded amino-acid, frequency <strong>wrt the encoded
     * amino-acid</strong>. Most importantly, this vector also gives sense to the other input data, as it shows the
     * order of the codons.
     */
    protected Codon[] codons;

    /**
     * Input 2): The codon mutation probabilities: each line in the matrix contains the probability of the label codon
     * to mutate in the other codons.
     */
    protected SquareMatrix codonMutationProb;

    // We may want to consider sequences with different compositions:
    /** Input 3a): The codon frequencies in the first sequence */
    protected double[] codonFreqs1;

    /** Input 3b): The codon frequencies in the second sequence */
    protected double[] codonFreqs2;

    // Derived
    /** Input 3b): The codon frequencies in the second sequence */
    protected double[] baseFreqs;

    // Temporary data
    /** Temporary data: foreground probability "matrix" (map) */
    private HashMap<SymbolPairAlignmentData, Double> fgProbs[];

    /** Temporary data: background probability "matrix", computed from the codon distribution table (map) */
    private HashMap<SymbolPairAlignmentData, Double> bgProbsCodonDistrib[];

    /** Temporary data: background probability "matrix" (map), computed as marginal probability from the fg */
    private HashMap<SymbolPairAlignmentData, Double> bgProbsMarginal[];

    /** Temporary data: the labels of each probability map, for a more comprehensive display */
    private String mapLabels[];

    // -------------------------------------------------------------------------------------------
    // CONSTRUCTORS
    // -------------------------------------------------------------------------------------------
    /**
     * Dummy constructor for internal use.
     */
    protected ContextScoreMatrixBuilder()
    {
    }

    /**
     * Initializes the builder with the codon mutation and frequency data, and the default matrix separation criterion.
     * We consider here the case where both sequences have the same codon composition.
     */
    public ContextScoreMatrixBuilder(Codon[] codons, SquareMatrix codonMutationProb, double codonFreqs[])
    {
        this(codons, codonMutationProb, codonFreqs, codonFreqs, DEFAULT_MATRIX_SEPARATION);
    }

    /**
     * Initializes the builder with the codon mutation and frequency data, and defines the matrix separation criterion.
     * We consider here the case where both sequences have the same codon composition. The default matrix separation is
     * chosen if the input separation is invalid.
     */
    public ContextScoreMatrixBuilder(Codon[] codons, SquareMatrix codonMutationProb, double[] codonFreqs, int separation)
    {
        this(codons, codonMutationProb, codonFreqs, codonFreqs, separation);
    }

    /**
     * Initializes the builder with the codon mutation and frequency data, and the default matrix separation criterion.
     * We consider here the case where the sequences have different codon compositions.
     */
    public ContextScoreMatrixBuilder(Codon[] codons, SquareMatrix codonMutationProb, double codonFreqs1[],
        double codonFreqs2[])
    {
        this(codons, codonMutationProb, codonFreqs1, codonFreqs2, DEFAULT_MATRIX_SEPARATION);
    }

    /**
     * Initializes the builder with the codon mutation and frequency data, and defines the matrix separation criterion.
     * The default matrix separation is chosen if the input separation is invalid. We consider here the case where the
     * sequences have different codon compositions.
     */
    public ContextScoreMatrixBuilder(Codon[] codons, SquareMatrix codonMutationProb, double[] codonFreqs1,
        double codonFreqs2[], int separation)
    {
        this.codons = codons;
        this.codonMutationProb = codonMutationProb;
        this.codonFreqs1 = codonFreqs1;
        this.codonFreqs2 = codonFreqs2;
        this.indepMatrixCount = separation;
        computeBaseFreqs();
        createTemporaryMatrices();
        buildMatrix();
    }

    private void computeBaseFreqs()
    {
        this.baseFreqs = new double[GenomicSequenceUtils.ALPHABET_SIZE];
        Arrays.fill(this.baseFreqs, 0);
        double sum = 0;
        for (int i = 0; i < this.codons.length; ++i) {
            for (int j = 0; j < Codon.CODON_LEN; ++j) {
                this.baseFreqs[GenomicSequenceUtils.getNucleotideID(this.codons[i].getSymbol(j))] +=
                    this.codonFreqs1[i];
                this.baseFreqs[GenomicSequenceUtils.getNucleotideID(GenomicSequenceUtils
                    .getComplementary(this.codons[i].getSymbol(j)))] += this.codonFreqs1[i];
                sum += 2 * this.codonFreqs1[i];
            }
        }
        for (int i = 0; i < this.baseFreqs.length; ++i) {
            this.baseFreqs[i] /= sum;
        }

    }

    // -------------------------------------------------------------------------------------------
    // TEMPORARY DATA MANAGEMENT (FG AND BG)
    // -------------------------------------------------------------------------------------------
    /**
     * Creation of temporary foreground and background probability maps.
     */
    @SuppressWarnings("unchecked")
    private void createTemporaryMatrices()
    {
        if (this.indepMatrixCount != MATRIX_SEPARATION_NONE && this.indepMatrixCount != MATRIX_SEPARATION_SENSE
            && this.indepMatrixCount != MATRIX_SEPARATION_PHASE && this.indepMatrixCount != MATRIX_SEPARATION_POS) {
            this.indepMatrixCount = DEFAULT_MATRIX_SEPARATION;
        }
        this.fgProbs = (new HashMap[this.indepMatrixCount]);
        this.bgProbsCodonDistrib = (new HashMap[this.indepMatrixCount]);
        this.bgProbsMarginal = (new HashMap[this.indepMatrixCount]);
        this.mapLabels = new String[this.indepMatrixCount];
        for (int p1 = 0; p1 < Codon.CODON_LEN; ++p1) {
            for (int p2 = 0; p2 < Codon.CODON_LEN; ++p2) {
                this.mapLabels[getMapIndex(p1, p2, false)] = p1 + " <-> " + p2 + "   ";
                this.mapLabels[getMapIndex(p1, p2, true)] = p1 + " <-> " + p2 + " RC";
            }
        }

        for (int i = 0; i < this.indepMatrixCount; ++i) {
            this.fgProbs[i] = new HashMap<SymbolPairAlignmentData, Double>();
            this.bgProbsCodonDistrib[i] = new HashMap<SymbolPairAlignmentData, Double>();
            this.bgProbsMarginal[i] = new HashMap<SymbolPairAlignmentData, Double>();
        }
    }

    /**
     * Computes the index of the temporary map that holds the key s.
     * 
     * @param s The key
     * @return The computed index, that can be used to identify the wanted map in <tt>fgProbs</tt> or <tt>bgProbs</tt>.
     */
    private int getMapIndex(SymbolPairAlignmentData s)
    {
        return getMapIndex(s.getPos1(), s.getPos2(), s.getS2().isRC() || s.getS1().isRC());
    }

    /**
     * Computes the index of the temporary map that holds the data which complied with the given characteristics
     * (aligned codon positions, and whether it is reverse complementary).
     * 
     * @param p1 Position in the codon of the first sequence.
     * @param p2 Position in the codon of the second sequence.
     * @param rc Reverse complementary
     * @return The computed index, that can be used to identify the wanted map in <tt>fgProbs</tt> or <tt>bgProbs</tt>.
     */
    private int getMapIndex(int p1, int p2, boolean rc)
    {
        int phase = (p1 - p2 + Codon.CODON_LEN) % Codon.CODON_LEN;
        int absolutePhase = phase + (rc ? Codon.CODON_LEN : 0);
        if (this.indepMatrixCount == MATRIX_SEPARATION_SENSE) {
            return (rc ? 1 : 0);
        }
        if (this.indepMatrixCount == MATRIX_SEPARATION_PHASE) {
            return absolutePhase;
        }
        if (this.indepMatrixCount == MATRIX_SEPARATION_POS) {
            return absolutePhase * Codon.CODON_LEN + p1;
        }
        return 0;
    }

    /**
     * Identifies the temporary foreground probabilities map that holds the key s.
     * 
     * @param s
     * @return
     */
    private HashMap<SymbolPairAlignmentData, Double> getTargetFgMap(SymbolPairAlignmentData s)
    {
        return this.fgProbs[getMapIndex(s)];
    }

    /**
     * Makes sure the sum of the probabilities in the map is 1.
     */
    private void normalize(HashMap<Object, Double> map)
    {
        double sum = 0;
        for (Object s : map.keySet()) {
            sum += map.get(s);
        }
        for (Object s : map.keySet()) {
            map.put(s, map.get(s) / sum);
        }
    }

    /**
     * Make sure the sum for each map is 1
     */
    @SuppressWarnings("unchecked")
    private void normalize(HashMap maps[])
    {
        for (int i = 0; i < maps.length; ++i) {
            normalize(maps[i]);
        }
    }

    // -------------------------------------------------------------------------------------------
    // TRANSPARENT ACCESS TO THE INPUT DATA
    // -------------------------------------------------------------------------------------------
    private double getMutationProbability(Codon c1, Codon c2)
    {
        return getMutationProbability(c1.getID(), c2.getID());
    }

    private double getMutationProbability(int c1, int c2)
    {
        return this.codonMutationProb.get(c1, c2);
    }

    // -------------------------------------------------------------------------------------------
    // ACTUAL SCORE MATRIX CONSTRUCTION
    // -------------------------------------------------------------------------------------------
    /**
     * Builds the score matrix from the input data. Steps: 1) computing the foreground probabilities; 2) computing the
     * background probabilities; 3) obtaining the scores from this temporary data; 4) scaling the scores (rounding,
     * etc); 5) correcting the matrix keys (see {@link #correctMatrixKeys()})
     */
    private void buildMatrix()
    {
        computeForegroundProbabilities();
        // FOR DEBUG
        // displayPartialSums(fgProbs, "FG before normalization");
        normalize(this.fgProbs);
        // FOR DEBUG
        // displayPartialSums(fgProbs, "FG after normalization");

        computeCDBackgroundProbabilities();
        // displayPartialSums(bgProbsCodonDistrib, "CD BG");
        // computeMarginalBackgroundProbabilities();
        // displayPartialSums(bgProbsMarginal, "Mg BG");

        // displayBGRatios();

        // FOR DEBUG
        // displayPartialSums(bgProbs, "ABG");
        // FOR DEBUG
        // displayPartialSums(bgProbs, "BG after normalization");
        computeScores();
        scaleScores();
        correctMatrixKeys();

        // MatrixAnalyzer ma = new MatrixAnalyzer();
        //    	
        // for (int i = 0; i < indepMatrixCount; ++i) {
        // ma.printMap(System.out, fgProbs[i], "[FG] " + mapLabels[i]);
        // ma.printMap(System.out, bgProbs[i], "[BG] " + mapLabels[i]);
        // }
    }

    /**
     * After the score matrix is built, it cannot be immediately used in alignments, because of the way reverse
     * complementary cases are treated. To be more precise: The key ((A, 0, M, false), (C, 2, I, true)) will point to
     * the score of the pair (A, 0, M), reverseComplementary((C, 2, I)). The latter is in fact (G, 0, i) (the lower case
     * of the amino-acid symbol marks the fact that the coding sequence is reversed and complemented). In the alignment
     * algorithm, we work directly with reversed and complemented sequences and ask for the scores for the symbols we
     * see (i.e. G on the first position of a codon), not for symbols that in fact in the original sequence (i.e. C on
     * the third position of a codon). consequently, all keys involving reverse complementaries must be transformed as
     * follows:
     * 
     * <pre>
     * for each of the two symbols si composing the key,
     *   if (si is reverse complementary symbol) then
     *     si.nucleotide = Complementary (si.nucleotide)
     *     si.position = codonlength - si.position (-1)
     *   endif
     * endfor
     * </pre>
     */
    private void correctMatrixKeys()
    {

        // this shouldn't be here
        displayScoreExpectation();

        HashMap<SymbolPairAlignmentData, Double> tmp = new HashMap<SymbolPairAlignmentData, Double>();
        tmp.putAll(this.scores);
        this.scores.clear();
        for (SymbolPairAlignmentData s : tmp.keySet()) {
            SymbolPairAlignmentData sCorrected =
                new SymbolPairAlignmentData(getCorrectedSymbol(s.getS1()), getCorrectedSymbol(s.getS2()));
            this.scores.put(sCorrected, tmp.get(s));
        }
    }

    /**
     * see {@link #correctMatrixKeys()}
     * 
     * @param s
     * @return
     */
    private SymbolData getCorrectedSymbol(SymbolData s)
    {
        if (s.isRC()) {
            SymbolData result =
                new SymbolData(GenomicSequenceUtils.getComplementary(s.getN()), s.getAA(), Codon.CODON_LEN - s.getPos()
                    - 1, s.isRC());
            return result;
        } else {
            return s.clone();
        }
    }

    // -------------------------------------------------------------------------------------------
    // XXX FOREGROUND PROBABILITIES ZONE
    // -------------------------------------------------------------------------------------------
    /**
     * Computation of all the foreground probabilities. Todo: explain me!
     */
    private void computeForegroundProbabilities()
    {
        for (Codon c1 : this.codons) {
            for (Codon c2 : this.codons) {
                for (int f = 0; f < Codon.CODON_LEN; ++f) {
                    computeFor(c1, c2, f, false);
                    computeFor(c1, c2, f, true);
                }
            }
        }
    }

    /**
     * Obtains the probability that two symbols (identified by base, position in a codon and encoded amino-acid) be
     * evolutionary related.
     * 
     * @param s The pair of symbols.
     * @return The value attached to the key s in the foreground probabilities map that holds this key.
     */
    private double getForegroundProbability(SymbolPairAlignmentData s)
    {
        return getTargetFgMap(s).get(s);
    }

    /**
     * Updates the foreground probabilities with the information that can be obtained in the case of two input codons,
     * aligned with the input phase.
     * 
     * @param c1 First codon
     * @param c2 Second codon
     * @param phase Difference between the positions of the first base in each codon
     * @param rc True if the second codon is considered reversed and complemented.
     */
    private void computeFor(Codon c1, Codon c2, int phase, boolean rc)
    {

        // first overlapping zone
        int from1 = phase;
        int from2 = 0;
        int len = (Codon.CODON_LEN - from1);
        if (len > 0) {
            computeForZone(c1, c2, from1, from2, len, rc);
        }
        // second overlapping zone
        from1 = 0;
        from2 = Codon.CODON_LEN - phase;
        len = (Codon.CODON_LEN - from2);
        if (len > 0) {
            computeForZone(c1, c2, from1, from2, len, rc);
        }
    }

    /**
     * Updates the foreground probabilities with the information that can be obtained in the case of two input codons,
     * that overlap on the zone defined by the input parameters.
     * 
     * @param c1 First codon
     * @param c2 Second codon
     * @param from1 Start of the overlapping zone in the first codon
     * @param from2 Start of the overlapping zone in the second codon
     * @param len Length of the overlapping zone
     * @param rc True if the second codon is considered reversed and complemented.
     */
    private void computeForZone(Codon c1, Codon c2, int from1, int from2, int len, boolean rc)
    {
        // boolean debug = false;
        // int id1 = new Codon("TGG", 0.5, 'W').getID();
        // int id2 = new Codon("TCC", 0.5, 'S').getID();
        // int p1 = 1;
        // int p2 = 0;
        // if (((c1.getID() == id1 && c2.getID() == id2 && from1 == p1 && from2 == p2) ||
        // (c1.getID() == id2 && c2.getID() == id1 && from1 == p2 && from2 == p1)) && rc) {
        // debug = true;
        // }
        double p = 0;
        for (Codon c : this.codons) {
            int overlapping[] = getOverlappingCodons(c, from1, from2, len, rc);
            for (int co : overlapping) {
                // double part1 = getOverlapProbability(c.getID(), co, overlapping, rc, from1 >= from2);
                double part1 = getOverlapProbability(c.getID(), co, from1, from2, rc);
                double part2 = getMutationProbability(c, c1);
                double part3 = getMutationProbability(co, c2.getID());
                double delta = part1 * part2 * part3;
                p += delta;
                // p += getOverlapProbability(c.getID(), co, overlapping, from1 >= from2) *
                // getMutationProbability(c, c1) *
                // getMutationProbability(co, c2.getID());

            }
        }
        for (int i = 0; i < len; ++i) {
            int pos1 = from1 + i;
            int pos2 = from2 + i;
            if (rc) {
                pos2 = Codon.CODON_LEN - pos2 - 1;
            }
            updateForegroundProbability(c1.getSymbol(pos1), c2.getSymbol(pos2), pos1, pos2, c1.getAA(), c2.getAA(), rc,
                p);
        }
    }

    /**
     * Computes the probability of the two codons overlapping. ATM, this is considered to be the product of the
     * probabilities of the two codons. Scenario: c1 c1' ===-- --=== c2' c2 Formula: P(c1 overlaps c2) = P(c1) * P(c2 |
     * c1 overlaps c2) =
     * 
     * @param c1
     * @param c2
     * @param from1
     * @param from2
     * @param rc
     */
    protected double getOverlapProbability(int c1, int c2, int from1, int from2, boolean rc)
    {
        double seq1Prob = this.codonFreqs1[c1];
        double seq2Prob = this.codonFreqs2[c2];
        int len = Codon.CODON_LEN - (from1 == 0 ? from2 : from1);
        if (len < Codon.CODON_LEN) {
            seq1Prob *= overlapCorrection(this.codons[c1], from1, len, rc, 1);
            seq2Prob *= overlapCorrection(this.codons[c2], from2, len, rc, 2);
        }
        return (seq1Prob + seq2Prob - seq1Prob * seq2Prob) / 2;
    }

    protected double overlapCorrection(Codon c, int mainFrom, int len, boolean rc, int which)
    {
        double correction = 0;
        int ovlen = Codon.CODON_LEN - len;
        int overlapFrom = mainFrom == 0 ? len : 0;

        int ov[] = getOverlappingCodons(c, overlapFrom, overlapFrom == 0 ? Codon.CODON_LEN - ovlen : 0, ovlen, rc);
        double codonFreqs[] = (which == 1) ? this.codonFreqs1 : this.codonFreqs2;
        for (int i : ov) {
            correction += codonFreqs[i];
        }
        return correction;
    }

    /**
     * Computes the probability of the two codons overlapping. ATM, this is considered to be the product of the
     * probabilities of the two codons. Formula: P(c1 overlaps c2) = P(c1) * P(c2 | c1 overlaps c2) = = P(c1) * (P(c2) /
     * sum(P(c' | c1 overlaps c'))) todo: is this the correct formula?
     * 
     * @param c1
     * @param c2
     * @return
     */
    protected double getOverlapProbability(int c1, int c2, int allOverlaps[], boolean rc, boolean order)
    {
        // double correction = 0;
        // for (int i : allOverlaps) {
        // correction += codonFreqs2[i];
        // }
        // if (correction == 0) {
        // //stop codon, needs to be ignored anyway
        // correction = 1;
        // }

        return this.codonFreqs1[c1] * this.codonFreqs2[c2];// / correction;
        //    	
        // Another way: do not compute this by codon usage, since, normally, the codon distribution
        // in overlapped coding sequences is atypical, due to the numerous constraints.
        // Instead, compute the probability of having, in a coding sequence, the subsequence that
        // describes the two overlapped codons. Depending on the length of the common subsequence
        // of the codons, the probability of a sequence containing 3, 4 or 5 nucleotides has to be
        // computed.
        // Although not realistic, we will start by considering all nucleotides equiprobable.

        // int overlapZoneSize = Codon.CODON_LEN - (int)Math.round((Math.log(allOverlaps.length) /
        // Math.log(GenomicSequenceUtils.ALPHABET_SIZE)));
        // double prob = 1;
        //    	
        // // Compute the probability of the sequence given by the first codon
        // for (int i = 0; i < Codon.CODON_LEN; ++i) {
        // prob *= getProbability(codons[c1].getSymbol(i));
        // }
        //    	
        // // order = from1 >= from2
        // /* Find out which part of the second codon's sequence is not comprized in the first
        // * There are 10 cases:
        // * 1) --->: order = true, rc = false
        // * --->
        // *
        // * 2) --->: order = true, rc = false
        // * --->
        // *
        // * 3) --->: order = true, rc = false
        // * --->
        // *
        // * 4) --->: order = false, rc = false
        // * --->
        // *
        // * 5) --->: order = false, rc = false
        // * --->
        // *
        // * 6) --->: order = true, rc = true
        // * <---
        // *
        // * 7) --->: order = true, rc = true
        // * <---
        // *
        // * 8) --->: order = true, rc = true
        // * <---
        // *
        // * 9) --->: order = false, rc = true
        // * <---
        // *
        // * 10) --->: order = false, rc = true
        // * <---
        // *
        // * Cases 1 and 6 are trivial, only the first codon's sequence needs to be taken into account.
        // * In cases 2, 3 (where only order is true) and cases 9, 10 (where only rc is true), the end
        // * of the second codon is the part not appearing in the first.
        // * In cases 4,5 (where both order and rc are false) and 7,8 (where both order and rc are true)
        // * the beginning of the second codon is the part not appearing in the first.
        // *
        // * Hence the formula:
        // */
        //    	
        // int startInC2 = 0;
        // int endInC2 = 0;
        // if (rc ^ order) {
        // startInC2 = overlapZoneSize;
        // endInC2 = Codon.CODON_LEN;
        // } else {
        // // startInC2 = 0; // already was zero
        // endInC2 = Codon.CODON_LEN - overlapZoneSize;
        // }
        // // Update the probability with the second codon fragment
        // for (int i = startInC2; i < endInC2; ++i) {
        // prob *= getProbability(codons[c2].getSymbol(i));
        // }
        // return prob;

    }

    // Todo: get the right base probability
    private double getProbability(char nucleotide)
    {
        return this.baseFreqs[GenomicSequenceUtils.getNucleotideID(nucleotide)];
    }

    /**
     * Retrieves all the codon that have a subsequence identical to a subsequence of codon c.
     * 
     * @param c
     * @param from1 Start of the overlapping zone in the input codon
     * @param from2 Start of the overlapping zone in the wanted codons
     * @param len Length of the overlapping zone
     * @param rc True if the second codon is considered reversed and complemented.
     * @return An array containing the indices if the wanted codons.
     */
    protected int[] getOverlappingCodons(Codon c, int from1, int from2, int len, boolean rc)
    {
        // if(rc) {
        // if (from1 > 0) {
        // from1 = Codon.CODON_LEN - from1 - 1;
        // } else {
        // from2 = Codon.CODON_LEN - from2 - 1;
        // }
        // len = Math.min(Codon.CODON_LEN - from1, Codon.CODON_LEN - from2);
        // }
        int count = 1;
        for (int i = 0; i < Codon.CODON_LEN - len; ++i) {
            count *= GenomicSequenceUtils.ALPHABET_SIZE;
        }
        int[] result = new int[count];
        // Extract the mask nucleotides from the base codon
        int base = 0;
        if (rc) {
            for (int i = from1 + len - 1; i >= from1; --i) {
                base *= GenomicSequenceUtils.ALPHABET_SIZE;
                base += GenomicSequenceUtils.getNucleotideID(GenomicSequenceUtils.getComplementary(c.getSymbol(i)));
            }
            from2 = Codon.CODON_LEN - (from2 + len);
        } else {
            for (int i = from1; i < from1 + len; ++i) {
                base *= GenomicSequenceUtils.ALPHABET_SIZE;
                base += GenomicSequenceUtils.getNucleotideID(c.getSymbol(i));
            }
        }
        // Shift the mask according to the positions on the target codons
        for (int i = from2 + len; i < Codon.CODON_LEN; ++i) {
            base *= GenomicSequenceUtils.ALPHABET_SIZE;
        }
        // Enumerate all the similar codons by completing the mask
        if (from2 > 0) {
            // 1/ the zone to complete is placed before the mask
            int shift = 1;
            for (int i = from2; i < Codon.CODON_LEN; ++i) {
                shift *= GenomicSequenceUtils.ALPHABET_SIZE;
            }
            for (int i = 0; i < result.length; ++i) {
                result[i] = base + i * shift;
            }
        } else {
            // 2/ the zone to complete is placed after the mask
            for (int i = 0; i < result.length; ++i) {
                result[i] = base + i;
            }
        }
        return result;
    }

    /**
     * Updates the probability to match two symbols (each defined by base, position in a codon and encoded amino-acid).
     * 
     * @param n1
     * @param n2
     * @param p1
     * @param p2
     * @param aa1
     * @param aa2
     * @param rc
     * @param value
     */
    private void updateForegroundProbability(char n1, char n2, int p1, int p2, char aa1, char aa2, boolean rc,
        double value)
    {
        SymbolPairAlignmentData s = new SymbolPairAlignmentData(n1, n2, aa1, aa2, p1, p2, rc);
        updateForegroundProbability(s, value);
    }

    /**
     * Updates the probability to match a symbol pair (each defined by base, position in a codon and encoded
     * amino-acid).
     * 
     * @param s
     * @param value
     */
    private void updateForegroundProbability(SymbolPairAlignmentData s, double value)
    {
        int t = getMapIndex(s);
        Double valO = this.fgProbs[t].get(s);
        if (valO == null) {
            this.fgProbs[t].put(s, value);
        } else {
            this.fgProbs[t].put(s, valO.doubleValue() + value);
        }
    }

    // -------------------------------------------------------------------------------------------
    // XXX BACKGROUND PROBABILITIES ZONE (codon distribition)
    // -------------------------------------------------------------------------------------------
    /**
     * Computation of all the background probabilities. Todo: explain me!
     */
    private void computeCDBackgroundProbabilities()
    {
        for (Codon c1 : this.codons) {
            for (Codon c2 : this.codons) {
                for (int f = 0; f < Codon.CODON_LEN; ++f) {
                    computeBGFor(c1, c2, f, false);
                    computeBGFor(c1, c2, f, true);
                }
            }
        }
    }

    /**
     * Obtains the probability that two symbols (identified by base, position in a codon and encoded amino-acid) are not
     * evolutionary related.
     * 
     * @param s The pair of symbols.
     * @return The value attached to the key s in the background probabilities map that holds this key.
     */
    private double getBackgroundProbability(SymbolPairAlignmentData s)
    {
        // return getTargetBgMap(s).get(s);
        return this.bgProbsCodonDistrib[getMapIndex(s)].get(s);
        // return bgProbsMarginal[getMapIndex(s)].get(s);
    }

    /**
     * Updates the background probabilities with the information that can be obtained in the case of two input codons,
     * aligned with the input phase.
     * 
     * @param c1 First codon
     * @param c2 Second codon
     * @param phase Difference between the positions of the first base in each codon
     * @param rc True if the second codon is considered reversed and complemented.
     */
    private void computeBGFor(Codon c1, Codon c2, int phase, boolean rc)
    {
        // first "overlapping" zone
        int from1 = phase;
        int from2 = 0;
        int len = (Codon.CODON_LEN - from1);
        if (len > 0) {
            computeBGForZone(c1, c2, from1, from2, len, rc);
        }
        // second "overlapping" zone
        from1 = 0;
        from2 = Codon.CODON_LEN - phase;
        len = (Codon.CODON_LEN - from2);
        if (len > 0) {
            computeBGForZone(c1, c2, from1, from2, len, rc);
        }
    }

    /**
     * Updates the background probabilities with the information that can be obtained in the case of two input codons,
     * that DO NOT have a common subsequence on the zone defined by the input parameters.
     * 
     * @param c1 First codon
     * @param c2 Second codon
     * @param from1 Start of the overlapping zone in the first codon
     * @param from2 Start of the overlapping zone in the second codon
     * @param len Length of the overlapping zone
     * @param rc True if the second codon is considered reversed and complemented.
     */
    private void computeBGForZone(Codon c1, Codon c2, int from1, int from2, int len, boolean rc)
    {
        for (int i = 0; i < len; ++i) {
            int pos1 = from1 + i;
            int pos2 = from2 + i;
            if (rc) {
                pos2 = Codon.CODON_LEN - pos2 - 1;
            }
            updateBackgroundProbability(c1.getSymbol(pos1), c2.getSymbol(pos2), pos1, pos2, c1.getAA(), c2.getAA(), rc,
                this.codonFreqs1[c1.getID()] * this.codonFreqs2[c2.getID()]);
        }
    }

    /**
     * Updates the probability to match two symbols by chance (each defined by base, position in a codon and encoded
     * amino-acid).
     * 
     * @param n1
     * @param n2
     * @param p1
     * @param p2
     * @param aa1
     * @param aa2
     * @param rc
     * @param value
     */
    private void updateBackgroundProbability(char n1, char n2, int p1, int p2, char aa1, char aa2, boolean rc,
        double value)
    {
        SymbolPairAlignmentData s = new SymbolPairAlignmentData(n1, n2, aa1, aa2, p1, p2, rc);
        updateBackgroundProbability(s, value);
    }

    /**
     * Updates the probability to match a symbol pair by chance (each defined by base, position in a codon and encoded
     * amino-acid).
     * 
     * @param s
     * @param value
     */
    private void updateBackgroundProbability(SymbolPairAlignmentData s, double value)
    {
        int t = getMapIndex(s);
        Double valO = this.bgProbsCodonDistrib[t].get(s);
        if (valO == null) {
            this.bgProbsCodonDistrib[t].put(s, value);
        } else {
            this.bgProbsCodonDistrib[t].put(s, valO.doubleValue() + value);
        }
    }

    // -------------------------------------------------------------------------------------------
    // XXX BACKGROUND PROBABILITIES ZONE (marginal)
    // -------------------------------------------------------------------------------------------
    /**
     * Computation of all the background probabilities. Todo: explain me!
     */

    private void computeMarginalBackgroundProbabilities()
    {
        for (int i = 0; i < this.indepMatrixCount; ++i) {
            computeMarginalBackgroundProbabilities(i);
        }
    }

    private void computeMarginalBackgroundProbabilities(int i)
    {
        HashMap<SymbolData, Double> marg1 = new HashMap<SymbolData, Double>();
        HashMap<SymbolData, Double> marg2 = new HashMap<SymbolData, Double>();

        for (SymbolPairAlignmentData s : this.fgProbs[i].keySet()) {
            double val = this.fgProbs[i].get(s);
            updateMap(marg1, s.getS1(), val);
            updateMap(marg2, s.getS2(), val);
        }
        for (SymbolPairAlignmentData s : this.fgProbs[i].keySet()) {
            this.bgProbsMarginal[i].put(s, marg1.get(s.getS1()) * marg2.get(s.getS2()));
        }

    }

    private void updateMap(HashMap<SymbolData, Double> map, SymbolData key, double value)
    {
        Double crtVal = map.get(key);
        if (crtVal != null) {
            map.put(key, crtVal + value);
        } else {
            map.put(key, value);
        }
    }

    // -------------------------------------------------------------------------------------------
    // XXX SCORES ZONE
    // -------------------------------------------------------------------------------------------
    /**
     * Computation of the actual scores for each possible symbol pair using the temporary foreground and background
     * maps, and completion of the score matrix.
     */
    private void computeScores()
    {
        for (int i = 0; i < this.indepMatrixCount; ++i) {
            for (SymbolPairAlignmentData sp : this.fgProbs[i].keySet()) {
                this.scores.put(sp, computeScore(sp));
            }
        }
    }

    /**
     * Computation of the score for a gived symbol pair, using the temporary foreground and background maps.
     * 
     * @param sp The symbol pair
     * @return The core, equal to the logarithm of te foreground/background ratio.
     */
    private double computeScore(SymbolPairAlignmentData sp)
    {
        // return Math.log(fgFreqs.get(sp)/(bgFreqs.get(sp.getS1())*bgFreqs.get(sp.getS2())));
        // return Math.log(getForegroundProbability(sp)/(getBackgroundProbability(sp)));
        return Math.log(getForegroundProbability(sp) / (getBackgroundProbability(sp)));
    }

    private void scaleScores()
    {
        double min = Double.POSITIVE_INFINITY, max = Double.NEGATIVE_INFINITY, c;
        for (SymbolPairAlignmentData s : this.scores.keySet()) {
            c = this.scores.get(s);
            if (!Double.isInfinite(c) && c < min) {
                min = c;
            }
            if (!Double.isInfinite(c) && c > max) {
                max = c;
            }
        }
        min *= 2;
        max *= 2;
        for (SymbolPairAlignmentData s : this.scores.keySet()) {
            c = this.scores.get(s);
            char aa1 = s.getAA1(), aa2 = s.getAA2();
            if (aa1 == Codon.STOP_CODON || aa1 == 'U' || aa2 == Codon.STOP_CODON || aa2 == 'U') {
                this.scores.put(s, min);
            } else if (Double.isInfinite(c)) {
                if (c < 0) {
                    this.scores.put(s, min);
                } else {

                    this.scores.put(s, max);
                }
            }
        }
        for (SymbolPairAlignmentData s : this.scores.keySet()) {
            c = get(s);
            // Truncate to 2 decimals
            this.scores.put(s, (Math.round((c * 100))) / 100.0);
        }
    }

    // SCORE MATRIX ACCESS
    /**
     * Retrieves the score for a symbol pair.
     */
    public double get(SymbolPairAlignmentData s)
    {
        Double V = this.scores.get(s);
        if (V == null) {
            System.out.println("UNDEFINED: " + s);
        }
        return (V != null) ? V.doubleValue() : Double.NEGATIVE_INFINITY;
    }

    /**
     * The computed score matrix.
     * 
     * @return the matrix instance
     */
    public ContextScoreMatrix getScoreMatrix()
    {
        return this.scores;
    }

    // -------------------------------------------------------------------------------------------
    // META ZONE
    // -------------------------------------------------------------------------------------------
    /**
     * Displays the sums of the probabilities gathered in the temporary matrices.
     */
    private void displayPartialSums(HashMap m[], String title)
    {
        double sums[] = new double[this.indepMatrixCount];
        Arrays.fill(sums, 0);
        double sum = 0;
        for (int i = 0; i < m.length; ++i) {
            for (Object s : m[i].keySet()) {
                sums[i] += (Double) m[i].get(s);
            }
            sum += sums[i];
        }
        System.out.println(title + ": " + sum);
        for (int i = 0; i < this.indepMatrixCount; ++i) {
            System.out.println(this.mapLabels[i] + ": " + String.format("%4.1f", +sums[i]) + "  (" + +sums[i] + ")");
        }
    }

    private void displayScoreExpectation()
    {
        System.out.println("Score expectation:");
        for (int i = 0; i < this.indepMatrixCount; ++i) {
            System.out.println(String.format("  %s: %6.2f", this.mapLabels[i], scoreExpectation(i)));
        }
    }

    private double scoreExpectation(int i)
    {
        double e = 0;
        // we suppose we call this function at a point when we have calculated the scores, but not yet
        // corrected the keys, so that the keys in the background maps correspond to keys in score map
        for (SymbolPairAlignmentData s : this.bgProbsCodonDistrib[i].keySet()) {
            e += getBackgroundProbability(s) * this.scores.get(s);
        }
        return e;
    }

    /**
     * Displays the ratio of the two variants of bg probs
     */
    private ArrayList<HashMap<SymbolPairAlignmentData, Double>> computeBGRatios()
    {
        ArrayList<HashMap<SymbolPairAlignmentData, Double>> ratios =
            new ArrayList<HashMap<SymbolPairAlignmentData, Double>>();
        for (int i = 0; i < this.indepMatrixCount; ++i) {
            ratios.add(computeBGRatios(i));
        }
        return ratios;
    }

    private HashMap<SymbolPairAlignmentData, Double> computeBGRatios(int i)
    {
        HashMap<SymbolPairAlignmentData, Double> ratio = new HashMap<SymbolPairAlignmentData, Double>();

        for (SymbolPairAlignmentData s : this.bgProbsCodonDistrib[i].keySet()) {
            ratio.put(s, this.bgProbsCodonDistrib[i].get(s) / this.bgProbsMarginal[i].get(s));
        }

        // new MatrixAnalyzer().printMap(System.out, ratio, " BG ratio " + mapLabels[i]);
        return ratio;
    }

    private void displayBGRatios()
    {
        ArrayList<HashMap<SymbolPairAlignmentData, Double>> ratios = computeBGRatios();

        MatrixAnalyzer ma = new MatrixAnalyzer();
        for (int i = 0; i < this.indepMatrixCount; ++i) {
            ma.printMap(System.out, this.bgProbsCodonDistrib[i], "BG CD " + this.mapLabels[i]);
            ma.printMap(System.out, this.bgProbsMarginal[i], "BG Mg " + this.mapLabels[i]);
            ma.printMap(System.out, ratios.get(i), "BG Ra: CD/Mg " + this.mapLabels[i]);
            System.out.println("-----------------------------------------------\n\n");
        }
        // System.exit(0);
    }
}
