package sequoia.path.scorematrixbuilder.mutation;

import sequoia.path.scorematrixbuilder.matrix.SquareMatrix;
import sequoia.path.util.Codon;
import sequoia.path.util.GeneticCodeMap;

public class CodonMutator
{
    SquareMatrix mutationProb;

    Codon codons[] = GeneticCodeMap.getInstance().getCodons();

    double[][] mutationChances;

    int unchangedCodons = 0;

    int synMutations = 0;

    int nonsynMutations = 0;

    int unchangedNucleotides = 0;

    int transMNucleotides = 0;

    int transvMNucleotides = 0;

    // ---------------------------------------------------------------
    // CONSTRUCTOR

    public CodonMutator(Codon c[], SquareMatrix m)
    {
        this.codons = c;
        this.mutationProb = m;
        this.mutationChances = new double[m.size()][m.size() + 1];

        for (int ci = 0; ci < m.size(); ++ci) {
            double N = 0;
            for (int i = 0; i < m.size(); ++i) {
                N += m.get(ci, i);
            }
            this.mutationChances[ci][0] = 0;
            for (int i = 0; i < m.size(); ++i) {
                this.mutationChances[ci][i + 1] = this.mutationChances[ci][i] + m.get(ci, i) / N;
            }
            this.mutationChances[ci][this.mutationChances[ci].length - 1] = 1;
            // System.err.println("C2 " + m.size() + mutationChances.length);
        }
    }

    public CodonMutator(SquareMatrix m)
    {
        this(GeneticCodeMap.getInstance().getCodons(), m);
    }

    // --------------------------------------------------------------
    // STATS

    public double getCodonIdentity()
    {
        return ((double) this.unchangedCodons) / (this.unchangedCodons + this.synMutations + this.nonsynMutations);
    }

    public double getAAIdentity()
    {
        return ((double) this.unchangedCodons + this.synMutations)
            / (this.unchangedCodons + this.synMutations + this.nonsynMutations);
    }

    public double getNucleotideIdentity()
    {
        return ((double) this.unchangedNucleotides)
            / (this.unchangedNucleotides + this.transMNucleotides + this.transvMNucleotides);
    }

    public void resetStats()
    {
        this.unchangedCodons = 0;
        this.synMutations = 0;
        this.nonsynMutations = 0;
        this.unchangedNucleotides = 0;
        this.transMNucleotides = 0;
        this.transvMNucleotides = 0;
    }

    public String stats()
    {
        StringBuilder str = new StringBuilder();
        str.append("     Codon identity: ");
        str.append(((int) (getCodonIdentity() * 10000)) / 100.0);
        str.append("%\n");
        str.append("        AA identity: ");
        str.append(((int) (getAAIdentity() * 10000)) / 100.0);
        str.append("%\n");
        str.append("Nucleotide identity: ");
        str.append(((int) (getNucleotideIdentity() * 10000)) / 100.0);
        str.append("%\n");
        return str.toString();
    }

    protected void updateMutationStats(int from, int to)
    {
        CodonRelation.CODON_DIF codonDif = CodonRelation.getCodonDif(from, to);
        if (codonDif == CodonRelation.CODON_DIF.IDENTICAL) {
            ++this.unchangedCodons;
        } else if (codonDif == CodonRelation.CODON_DIF.SYN) {
            ++this.synMutations;
        } else {
            ++this.nonsynMutations;
        }

        CodonRelation.NUCLEOTIDE_DIF nuclDif[] = CodonRelation.getNucleotideDif(from, to);
        for (CodonRelation.NUCLEOTIDE_DIF nd : nuclDif) {
            if (nd == CodonRelation.NUCLEOTIDE_DIF.IDENTICAL) {
                ++this.unchangedNucleotides;
            } else if (nd == CodonRelation.NUCLEOTIDE_DIF.TRANS) {
                ++this.transMNucleotides;
            } else {
                ++this.transvMNucleotides;
            }
        }
    }

    // ---------------------------------------------------------------
    // MUTATION

    protected int randomCodonMutation(int c)
    {
        double r = Math.random();
        for (int i = 1; i < this.mutationChances[c].length; ++i) {
            if (r < this.mutationChances[c][i]) {
                return i - 1;
            }
        }
        return this.mutationChances[c].length - 2;
    }

    public int getMutation(int i)
    {
        int m;
        // do {
        m = randomCodonMutation(i);
        // } while (!GeneticCodeMap.isAACodon(codons[m]));
        updateMutationStats(i, m);
        return m;
    }

    // ---------------------------------------------------------------
    // DISPLAY

    @Override
    public String toString()
    {
        StringBuilder str = new StringBuilder();

        str.append("              ");
        for (Codon c : this.codons) {
            str.append(String.format("%7s", c));
        }
        str.append("\n");

        for (int i = 0; i < this.mutationChances.length; ++i) {
            str.append(this.codons[i] + " ");
            for (int j = 0; j < this.mutationChances[i].length; ++j) {
                str.append(String.format(" %6.4f", this.mutationChances[i][j]));
            }
            str.append("\n");
        }

        str.append("\n--------------------\n\n");

        str.append("       ");
        for (Codon c : this.codons) {
            str.append(String.format("%7s", c));
        }
        str.append("\n");

        for (int i = 0; i < this.mutationProb.size(); ++i) {
            str.append(this.codons[i] + " ");
            for (int j = 0; j < this.mutationProb.size(); ++j) {
                str.append(String.format(" %6.4f", this.mutationProb.get(i, j)));
            }
            str.append("\n");
        }

        return str.toString();
    }

}
