package sequoia.path.scores;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Set;
import java.util.zip.DataFormatException;

import sequoia.path.util.Codon;

public class ContextScoreMatrix extends HashMap<SymbolPairAlignmentData, Double> implements Serializable, ScoreMatrix
{
    public static final long serialVersionUID = 20080720;

    public ContextScoreMatrix()
    {
        super();
    }

    public ContextScoreMatrix(String inputFile)  throws FileNotFoundException, IOException, DataFormatException
    {
        super();
        loadFile(inputFile);
    }
    
    protected void loadFile(String filename) throws FileNotFoundException, IOException, DataFormatException {
    	// message to display when formating errore are found in the input file
    	final String DATA_FORMAT_EXCEPTION_MESSAGE = "Wrongly formatted input file: ";
    	// for verification: 
    	int submatrices = 0;
    	boolean readSubmatrixMarker = false;
    	boolean readSubmatrixHeader = false;
    	
		String line;
		String tokens[];
		boolean rc = false;
		int p1 = -1, p2 = -1;
		ArrayList<SymbolData> symbols = new ArrayList<SymbolData>();
		BufferedReader br = new BufferedReader(new FileReader(filename));
		while ((line = br.readLine()) != null) {
			line = line.trim();
			if (line.startsWith(ScoreMatrixFormatter.COMMENT_LINE_MARKER)) {
				// comment; ignore
				continue;
			}
			if (line.contains(ScoreMatrixFormatter.SUBMATRIX_POSITION_MARKER)) {
				// sub-matrix start
				tokens = line.split("\\s+");
				if (tokens.length > 0) {
				    rc = tokens[tokens.length - 1].equalsIgnoreCase(ScoreMatrixFormatter.REVERSE_COMPLEMENTARY_MARKER);
				    tokens = tokens[0].split(ScoreMatrixFormatter.SUBMATRIX_POSITION_MARKER);
				    if (tokens.length >= 2) {
				    	p1 = Integer.valueOf(tokens[0]);
				    	p2 = Integer.valueOf(tokens[1]);
				    }
				    if (p1 < 0 || p1 >= Codon.CODON_LEN || p2 < 0 || p2 >= Codon.CODON_LEN) {
				    	throw new DataFormatException(DATA_FORMAT_EXCEPTION_MESSAGE + "position outside the codon (" + line +")");
				    }
				    // for file format verification: 
				    ++submatrices;
				    readSubmatrixMarker = true;
					readSubmatrixHeader = false;
				}
			} else if (line.matches("^([a-zA-Z]\\|[a-zA-Z]\\s+)+([a-zA-Z]\\|[a-zA-Z])")) {
				// matrix header line
				// verification of file format
				if (!readSubmatrixMarker) {
					throw new DataFormatException(DATA_FORMAT_EXCEPTION_MESSAGE + "msiplaced submatrix header (" + line +")");
				}
				readSubmatrixMarker = false;
				readSubmatrixHeader = true;
				// verification done
				
				symbols.clear();
				tokens = line.split("\\s+");
				for (String token : tokens) {
					String pair[] = token.split("\\|");
					symbols.add(new SymbolData(pair[0].charAt(0), pair[1].charAt(0), p2, rc));
				}
			} else if (line.matches("^[a-zA-Z]\\|[a-zA-Z](\\s+-?[0-9]+\\.[0-9]+)+")) {
				// regular matrix line
				tokens = line.split("\\s+");
				// verification:
				if (!readSubmatrixHeader) {
				    throw new DataFormatException(DATA_FORMAT_EXCEPTION_MESSAGE + "misplaced submatrix line, without submatrix header (" + line +")");
			    }
				if (tokens.length != symbols.size() + 1) {
					throw new DataFormatException(DATA_FORMAT_EXCEPTION_MESSAGE + "wrong number of values found in matrix row. Expected " + symbols.size() + ", found " + (tokens.length - 1) + " (" + line +")");
				}
				// done verification
				String pair[] = tokens[0].split("\\|");
				SymbolData symb = new SymbolData(pair[0].charAt(0), pair[1].charAt(0), p1, false);
				for (int i = 1; i < tokens.length; ++i) {
					this.put(new SymbolPairAlignmentData(symb, symbols.get(i-1)), Double.valueOf(tokens[i]));
				}
			}
		}
		if (submatrices != 18) {
			throw new DataFormatException(DATA_FORMAT_EXCEPTION_MESSAGE + "wrong number of submatrices found. Expected 18, found " + submatrices);
		}
    }
    
    public String getNPair(SymbolData s1, SymbolData s2)
    {
        return s1.getN() + "" + s2.getN();
    }

    public String getNPair(SymbolPairAlignmentData spd)
    {
        return spd.getN1() + "" + spd.getN2();
    }

    public double getScore(SymbolData s1, SymbolData s2)
    {
        return super.get(new SymbolPairAlignmentData(s1, s2));
    }

    public double getScore(SymbolPairAlignmentData spd)
    {
        return super.get(spd);
    }

    public Double get(SymbolData s1, SymbolData s2)
    {
        return super.get(new SymbolPairAlignmentData(s1, s2));
    }

    public Double get(SymbolPairAlignmentData spd)
    {
        return super.get(spd);
    }

    public String getNPair(char n1, char aa1, int p1, boolean rc1, char n2, char aa2, int p2, boolean rc2)
    {
        return n1 + "" + n2;
    }

    public double getScore(char n1, char aa1, int p1, boolean rc1, char n2, char aa2, int p2, boolean rc2)
    {
        return getScore(new SymbolData(n1, aa1, p1, rc1), new SymbolData(n2, aa2, p2, rc2));
    }

    public Double get(char n1, char aa1, int p1, boolean rc1, char n2, char aa2, int p2, boolean rc2)
    {
        return get(new SymbolData(n1, aa1, p1, rc1), new SymbolData(n2, aa2, p2, rc2));
    }
    
    public Set<SymbolPairAlignmentData> getAllSymbolPairs() {
    	return keySet();
    }
    
    public void set(SymbolPairAlignmentData spd, double score) {
    	put(spd, score);
    }
    @Override
    public String toString()
    {
    	return ScoreMatrixFormatter.format(this);
    }
}
