package sequoia.path.scores;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;
import java.util.Formatter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import sequoia.path.util.Codon;
import sequoia.path.util.GenomicSequenceUtils;

public class SimpleScoreMatrix implements Serializable, ScoreMatrix
{
    public static final long serialVersionUID = 20080721;

    protected static final String DEFAULT_SYMBOLS = GenomicSequenceUtils.BASIC_SYMBOLS;

    double score[][];

    char commonSymbols[][];

    String symbols = null;
    
    private SimpleScoreMatrix(String symbols, char commonSymbols[][], double score[][]) {
    	this.symbols = symbols;
    	this.commonSymbols = new char[commonSymbols.length][];
    	for (int i = 0; i < commonSymbols.length; ++i) {
    		this.commonSymbols[i] = new char[commonSymbols[i].length];
    		for (int j = 0; j < commonSymbols[i].length; ++j) {
    			this.commonSymbols[i][j] = commonSymbols[i][j];
    		}
    	}
		this.score = new double[score.length][];
    	for (int i = 0; i < score.length; ++i) {
    		this.score[i] = new double[score[i].length];
    		for (int j = 0; j < score[i].length; ++j) {
    			this.score[i][j] = score[i][j];
    		}
    	}
    }

    public SimpleScoreMatrix(int match, int mismatch, int tmismatch)
    {
        this(match, mismatch, tmismatch, GenomicSequenceUtils.BASIC_SYMBOLS, GenomicSequenceUtils.EXTENDED_SYMBOLS,
            GenomicSequenceUtils.IUPACmapping);
    }

    public SimpleScoreMatrix(int match, int mismatch, int tmismatch, String basicSymbols, String extendedSymbols,
        HashMap<Character, HashSet<Character>> extendedSymbolsMap)
    {
        symbols = extendedSymbols;
        score = new double[symbols.length()][symbols.length()];
        for (int i = 0; i < basicSymbols.length(); ++i) {
            score[i][i] = match;
            char si = symbols.charAt(i);
            for (int j = i + 1; j < basicSymbols.length(); ++j) {
                char sj = symbols.charAt(j);
                if (GenomicSequenceUtils.isTransitionMutation(si, sj)) {
                    score[i][j] = tmismatch;
                    score[j][i] = tmismatch;
                } else {
                    score[i][j] = mismatch;
                    score[j][i] = mismatch;
                }
            }
        }
        extendMatrix(basicSymbols, extendedSymbolsMap);
    }

    public SimpleScoreMatrix(String file)
    {
        extendMatrix(load(file), GenomicSequenceUtils.IUPACmapping);
    }

    public SimpleScoreMatrix(String file, String Symbols)
    {
        symbols = Symbols;
        extendMatrix(load(file), GenomicSequenceUtils.IUPACmapping);
    }

    public SimpleScoreMatrix(String file, String Symbols, HashMap<Character, HashSet<Character>> extendedSymbols)
    {
        symbols = Symbols;
        extendMatrix(load(file), extendedSymbols);
    }

    private String load(String file)
    {
        String loaded = "";
        try {
            BufferedReader br = new BufferedReader(new FileReader(file));
            String line = br.readLine();
            String cols[] = line.split("\\s+");
            score = new double[cols.length][cols.length];
            if (symbols == null) {
                symbols = line.replaceAll("\\s+", "");
            }
            loaded = line.replaceAll("\\s+", "");
            line = br.readLine();
            while (line != null) {
                String[] toks = line.split("\\s+");
                int rowIndex = symbols.indexOf(toks[0].charAt(0));
                if (rowIndex != -1) {
                    for (int i = 1; i < toks.length; ++i) {
                        int colIndex = symbols.indexOf(cols[i].charAt(0));
                        if (colIndex != -1) {
                            score[rowIndex][colIndex] = Double.parseDouble(toks[i]);
                        }
                    }
                }
                line = br.readLine();
            }
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
        return loaded;
    }

    private void extendMatrix(String loadedSymbols, HashMap<Character, HashSet<Character>> extendedSymbols)
    {
        double score, scorePiece, maxNeg, max;
        for (int i = 0; i < symbols.length(); ++i) {
            char si = symbols.charAt(i);
            // if (loadedSymbols.indexOf(si) < 0) {
            HashSet<Character> siSet = extendedSymbols.get(si);
            for (int j = i; j < symbols.length(); ++j) {
                char sj = symbols.charAt(j);
                if (loadedSymbols.indexOf(sj) < 0) {
                    HashSet<Character> sjSet = extendedSymbols.get(sj);
                    // -----------------------------
                    score = 0;
                    maxNeg = Double.NEGATIVE_INFINITY;
                    max = Double.NEGATIVE_INFINITY;
                    for (char ni : siSet) {
                        for (char nj : sjSet) {
                            if ((scorePiece = get(ni, nj)) > 0) {
                                score += scorePiece;
                            } else if (scorePiece > maxNeg) {
                                maxNeg = scorePiece;
                            }
                            if (scorePiece > max) {
                                max = scorePiece;
                            }
                        }
                    }
                    // set(i, j, (score > 0) ? score: maxNeg);
                    set(i, j, max);
                    // -----------------------------
                }
            }
            // }
        }
    }

    public double get(char a, char b)
    {
        try {
            return score[symbols.indexOf(a)][symbols.indexOf(b)];
        } catch (Exception e) {
            // e.printStackTrace();
            // System.out.println(a + " " + b); System.exit(0);
            return 0;
        }
    }

    // private void set(char a, char b, double val){
    // try {
    // score[symbols.indexOf(a)][symbols.indexOf(b)] = val;
    // score[symbols.indexOf(b)][symbols.indexOf(a)] = val;
    // } catch (Exception e) {
    // //e.printStackTrace();
    // //System.out.println(a + " " + b); System.exit(0);
    // }
    // }
    private void set(int i, int j, double val)
    {
        try {
            score[i][j] = val;
            score[j][i] = val;
        } catch (Exception e) {
            // e.printStackTrace();
            // System.out.println(a + " " + b); System.exit(0);
        }
    }

    @Override
    public String toString()
    {
        StringBuilder str = new StringBuilder();

        str.append("     ");

        for (int ia2 = 0; ia2 < symbols.length(); ++ia2) {
            str.append(new Formatter().format("    %c  ", symbols.charAt(ia2)));
        }
        str.append("\n");
        for (int ia1 = 0; ia1 < symbols.length(); ++ia1) {
            str.append(new Formatter().format("%c   ", symbols.charAt(ia1)));
            for (int ia2 = 0; ia2 < symbols.length(); ++ia2) {
                str.append(new Formatter().format("%6.3g ", score[ia1][ia2]));
            }
            str.append("\n");
        }
        return str.toString();
    }

    public String getNPair(char n1, char aa1, int p1, boolean rc1, char n2, char aa2, int p2, boolean rc2)
    {
        return n1 + "" + n2;
    }

    public String getNPair(SymbolData s1, SymbolData s2)
    {
        return s1.getN() + "" + s2.getN();
    }

    public String getNPair(SymbolPairAlignmentData spd)
    {
        return spd.getN1() + "" + spd.getN2();
    }

    public double getScore(char n1, char aa1, int p1, boolean rc1, char n2, char aa2, int p2, boolean rc2)
    {
        return get(n1, n2);
    }

    public double getScore(SymbolData s1, SymbolData s2)
    {
        return get(s1.getN(), s2.getN());
    }

    public double getScore(SymbolPairAlignmentData spd)
    {
        return get(spd.getN1(), spd.getN2());
    }

    public Object get(char n1, char aa1, int p1, boolean rc1, char n2, char aa2, int p2, boolean rc2)
    {
        return new Double(get(n1, n2));
    }

    public Object get(SymbolData s1, SymbolData s2)
    {
        return new Double(get(s1.getN(), s2.getN()));
    }

    public Object get(SymbolPairAlignmentData spd)
    {
        return new Double(get(spd.getN1(), spd.getN2()));
    }
    
    public Set<SymbolPairAlignmentData> getAllSymbolPairs() {
    	HashSet<SymbolPairAlignmentData> all = new HashSet<SymbolPairAlignmentData>();
    	for (char s1 : symbols.toCharArray()) {
    		for (char s2 : symbols.toCharArray()) {
    			all.add(new SymbolPairAlignmentData(s1, s2, Codon.ANY_AA, Codon.ANY_AA, -1, -1, false));
    		}
    	}
    	return all;
    }
    
    public void set(SymbolPairAlignmentData spd, double score) {
    	try {
            this.score[symbols.indexOf(spd.getN1())][symbols.indexOf(spd.getN2())] = score;
        } catch (Exception e) {
            // e.printStackTrace();
            // System.out.println(a + " " + b); System.exit(0);
        }
    }
    
    public SimpleScoreMatrix clone() {
    	return new SimpleScoreMatrix(this.symbols, this.commonSymbols, this.score);
    }
    
    public void clear() {
    	// TODO
    }
}
