package sequoia.path.translation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map.Entry;

import sequoia.path.graph.BackTranslationGraph;
import sequoia.path.util.Codon;
import sequoia.path.util.DBSequence;
import sequoia.path.util.GeneticCodeMap;
import sequoia.path.util.GenomicSequenceUtils;

public class BackTranslator
{
    public static final int BACK_TRANSLATE_PROTEIN = 0;

    public static final int BACK_TRANSLATE_DNA = 1;

    public static final int BACK_TRANSLATE_RNA = 2;

    public static final int BACK_TRANSLATE_UNDEFINED = -1;

    private static final int DEFAULT_SEQ_TYPE = BACK_TRANSLATE_UNDEFINED;

    public static final boolean COMPRESS_GRAPHS = true;

    public static final boolean DO_NOT_COMPRESS_GRAPHS = false;

    public static final boolean DEFAULT_GRAPH_FORMAT = DO_NOT_COMPRESS_GRAPHS;

    public static final boolean EXPAND_GENOMIC_SEQUENCE_GRAPHS = true;

    public static final boolean DO_NOT_EXPAND_GENOMIC_SEQUENCE_GRAPHS = false;

    public static final boolean DEFAULT_EXPAND_GENOMIC_SEQUENCE_GRAPHS = EXPAND_GENOMIC_SEQUENCE_GRAPHS;

    public final HashMap<Character, BackTranslationGraph> codonGraphMap =
        new HashMap<Character, BackTranslationGraph>();

    private boolean compressGraphs = DEFAULT_GRAPH_FORMAT;

    private boolean expandGenomicSequenceGraphs = DEFAULT_EXPAND_GENOMIC_SEQUENCE_GRAPHS;

    private int seqType = DEFAULT_SEQ_TYPE;

    public BackTranslator()
    {
        buildAAGraphMap();
    }

    public BackTranslator(boolean compress)
    {
        this.compressGraphs = compress;
        buildAAGraphMap();
    }

    public BackTranslator(boolean compress, boolean expandGSG)
    {
        this.compressGraphs = compress;
        this.expandGenomicSequenceGraphs = expandGSG;
        buildAAGraphMap();
    }

    protected BackTranslationGraph buildGraph(ArrayList<Codon> codons, char AA)
    {
        return new BackTranslationGraph(codons, AA, this.compressGraphs);
    }

    public BackTranslationGraph backTranslate(DBSequence Sequence)
    {
        int seqType = this.seqType;
        if (seqType == BACK_TRANSLATE_UNDEFINED) {
            seqType = guessSeqType(Sequence.getSeq());
        }
        return backTranslate(Sequence, seqType);
    }

    private void buildAAGraphMap()
    {
        for (Entry<String, ArrayList<Codon>> entry : GeneticCodeMap.getInstance().A2CentrySet()) {
            BackTranslationGraph aag = buildGraph(entry.getValue(), entry.getKey().charAt(0));
            this.codonGraphMap.put(entry.getKey().toLowerCase().charAt(0), aag.reverseComplementary());
            this.codonGraphMap.put(entry.getKey().charAt(0), aag);
        }
    }

    // TODO this is pretty ugly, can be easily fooled; write a prettier one
    protected int guessSeqType(String seq)
    {
        // we consider that the sequences are large enough to distinguish
        // between types
        HashSet<Character> symbols = new HashSet<Character>();
        String uSeq = seq.toUpperCase();
        for (int i = 0; i < uSeq.length(); ++i) {
            symbols.add(new Character(uSeq.charAt(i)));
            // check from time to time...
            if (i % 10 == 0 && symbols.size() > 4) {
                return BACK_TRANSLATE_PROTEIN;
            }
        }
        if (symbols.size() > 4) {
            return BACK_TRANSLATE_PROTEIN;
        } else if (symbols.contains(new Character('U'))) {
            return BACK_TRANSLATE_RNA;
        }
        return BACK_TRANSLATE_DNA;
    }

    protected BackTranslationGraph getCodonGraph(char AA)
    {
        return this.codonGraphMap.get(AA);
    }

    public BackTranslationGraph backTranslate(DBSequence Sequence, int seqType)
    {
        switch (seqType) {
            case BACK_TRANSLATE_RNA:
                return backTranslateRNA(Sequence);
            case BACK_TRANSLATE_DNA:
                return backTranslateDNA(Sequence);
            default:
                return backTranslateProtein(Sequence);
        }
    }

    protected BackTranslationGraph backTranslateProtein(DBSequence theAASequence)
    {
        String AASequence = theAASequence.getSeq();
        //System.out.println(" >>>> " + theAASequence.getId());
        BackTranslationGraph dnaseq = new BackTranslationGraph(AASequence.length(), true, theAASequence.getId());
        for (int i = 0; i < AASequence.length(); ++i) {
            if (getCodonGraph(AASequence.charAt(i)) != null) {
                dnaseq.append(getCodonGraph(AASequence.charAt(i)));
            } else {
                System.out.println("<<<" + AASequence.charAt(i) + ">>>" + i);
            }
        }
        return dnaseq;
    }

    public BackTranslationGraph backTranslateDNA(DBSequence DNASequence)
    {
        return backTranslateDNAcoding(DNASequence);
    }

    public BackTranslationGraph backTranslateDNAcoding(DBSequence DNASequence)
    {
        int codingStart = 0, codingEnd = 0;
        while (codingStart < DNASequence.getSeq().length()
            && Character.isLowerCase(DNASequence.getSeq().charAt(codingStart))) {
            ++codingStart;
        }
        int phase = codingStart % Codon.CODON_LEN;
        codingEnd = codingStart;
        if (codingEnd < DNASequence.getSeq().length() - 1) {
            while (codingEnd < DNASequence.getSeq().length()
                && Character.isUpperCase(DNASequence.getSeq().charAt(codingEnd))) {
                ++codingEnd;
            }
            codingEnd -= (codingEnd - codingStart) % Codon.CODON_LEN;
        }
        String prefix = DNASequence.getSeq().substring(0, codingStart).toUpperCase();
        String protein = Translator.translate(DNASequence.getSeq().substring(codingStart, codingEnd), 0);
        String suffix = DNASequence.getSeq().substring(codingEnd).toUpperCase();
        BackTranslationGraph result = new BackTranslationGraph(DNASequence.getSeq().length(), DNASequence.getId());
        if (prefix.length() > 0) {
            result.append(new BackTranslationGraph(prefix));
        }
        if (this.expandGenomicSequenceGraphs) {
            System.out.println(protein);
            for (int i = 0; i < protein.length(); ++i) {
                result.append(getCodonGraph(protein.charAt(i)));
            }
        } else {
            int j = codingStart;
            for (int i = 0; i < protein.length(); ++i, j += Codon.CODON_LEN) {
                result.append(new BackTranslationGraph(new Codon(
                    DNASequence.getSeq().substring(j, j + Codon.CODON_LEN), 1.0, protein.charAt(i))));
            }
        }
        if (suffix.length() > 0) {
            result.append(new BackTranslationGraph(suffix));
        }
        result.setPhase(phase);
        return result;
    }

    public BackTranslationGraph backTranslateDNA(DBSequence DNASequence, int readingFrame)
    {
        if (readingFrame >= DNASequence.getSeq().length()) {
            readingFrame = 0;
        }
        String prefix = DNASequence.getSeq().substring(0, readingFrame);
        String protein = Translator.translate(DNASequence.getSeq(), readingFrame);
        String suffix = DNASequence.getSeq().substring(readingFrame + protein.length() * Codon.CODON_LEN);
        BackTranslationGraph result = new BackTranslationGraph(DNASequence.getSeq().length(), DNASequence.getId());
        if (prefix.length() > 0) {
            result.append(new BackTranslationGraph(prefix));
        }
        if (this.expandGenomicSequenceGraphs) {
            System.out.println(protein);
            for (int i = 0; i < protein.length(); ++i) {
                result.append(getCodonGraph(protein.charAt(i)));
            }
        } else {
            int j = readingFrame;
            for (int i = 0; i < protein.length(); ++i, j += Codon.CODON_LEN) {
                result.append(new BackTranslationGraph(new Codon(
                    DNASequence.getSeq().substring(j, j + Codon.CODON_LEN), 1.0, protein.charAt(i))));
            }
        }
        if (suffix.length() > 0) {
            result.append(new BackTranslationGraph(suffix));
        }
        result.setPhase(readingFrame);
        return result;
    }

    public BackTranslationGraph backTranslateRNA(DBSequence RNASequence)
    {
        return backTranslateRNA(RNASequence, 0);
    }

    public BackTranslationGraph backTranslateRNA(DBSequence RNASequence, int readingFrame)
    {
        return backTranslateDNA(GenomicSequenceUtils.rna2dna(RNASequence), readingFrame);
    }

    public int getSeqType()
    {
        return this.seqType;
    }

    public void setSeqType(int seqType)
    {
        this.seqType = seqType;
    }
}
