package sequoia.path.util;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.Map.Entry;

public class GeneticCodeMap
{
    // ------------------------------------------------------------------------------------
	// Constants
	/** String that marks the start codon in the input files */
	public static final String START_CODON = "START";

	/** String that marks a stop codon in the input files */
    public static final String STOP_CODON = "STOP";

    /**  */
    public static final String UNKNOWN_CODON = "___";

    public static final String ANY_AA = "X";
    
    /** Resource file containing the codon - amino acid correspondence. */
    protected static final String inputFileName = "codon_usage";
    /** Resource file containing the codon - amino acid correspondence -- ambiguous nucleotide symbols. */
    protected static final String inputFileNameAmbiguousSymbols = "codon_usage_amb";
    
    //-------------------------------------------------------------------------------------
    // Access to a genetic code map
    protected static GeneticCodeMap instance = null;
    protected static GeneticCodeMap instanceAmbiguousSymbols = null;
    
    /**
     * Access to an instance of the GeneticCodeMap
     * @return The non-ambiguous map, if the parameter is false; the ambiguous map, if the parameter is true. 
     */
    public static GeneticCodeMap getInstance(boolean ambigous) {
    	return ambigous ? getInstanceAmbiguous() : getInstance();
    }
    public static GeneticCodeMap getInstance() {
    	if (instance == null) {
    		instance = new GeneticCodeMap(inputFileName);
    	}
    	return instance;
    }
    public static GeneticCodeMap getInstanceAmbiguous() {
    	if (instanceAmbiguousSymbols == null) {
    		instanceAmbiguousSymbols = new GeneticCodeMap(inputFileNameAmbiguousSymbols);
    	}
    	return instanceAmbiguousSymbols;
    }
    // ------------------------------------------------------------------------------------
    // Instance fields. Amino acid <-> codon maps for quick access
    /** Codons are mapped to their corresponding amino acids  */
    protected final HashMap<Codon, String> C2A = new HashMap<Codon, String>();

    /** Each amino acid has a corresponding list of codons */
    protected final HashMap<String, ArrayList<Codon>> A2C = new HashMap<String, ArrayList<Codon>>();
    
    /** Amino acid frequencies. for internal use only. */
    protected final HashMap<String, Double> aafreqs = new HashMap<String, Double>();

    // ------------------------------------------------------------------------------------
    // Constructor
    /**
     * The constructor is for internal use only.
     * It builds the Amino acid <-> codon maps.
     */
    protected GeneticCodeMap(String resourceFile) {
    	HashMap<String, String> ambigousAAs = new HashMap<String, String>();
        ambigousAAs.put("Q", "Z");
        ambigousAAs.put("E", "Z");
        ambigousAAs.put("N", "B");
        ambigousAAs.put("D", "B");
        ambigousAAs.put("I", "J");
        ambigousAAs.put("L", "J");
        
        C2A.put(new Codon(UNKNOWN_CODON, 0.0, ANY_AA.charAt(0)), ANY_AA);
        
        ArrayList<Codon> unknownCodons = new ArrayList<Codon>();
        unknownCodons.add(new Codon(UNKNOWN_CODON, 0.0, ANY_AA.charAt(0)));
        A2C.put(ANY_AA, unknownCodons);

        try {
            String line;
            String tokens[];
            String aa = "";
            double f = 1;

            BufferedReader bf =
                new BufferedReader(new InputStreamReader(ClassLoader.getSystemResourceAsStream(resourceFile)));
            line = bf.readLine();
            while (line != null) {
                Codon c;
                tokens = line.split("\\s+");
                try {
                    f = Double.parseDouble(tokens[2]) * 0.01;
                } catch (NumberFormatException nfex) {
                    nfex.printStackTrace();
                    f = 0;
                }
                try {
                    aa = (tokens[1].split("/"))[1];
                    c = new Codon(tokens[0], f, aa.charAt(0));
                } catch (ArrayIndexOutOfBoundsException aiobex) {
                    aa = tokens[1];
                    if (tokens[1].equals(START_CODON)) {
                        aa = Codon.START_CODON + "";
                        c = new Codon(tokens[0], f, Codon.START_CODON);
                    } else {
                        aa = Codon.STOP_CODON + "";
                        c = new Codon(tokens[0], f, Codon.STOP_CODON);
                    }
                }
                if (isNotAASet(aa)) {
                    C2A.put(c, aa);
                }
                putData(c, aa, aafreqs);
                String sAA = ambigousAAs.get(aa);
                if (sAA != null) {
                    putData(c.createCopy(), sAA, aafreqs);
                }

                line = bf.readLine();
            }
            bf.close();

            Object[] allAAs = A2C.keySet().toArray();
            for (int i = 0; i < allAAs.length; ++i) {
                try {
                    double aaf = aafreqs.get(allAAs[i]).doubleValue();
                    ArrayList<Codon> cs = A2C.get(allAAs[i]);
                    for (int j = 0; j < cs.size(); ++j) {
                        Codon c = cs.get(j);

                        if (isNotAASet((String) allAAs[i])) {
                            c.setFrequency(c.getFrequency() / aaf);
                            C2A.put(c, (String) allAAs[i]);
                        }
                    }
                } catch (NullPointerException npex) {
                    // System.out.println(npex.getMessage() + " " + allAAs[i]);
                }
            }
        } catch (IOException ex) {
            ex.printStackTrace();
        }
    }

    public boolean isNotAASet(String AA)
    {
        return (!AA.toLowerCase().equals(AA) || AA.equals(">")) && !AA.equals("X") && !AA.equals("B")
            && !AA.equals("Z") && !AA.equals("J");
        // && !AA.equals("START") && !AA.equals("STOP");
    }

    public boolean isSingleAA(String AA)
    {
        return (!AA.toLowerCase().equals(AA)) && !AA.equals("X") && !AA.equals("B") && !AA.equals("Z")
            && !AA.equals("J");
    }

    protected void putData(Codon c, String aa, HashMap<String, Double> aafreqs)
    {
        ArrayList<Codon> cs = A2C.get(aa);
        if (cs == null) {
            cs = new ArrayList<Codon>();
            A2C.put(aa, cs);
        }
        cs.add(c);

        Double aaf = aafreqs.get(aa);
        double aafval = 0;
        if (aaf != null) {
            aafval = aaf.doubleValue();
        }
        aafreqs.put(aa, new Double(aafval + c.getFrequency()));
    }

    public String getAA(String codon)
    {
        return getAA(new Codon(codon));
    }

    public String getAA(Codon codon)
    {
        return C2A.get(codon);
    }

    public char[] getAAs()
    {
        HashSet<String> aas = new HashSet<String>();
        for (String aa : A2C.keySet()) {
            if (isSingleAA(aa)) {
                aas.add(aa);
            }
        }
        char[] aasv = new char[aas.size()];
        int i = 0;
        for (String aa : aas) {
            aasv[i++] = aa.charAt(0);
        }
        Arrays.sort(aasv);
        return aasv;
    }

    public ArrayList<Codon> getCodons(String AA)
    {
        ArrayList<Codon> c = A2C.get(AA);
        if (c == null) {
            c = getReversedComplementaryCodons(AA);
        }
        return c; // (A2C.get(AA));
    }

    public Codon[] getCodons()
    {
        Object ocs[] = C2A.keySet().toArray();
        Codon codons[] = new Codon[ocs.length - 1];
        int i = 0;
        for (Object c : ocs) {
            if (((Codon) c).getAA() != Codon.ANY_AA) {
                codons[i++] = (Codon) c;
            }
        }
        Arrays.sort(codons);
        return codons;
    }

    public Codon[] getAACodons()
    {
        Object ocs[] = C2A.keySet().toArray();
        Codon codons[] = new Codon[ocs.length - 3];
        int i = 0;
        for (Object c : ocs) {
            if (isAACodon((Codon) c)) {
                codons[i++] = (Codon) c;
            }
        }
        Arrays.sort(codons);
        return codons;
    }

    public boolean isAACodon(Codon c)
    {
        return (c).getAA() != Codon.ANY_AA && (c).getAA() != Codon.STOP_CODON;
    }

    public ArrayList<Codon> getReversedComplementaryCodons(String AA)
    {
        ArrayList<Codon> c = A2C.get(AA.toUpperCase());
        ArrayList<Codon> rcc = new ArrayList<Codon>(c.size());
        for (int i = 0; i < c.size(); ++i) {
            Codon o = c.get(i);
            rcc
                .add(new Codon(GenomicSequenceUtils.getReverseComplementary(o.getSeq()), o.getFrequency(), AA.charAt(0)));
        }
        return rcc; // (A2C.get(AA));
    }

    public ArrayList<Codon> getCodons(char AA)
    {
        return (A2C.get(AA + ""));
    }
    
    public Set<Entry<String, ArrayList<Codon>>> A2CentrySet() {
    	return this.A2C.entrySet();
    }
    
    public Set<Entry<Codon,String>> C2AentrySet() {
    	return this.C2A.entrySet();
    }
}
